/*
 *	HT Editor
 *	htapp.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTAPP_H__
#define __HTAPP_H__

#include "htctrl.h"
#include "htdialog.h"
#include "htviewer.h"

#define cmd_quit					100
#define cmd_load					101

#define cmd_clipboard_copy			200
#define cmd_clipboard_paste			201
#define cmd_clipboard_show			202
#define cmd_clipboard_clear			203
#define cmd_clipboard_paste_into_file	204
#define cmd_clipboard_copy_from_file	205

#define cmd_options_colors			300

#define cmd_help					400

#define VIEWERGROUP_NAME		"viewergroup"

/*
 *	CLASS ht_keyline
 */

class ht_keyline: public ht_view {
public:
			void init(bounds *b);
	virtual	void done();
/* overwritten */
	virtual	void defaultpalette();
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_desktop
 */

class ht_desktop: public ht_view {
public:
			void init(bounds *b);
	virtual	void done();
/* overwritten */
	virtual	void defaultpalette();
	virtual	void draw();
};

/*
 *	CLASS ht_app
 */

class ht_app: public ht_group {
private:
	ht_window *logwindow;
	ht_logviewer *logviewer;
	ht_view *menu;
	ht_keyline *keyline;
	ht_desktop *desktop;

	ht_group *battlefield;

	ht_window *findwindowbynumber(int index);
	int findfreewinnumber();
public:
			void init(bounds *b);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	char *func(int i, int getname);
	virtual	void handlemsg(htmsg *msg);
			int load(ht_stream *f);
	virtual	void store(ht_stream *f);
/* new */
			ht_window *create_window(char *filename);
			void	delete_window(ht_window *window);
			void	focus_window(int index);

			int logf(const char *format, ...);
			void	popup_clipboard_viewer();
			void	popup_color_list();
			void	popup_view_list();
			void	popup_view_list_dump(ht_view *view, ht_text_listbox *listbox, ht_collection *structure, int depth, int *currenti, ht_view *currentv);
			void	popup_window_list();
};

/*
 * exports
 */

void jump_dialog(ht_viewer *viewer);
void search_dialog(ht_viewer *viewer, STREAMOFS current_ofs);

extern char *globalerror;

#define LOG(a...) ((ht_app*)app)->logf(a);

/*
 *	INIT
 */

int init_app();

/*
 *	DONE
 */

void done_app();

#endif /* __HTAPP_H__ */
