/*
 *	HT Editor
 *	htclipboard.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htclipboard.h"
#include "htstring.h"
#include "tools.h"

#include <stdlib.h>
#include <time.h>

ht_clipboard *clipboard;

class ht_clipboard_copy_history: public ht_data {
public:
	virtual ~ht_clipboard_copy_history() {
		if (source) delete source;
	}

	char *source;
	dword start;
	dword size;
	time_t time;
};

/*
 *	CLASS ht_clipboard
 */

void ht_clipboard::init()
{
	ht_mem_stream::init(0, 16);
	copy_history=new ht_collection();
	clipboard_start=0;
	clipboard_size=0;
}

void ht_clipboard::done()
{
	delete copy_history;
	ht_mem_stream::done();
}

/*
 *	CLASS ht_clipboard_viewer
 */

void ht_clipboard_viewer::init(bounds *b, char *desc, ht_clipboard *clipboard, int caps)
{
	ht_uviewer::init(b, desc, clipboard, caps);

	int c=clipboard->copy_history->count();
	char title[512]; /* possible buffer overflow ! */
	if (!c) {
// FIXME: uviewer bug when no selectables !
/*		sprintf(title, "<empty clipboard>");

		hts_mask *m=new hts_mask();
		m->init(clipboard);
		m->add_mask(title);
		insertsub(m);*/
	}
	for (int i=0; i<c; i++) {
		ht_clipboard_copy_history *j=(ht_clipboard_copy_history*)clipboard->copy_history->get(i);

		tm *t=localtime(&j->time);
		sprintf(title, "*** %02d:%02d:%02d, size %ld(%lxh), from %s", t->tm_hour, t->tm_min, t->tm_sec, j->size, j->size, j->source);

		hts_mask *m=new hts_mask();
		m->init(clipboard);
		m->add_mask(title);
		insertsub(m);

		hts_hex *h=new hts_hex();
		h->init(clipboard, j->start, j->size);
		insertsub(h);
	}

	sendmsg(htm_complete_init, 0);
	select_add(clipboard->clipboard_start, clipboard->clipboard_start+clipboard->clipboard_size);
}

/* clipboard functions */

void clipboard_add_copy_history_entry(char *source, dword start, dword size, time_t time)
{
	ht_clipboard_copy_history *h=new ht_clipboard_copy_history();
	h->source=strdup(source);
	h->start=start;
	h->size=size;
	h->time=time;
	clipboard->copy_history->insert(h);
}

#define CLIPBOARD_TRANSFER_BUF_SIZE	32*1024
//#define CLIPBOARD_TRANSFER_BUF_SIZE	2

int clipboard_copy(char *source_desc, void *buf, dword len)
{
	dword size=clipboard->get_size();
	clipboard->seek(size);
	int r=clipboard->write(buf, len);
	clipboard->clipboard_start=size;
	clipboard->clipboard_size=r;
	clipboard_add_copy_history_entry(source_desc, size, r, time(0));
	return r;
}

int clipboard_copy(char *source_desc, ht_stream *stream, dword offset, dword len)
{
	dword size=clipboard->get_size();
	dword temp=stream->tell();
	dword cpos=size, spos=offset;
	byte *buf=(byte*)malloc(CLIPBOARD_TRANSFER_BUF_SIZE);
	dword l=len, r=0;
	while ((len) && (l)) {
		l=len;
		if (l>CLIPBOARD_TRANSFER_BUF_SIZE) l=CLIPBOARD_TRANSFER_BUF_SIZE;
		stream->seek(spos);
		l=stream->read(buf, l);
		spos+=l;
		clipboard->seek(cpos);
		clipboard->write(buf, l);
		cpos+=l;
		len-=l;
		r+=l;
	}
	stream->seek(temp);
	clipboard->clipboard_start=size;
	clipboard->clipboard_size=r;
	clipboard_add_copy_history_entry(source_desc, size, r, time(0));
	delete buf;
	return r;
}

int clipboard_paste(void *buf, dword maxlen)
{
	clipboard->seek(clipboard->clipboard_start);
	return clipboard->read(buf, MIN(clipboard->clipboard_size, maxlen));
}

int clipboard_paste(ht_stream *stream, dword offset)
{
	dword len=clipboard->clipboard_size;
	dword temp=stream->tell();
	dword cpos=clipboard->clipboard_start, spos=offset;
	byte *buf=(byte*)malloc(CLIPBOARD_TRANSFER_BUF_SIZE);
	dword l=len, r=0;
	while ((len) && (l)) {
		l=len;
		if (l>CLIPBOARD_TRANSFER_BUF_SIZE) l=CLIPBOARD_TRANSFER_BUF_SIZE;
		clipboard->seek(cpos);
		l=clipboard->read(buf, l);
		cpos+=l;
		stream->seek(spos);
		stream->write(buf, l);
		spos+=l;
		len-=l;
		r+=l;
	}
	stream->seek(temp);
	delete buf;
	return r;
}

int clipboard_clear()
{
	clipboard->done();
	delete clipboard;
	clipboard=new ht_clipboard();
	clipboard->init();
	return 1;
}

dword clipboard_getsize()
{
	return clipboard->clipboard_size;
}

/*
 *	INIT
 */

int init_clipboard()
{
	clipboard=new ht_clipboard();
	clipboard->init();
	return 1;
}

/*
 *	DONE
 */

int done_clipboard()
{
	clipboard->done();
	delete clipboard;
	return 1;
}

