/* 
 *	HT Editor
 *	htdata.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTDATA_H__
#define __HTDATA_H__

class ht_tree;

#include "global.h"
#include "htdebug.h"

/*
 *	CLASS ht_data
 */

class ht_data {
public:
	ht_data();
	virtual ~ht_data();
};

/*
 *	CLASS ht_data_dword
 */

class ht_data_dword: public ht_data {
public:
	dword data;
	ht_data_dword(dword d);
	virtual ~ht_data_dword();
};

/*
 *	CLASS ht_data_ptr
 */

class ht_data_ptr: public ht_data {
public:
	void *data;
	ht_data_ptr(void *data);
	virtual ~ht_data_ptr();
};

/*
 *	CLASS ht_tree
 */

struct ht_node {
	int key;
	ht_data *data;
	ht_node *left, *right;
};

class ht_tree {
private:
	int it_key;
	int count;
	int cached_key;
	ht_data *cached_data;
	ht_node *root;
	void set_r(ht_node **node, int key, ht_data *data);
	ht_node *get_r(ht_node *node, int key);
	ht_node *get_next_r(ht_node *node, int key);
	ht_node *get_smallest(ht_node *node);
	ht_node *get_highest(ht_node *node);
	void free_r(ht_node *node);
	ht_node *balance_r(ht_node **table, int l, int r);
	void free_skeleton_r(ht_node *n);
	ht_node *find_top_node(ht_node *n);
public:
	ht_tree();
	~ht_tree();
/* new */
	ht_node *enum_first();
	ht_node *enum_next();
	void set(int key, ht_data *data);
	ht_data *get(int key);
	int del(int key);
	int getcount();

	void balance();
};

/*
 *   ht_collection
 */

class ht_collection {
private:
	ht_data **table;
	int table_size, table_maxsize;

	void expand();
public:
	ht_collection();
	~ht_collection();
	void insert(ht_data *data);
	ht_data *get(int index);
	void set(int index, ht_data *data);
	void del(int index);
	int count();
};

/*
 *   ht_string_collection
 */

class ht_string_collection {
private:
	char **str_table;
	int str_table_size, str_table_maxsize;

	void expand();
public:
	ht_string_collection();
	~ht_string_collection();
	void insert(char *str);
	char *get(int index);
	void del(int index);
	int count();
};

/*
 *	qsort for byvalue-tables
 */

struct byvalue {
	dword key;
	dword value;
};

void qsortbyvalue(byvalue *table, int l, int r);

/*
 *	qsort for byname-tables
 */

struct byname {
	char *key;
	int keylen;
	dword value;
};

void qsortbyname(byname *table, int l, int r);

/*
 *	simple int hash
 */

struct int_hash {
	int value;
	char *desc;
};

char *matchhash(int value, int_hash *hash_table);

#endif /* ! __HTDATA_H__ */

