/*
 *	HT Editor
 *	htdialog.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "htapp.h"
#include "htclipboard.h"
#include "htcolor.h"
#include "htctrl.h"
#include "htdialog.h"
#include "htidle.h"
#include "htkeyb.h"
#include "htstring.h"
#include "tools.h"

int iexecute_dialog(ht_dialog *dialog, int modal)
{
	ht_view *orig_focused=app->getselected(), *orig_baseview=baseview;
	int oldx, oldy;
	ht_view *drawer=modal ? dialog : app;
	screen->getcursor(&oldx, &oldy);
	dialog->setstate(ds_normal, 0);
	((ht_app*)app)->insert(dialog);
	((ht_app*)app)->focus(dialog);
	baseview=dialog;
	drawer->sendmsg(htm_draw, 0);
	screen->show();
	while (dialog->getstate(0)==ds_normal) {
		if (htkeypressed()) {
			dialog->sendmsg(htm_keypressed, htgetkey());
			drawer->sendmsg(htm_draw, 0);
			screen->show();
		}
		if (!modal) do_idle();
	}
	int return_val;
	int state=dialog->getstate(&return_val);
	screen->setcursor(oldx, oldy);
	((ht_app*)app)->remove(dialog);
	app->focus(orig_focused);
	baseview=orig_baseview;
	if (state!=ds_term_cancel) {
		return return_val;
	} else {
		return 0;
	}
}

int execute_dialog(ht_dialog *dialog)
{
    return iexecute_dialog(dialog, 0);
}

int execute_dialog_modal(ht_dialog *dialog)
{
    return iexecute_dialog(dialog, 1);
}

/*
 *	CLASS ht_dialog
 */

void ht_dialog::init(bounds *b, char *desc, int framestyle)
{
	ht_window::init(b, desc, -1, framestyle);
	setpalette(dialog_palette_str);
	VIEW_DEBUG_NAME("ht_dialog");
	options&=~VO_SELBOUND;
}

void ht_dialog::done()
{
	ht_window::done();
}

int ht_dialog::alone()
{
	return 1;
}

void ht_dialog::draw()
{
	clear(windialog_color_body);
	ht_group::draw();
}

int ht_dialog::getstate(int *_return_val)
{
	if (_return_val) *_return_val=return_val;
	return state;
}

void ht_dialog::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_button_pressed) {
		switch (msg->data_int) {
			case button_cancel:
				setstate(ds_term_cancel, msg->data_int);
				clearmsg(msg);
				return;
			default:
				setstate(ds_term_ok, msg->data_int);
				clearmsg(msg);
				return;
		}
	}
	ht_window::handlemsg(msg);
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Escape:
				setstate(ds_term_cancel, msg->data_int);
				clearmsg(msg);
				return;
			case K_Return:
				sendmsg(htm_button_pressed, button_ok);
				clearmsg(msg);
				return;
		}
	}
}

void ht_dialog::setstate(int _state, int _return_val)
{
	state=_state;
	return_val=_return_val;
}

/*
 *	CLASS ht_cluster
 */

void ht_cluster::init(bounds *b, ht_string_collection *_strings)
{
	ht_view::init(b, VO_SELECTABLE | VO_OWNBUFFER | VO_POSTPROCESS, 0);
	VIEW_DEBUG_NAME("ht_cluster");
	strings=_strings;
	scount=strings->count();
	if (scount>32) scount=32;			/* cant use more than 32... */
	sel=0;
	for (int i=0; i<scount; i++) {
		char *s=strings->get(i);
		s=strchr(s, '~');
		if (s) {
			shortcuts[i]=htkey2alt(*(s+1));
		} else shortcuts[i]=-1;
	}
}

void ht_cluster::done()
{
	ht_view::done();
}

void ht_cluster::defaultpalette()
{
	pal_name=dialog_palette_str;
}

/*
 *	CLASS ht_checkboxes
 */

void ht_checkboxes::init(bounds *b, ht_string_collection *strings)
{
	ht_cluster::init(b, strings);
	VIEW_DEBUG_NAME("ht_checkboxes");
	state=0;
}

void ht_checkboxes::done()
{
	ht_cluster::done();
}

int ht_checkboxes::datasize()
{
	return sizeof (ht_checkboxes_data);
}

void ht_checkboxes::draw()
{
	clear(windialog_color_cluster_unfocused);
	int i=0;
	int vx=0, vy=0;
	int maxcolstrlen=0;
	while (i<scount) {
		int c=windialog_color_cluster_unfocused;
		if ((focused) && (sel==i)) c=windialog_color_cluster_focused;
		char *s=strings->get(i);
		int slen=strlen(s);
		if (slen>maxcolstrlen) maxcolstrlen=slen;
		if ((1<<i) & state) {
			print(vx, vy, c, "[X]");
		} else {
			print(vx, vy, c, "[ ]");
		}
		int k=0, oc=c;
		for (int q=0; q<size.w-4; q++) {
			if (!*(s+q)) break;
			if (*(s+q)=='~') {
				c=windialog_color_cluster_shortcut;
				continue;
			} else {
				printchar(vx+k+4, vy, c, *(s+q));
				k++;
			}
			c=oc;
		}
		i++;
		vy++;
		if (vy>=size.h) {
			vx+=maxcolstrlen+5;
			vy=0;
			maxcolstrlen=0;
		}
	}
}

void ht_checkboxes::getdata(void *buf)
{
	((ht_checkboxes_data*)buf)->state=state;
}

void ht_checkboxes::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
			for (int i=0; i<scount; i++) {
				if ((shortcuts[i]!=-1) && (msg->data_int==shortcuts[i])) {
					sel=i;
					state=state^(1<<sel);
					app->focus(this);
					dirtyview();
					clearmsg(msg);
					return;
				}
			}
		}
	} else {
		if (msg->msg==htm_keypressed) {
			switch (msg->data_int) {
				case K_Left:
					sel-=size.h;
					if (sel<0) sel=0;
					dirtyview();
					clearmsg(msg);
					return;
				case K_Right:
					sel+=size.h;
					if (sel>=scount) sel=scount-1;
					dirtyview();
					clearmsg(msg);
					return;
				case K_Up:
					sel--;
					if (sel<0) sel=0;
					dirtyview();
					clearmsg(msg);
					return;
				case K_Down:
					sel++;
					if (sel>=scount) sel=scount-1;
					dirtyview();
					clearmsg(msg);
					return;
				case K_Space:
					state=state^(1<<sel);
					dirtyview();
					clearmsg(msg);
					return;
			}
		}
	}
	ht_cluster::handlemsg(msg);
}

void ht_checkboxes::setdata(void *buf)
{
	state=((ht_checkboxes_data*)buf)->state;
}

/*
 *	CLASS ht_radioboxes
 */

void ht_radioboxes::init(bounds *b, ht_string_collection *strings)
{
	ht_cluster::init(b, strings);
	VIEW_DEBUG_NAME("ht_radioboxes");
}

void ht_radioboxes::done()
{
	ht_cluster::done();
}

int ht_radioboxes::datasize()
{
	return sizeof (ht_radioboxes_data);
}

void ht_radioboxes::draw()
{
	clear(windialog_color_cluster_unfocused);
	int i=0;
	int vx=0, vy=0;
	int maxcolstrlen=0;
	while (i<scount) {
		int c=windialog_color_cluster_unfocused;
		if ((focused) && (sel==i)) c=windialog_color_cluster_focused;
		char *s=strings->get(i);
		int slen=strlen(s);
		if (slen>maxcolstrlen) maxcolstrlen=slen;
		print(vx, vy, c, "( )");
		if (i==sel) {
			printchar(vx+1, vy, c, CHAR_RADIO);
		}
		print(vx+4, vy, c, s);
		i++;
		vy++;
		if (vy>=size.h) {
			vx+=maxcolstrlen+5;
			vy=0;
			maxcolstrlen=0;
		}
	}
}

void ht_radioboxes::getdata(void *buf)
{
	((ht_radioboxes_data*)buf)->sel=sel;
}

void ht_radioboxes::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
				sel-=size.h;
				if (sel<0) sel=0;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right:
				sel+=size.h;
				if (sel>=scount) sel=scount-1;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Up:
				sel--;
				if (sel<0) sel=0;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
				sel++;
				if (sel>=scount) sel=scount-1;
				dirtyview();
				clearmsg(msg);
				return;
/*			case K_Space:
				state=state^(1<<sel);
				dirtyview();
				clearmsg(msg);
				break;*/
		}
	}
	ht_cluster::handlemsg(msg);
}

/*
 *	CLASS ht_inputfield
 */

void ht_inputfield::init(bounds *b, int _maxtextlen, ht_collection *_history)
{
	ht_view::init(b, VO_SELECTABLE, "some inputfield");
	VIEW_DEBUG_NAME("ht_inputfield");
	textv=(byte*)malloc(_maxtextlen);
	curcharv=textv;
	textlenv=0;
	maxtextlenv=_maxtextlen;
	selstartv=0;
	selendv=0;

	text=&textv;
	curchar=&curcharv;
	textlen=&textlenv;
	maxtextlen=&maxtextlenv;
	selstart=&selstartv;
	selend=&selendv;

	insert=1;
	ofs=0;
	attachedto=0;
	history=_history;
}

void ht_inputfield::done()
{
	freebuf();
	ht_view::done();
}

void ht_inputfield::attach(ht_inputfield *inputfield)
{
	freebuf();
	inputfield->query(&curchar, &text, &selstart, &selend, &textlen, &maxtextlen);
	attachedto=inputfield;
	ofs=0;
	insert=1;
}

int ht_inputfield::datasize()
{
	return sizeof (ht_inputfield_data);
}

void ht_inputfield::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_inputfield::freebuf()
{
	if ((text) && (*text) && (!attachedto)) delete *text;
}

void ht_inputfield::getdata(void *buf)
{
	if (attachedto) {
		((ht_inputfield_data*)buf)->isattached=1;
		((ht_inputfield_data*)buf)->attached.attachedto=attachedto;
	} else {
		((ht_inputfield_data*)buf)->isattached=0;
		((ht_inputfield_data*)buf)->pure.textlen=*textlen;
		((ht_inputfield_data*)buf)->pure.maxtextlen=*maxtextlen;
		((ht_inputfield_data*)buf)->pure.text=*text;
	}
}

int ht_inputfield::insertbyte(byte *pos, byte b)
{
	if (*textlen<*maxtextlen) {
		*selstart=0;
		*selend=0;
		memmove(pos+1, pos, *textlen-(pos-*text));
		*pos=b;
		(*textlen)++;
		dirtyview();
		return 1;
	}
	return 0;
}

void ht_inputfield::query(byte ***c, byte ***t, byte ***ss, byte ***se, int **tl, int **mtl)
{
	*c=curchar;
	*t=text;
	*ss=selstart;
	*se=selend;
	*tl=textlen;
	*mtl=maxtextlen;
}

void ht_inputfield::select_add(byte *start, byte *end)
{
	if (end<start) {
		byte *temp=start;
		start=end;
		end=temp;
	}
	if (end==*selstart) {
		*selstart=start;
	} else if (start==*selend) {
		*selend=end;
	} else if (start==*selstart) {
		*selstart=end;
	} else if (end==*selend) {
		*selend=start;
	} else {
		*selstart=start;
		*selend=end;
	}
	if (*selstart>*selend) {
		byte *temp=*selstart;
		*selstart=*selend;
		*selend=temp;
	}
}

void ht_inputfield::setdata(void *buf)
{
	if (((ht_inputfield_data*)buf)->isattached) {
		attach(((ht_inputfield_data*)buf)->attached.attachedto);
	} else {
		textlenv=((ht_inputfield_data*)buf)->pure.textlen;
		textlen=&textlenv;
		maxtextlenv=((ht_inputfield_data*)buf)->pure.maxtextlen;
		maxtextlen=&maxtextlenv;
		if (*textlen>*maxtextlen) *textlen=*maxtextlen;

		freebuf();
		textv=(byte*)malloc(*maxtextlen+1);
		memmove(textv, ((ht_inputfield_data*)buf)->pure.text, *maxtextlen+1);
		textv[*maxtextlen]=0;
		text=&textv;
		curcharv=*text+*textlen;
		curchar=&curcharv;
		attachedto=0;
		*selstart=*text;
		*selend=*text+*textlen;
	}
}

/*
 *	CLASS ht_strinputfield
 */

void ht_strinputfield::init(bounds *b, int maxstrlen, ht_collection *history)
{
	ht_inputfield::init(b, maxstrlen, history);
	VIEW_DEBUG_NAME("ht_strinputfield");
	is_virgin=1;
}

void ht_strinputfield::done()
{
	ht_inputfield::done();
}

void ht_strinputfield::correct_viewpoint()
{
	if (*curchar-*text<ofs) ofs=*curchar-*text; else
	if (*curchar-*text-(size.w-2)*size.h>ofs) ofs=*curchar-*text-(size.w-2)*size.h;
}

void ht_strinputfield::draw()
{
	int c=focused ? windialog_color_input_focused : windialog_color_input_unfocused;
	byte *t=*text+ofs;
	int l=*textlen-ofs;
	if (l>size.w) l=size.w;
	int y=0;
	fill(0, 0, size.w, size.h, c, ' ');
	if (ofs) printchar(0, y, windialog_color_input_clip, '<');
	for (int k=0; k<*textlen-ofs; k++) {
		if (1+k-y*(size.w-2)>size.w-2) {
			if (y+1<size.h) y++; else break;
		}
		if ((t<*selstart) || (t>=*selend)) {
			printchar(1+k-y*(size.w-2), y, c, *(t++));
		} else {
			printchar(1+k-y*(size.w-2), y, windialog_color_input_selected, *(t++));
		}
	}
	if (*textlen-ofs > (size.w-2)*size.h) {
		printchar(size.w-1, y, windialog_color_input_clip, '>');
	}
	if (history && history->count()) {
		printchar(size.w-1, y, windialog_color_input_clip, CHAR_ARROW_DOWN);
	}
	if (focused) {
		if (*curchar-*text-ofs>=(size.w-2)*size.h) setcursor(size.w-1, size.h-1); else setcursor((*curchar-*text-ofs)%(size.w-2)+1, (*curchar-*text-ofs)/(size.w-2));
	}
}

void ht_strinputfield::handlemsg(htmsg *msg)
{
	if ((msg->type==htmt_nothing) && (msg->msg==htm_keypressed)) {
		switch (msg->data_int) {
			case K_Up:
				if (*curchar-*text-ofs<size.w-2) {
					ofs-=size.w-2;
					if (ofs<0) ofs=0;
				}
				*curchar-=size.w-2;
				if (*curchar<*text) *curchar=*text;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
				*curchar+=size.w-2;
				if (*curchar-*text>*textlen) *curchar=*text+*textlen;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Left:
				if (*curchar>*text) (*curchar)--;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right:
				if (*curchar-*text<*textlen) (*curchar)++;
				is_virgin=0;
				dirtyview();
				clearmsg(msg);
				return;
			case K_BackSpace:
				if (*curchar>*text) {
					*selstart=0;
					*selend=0;
					memmove(*curchar-1, *curchar, *textlen-(*curchar-*text));
					(*textlen)--;
					(*curchar)--;
					is_virgin=0;
					correct_viewpoint();
					dirtyview();
					clearmsg(msg);
					return;
				}
				break;
			case K_Delete:
				if (is_virgin) {
					*selstart=0;
					*selend=0;
					*textlen=0;
					*curchar=*text;
					is_virgin=0;
					dirtyview();
					clearmsg(msg);
				} else if ((*curchar-*text<*textlen) && (*textlen)) {
					*selstart=0;
					*selend=0;
					memmove(*curchar, *curchar+1, *textlen-(*curchar-*text)-1);
					(*textlen)--;
					is_virgin=0;
					dirtyview();
					clearmsg(msg);
					return;
				}
				break;
			case K_Home:
				*curchar=*text;
				is_virgin=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_End:
				*curchar=*text+*textlen;
				is_virgin=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Insert:
				insert=!insert;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Shift_Left:
				if (*curchar>*text) {
					(*curchar)--;
					select_add(*curchar, *curchar+1);
				}
				is_virgin=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Shift_Right:
				if (*curchar-*text<*textlen) {
					(*curchar)++;
					select_add(*curchar-1, *curchar);
				}
				is_virgin=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Shift_Home:
				select_add(*curchar, *text);
				*curchar=*text;
				is_virgin=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Shift_End:
				select_add(*curchar, *text+*textlen);
				*curchar=*text+*textlen;
				is_virgin=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Shift_Delete:
				if (*selend>*selstart) {
					memmove(*selstart, *selend, *textlen-(*selend-*text));
					*textlen-=*selend-*selstart;
					*curchar=*selstart;
					*selstart=0;
					*selend=0;
					is_virgin=0;
					correct_viewpoint();
				}
				dirtyview();
				clearmsg(msg);
				return;
			case K_Control_Insert:
				if (*selend>*selstart) clipboard_copy("inputfield", *selstart, *selend-*selstart);
				is_virgin=0;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Shift_Insert: {
				int maxsize=MIN(*maxtextlen-*textlen, (int)clipboard_getsize());
				byte *buf=(byte*)malloc(maxsize);
				int r=clipboard_paste(buf, maxsize);
				if (r) {
					for (int i=0; i<r; i++) {
						setbyte(buf[i]);
					}
					*selstart=*curchar-r;
					*selend=*curchar;
				}
				delete buf;
				is_virgin=0;
				dirtyview();
				clearmsg(msg);
				return;
			}
			default:
				if ((msg->data_int>=' ') && (msg->data_int<256)) {
					setbyte(msg->data_int);
					dirtyview();
					clearmsg(msg);
					return;
				}
		}
	}
	ht_inputfield::handlemsg(msg);
}

void ht_strinputfield::receivefocus()
{
	correct_viewpoint();
	ht_inputfield::receivefocus();
}

void ht_strinputfield::setbyte(byte a)
{
	*selstart=0;
	*selend=0;
	if ((insert) || (*curchar-*text>=*textlen)) {
		if ((insertbyte(*curchar, a)) && (*curchar-*text<*textlen)) (*curchar)++;
	} else {
		**curchar=a;
		(*curchar)++;
	}
	correct_viewpoint();
	is_virgin=0;
}

/*
 *	CLASS ht_hexinputfield
 */

void ht_hexinputfield::init(bounds *b, int maxstrlen)
{
	ht_inputfield::init(b, maxstrlen);
	VIEW_DEBUG_NAME("ht_strinputfield");
	nib=0;
	insert=1;
}

void ht_hexinputfield::done()
{
	ht_inputfield::done();
}

void ht_hexinputfield::correct_viewpoint()
{
	if (*curchar-*text<ofs) ofs=*curchar-*text; else
	if ((*curchar-*text)*3-(size.w-2)*size.h+5>ofs*3) ofs=((*curchar-*text)*3-(size.w-2)*size.h+5)/3;
}

void ht_hexinputfield::draw()
{
	int c=focused ? windialog_color_input_focused : windialog_color_input_unfocused;
	char hbuf[256], *h=hbuf;
	int y=0;
	fill(0, 0, size.w, size.h, c, ' ');
	if (ofs) print(0, y, windialog_color_input_clip, "<");
	int vv=*textlen-ofs;
	if (vv<0) vv=0; else if (vv>(size.w-2)*size.h/3) vv=(size.w-2)*size.h/3+1;
	for (int k=0; k<vv; k++) {
		h+=sprintf(h, "%02x ", *(*text+k+ofs));
	}
	if (vv) {
		h=hbuf;
		while ((*h) && (y<size.h)) {
			h+=lprint(1, y, c, size.w-2, h);
			y++;
		}
		y--;
	}
	if ((*textlen-ofs)*3 > (size.w-2)*size.h) {
		print(size.w-1, y, windialog_color_input_clip, ">");
	}
	if (focused) {
		if ((*curchar-*text-ofs)*3+nib+1>=(size.w-2)*size.h) setcursor(size.w-1, size.h-1); else
		setcursor(((*curchar-*text-ofs)*3+nib+1)%(size.w-2), ((*curchar-*text-ofs)*3+nib+1)/(size.w-2));
	}
}

void ht_hexinputfield::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Up:
				if (*curchar-*text-ofs<(size.w-2)/3) {
					ofs-=(size.w-2)/3;
					if (ofs<0) ofs=0;
				}
				*curchar-=(size.w-2)/3;
				if (*curchar<*text) *curchar=*text;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
/*				if (*curchar-*text-ofs>=(size.w-2)*(size.h-1)/3) {
					ofs+=(size.w-2)/3;
				}*/
				*curchar+=(size.w-2)/3;
				if (*curchar-*text>*textlen) *curchar=*text+*textlen;
				if (*curchar-*text>=*textlen) nib=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Left:
				if (nib) {
					nib=0;
				} else if (*curchar>*text) {
					(*curchar)--;
					nib=1;
				}
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right:
				if (!nib) {
					if (*curchar-*text<*textlen) nib=1;
				} else if (*curchar-*text<*textlen) {
					(*curchar)++;
					nib=0;
				}
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_BackSpace:
				if (*textlen) {
					if ((nib) || (*curchar==*text)) {
						if (*curchar-*text<*textlen) {
							memmove(*curchar, *curchar+1, *textlen-(*curchar-*text)-1);
						}
					} else {
						memmove(*curchar-1, *curchar, *textlen-(*curchar-*text));
					}
					(*textlen)--;
					if (*curchar-*text) (*curchar)--;
					nib=0;
					if (*curchar-*text<*textlen) nib=1;
					correct_viewpoint();
					dirtyview();
					clearmsg(msg);
					return;
				}
				break;
			case K_Delete:
				if ((*curchar-*text<*textlen) && (*textlen)) {
					memmove(*curchar, *curchar+1, *textlen-(*curchar-*text)-1);
					(*textlen)--;
					nib=0;
					dirtyview();
					clearmsg(msg);
					return;
				}
				break;
			case K_Home:
				*curchar=*text;
				nib=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_End:
				*curchar=*text+*textlen;
				nib=0;
				correct_viewpoint();
				dirtyview();
				clearmsg(msg);
				return;
			case K_Insert:
				insert=!insert;
				dirtyview();
				clearmsg(msg);
				return;
			default:
				if ((msg->data_int>='0') && (msg->data_int<='9')) {
					setnibble(msg->data_int-'0');
					dirtyview();
					clearmsg(msg);
				} else if ((msg->data_int>='a') && (msg->data_int<='f')) {
					setnibble(msg->data_int-'a'+10);
					dirtyview();
					clearmsg(msg);
				}
				return;
		}
	}
	ht_inputfield::handlemsg(msg);
}

void ht_hexinputfield::receivefocus()
{
	correct_viewpoint();
	if ((nib) && (*curchar-*text==*textlen)) {
		nib=0;
	}
	ht_inputfield::receivefocus();
}

void ht_hexinputfield::setnibble(byte a)
{
	if (((insert) || (*curchar-*text>=*textlen)) && (nib==0)) {
		if ((insertbyte(*curchar, a<<4)) && (*curchar-*text<*textlen)) nib=1;
	} else {
		if (nib) {
			**curchar=(**curchar & 0xf0) | a;
			(*curchar)++;
			nib=0;
		} else {
			**curchar=(**curchar & 0xf) | (a<<4);
			nib=1;
		}
	}
	correct_viewpoint();
}

/*
 *	CLASS ht_button
 */

void ht_button::init(bounds *b, char *_text, int _value)
{
	ht_view::init(b, VO_SELECTABLE | VO_OWNBUFFER | VO_POSTPROCESS, "some button");
	VIEW_DEBUG_NAME("ht_button");
	value=_value;
	text=strdup(_text);
	pressed=0;
	magicchar=strchr(text, '~');
	if (magicchar) {
		int l=strlen(text);
		memmove(magicchar, magicchar+1, l-(magicchar-text));
		shortcut=htkey2alt(*magicchar);
	} else shortcut=-1;
}

void ht_button::done()
{
	if (text) delete text;
	ht_view::done();
}

int ht_button::datasize()
{
	return sizeof (ht_button_data);
}

void ht_button::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_button::draw()
{
	int c=focused ? windialog_color_button_focused : windialog_color_button_unfocused;
	fill(0, 0, size.w-1, size.h-1, c, ' ');
	int xp=(size.w-strlen(text))/2, yp=(size.h-1)/2;
	print(xp, yp, c, text);
	if (magicchar) printchar(xp+(magicchar-text), yp, windialog_color_button_shortcut, *magicchar);
/* shadow */
	printchar(0, 1, windialog_color_button_shadow, ' ');
	for (int i=1; i<size.w-1; i++) {
		printchar(i, 1, windialog_color_button_shadow, CHAR_FILLED_HU);
	}
	printchar(size.w-1, 0, windialog_color_button_shadow, CHAR_FILLED_HL);
	printchar(size.w-1, 1, windialog_color_button_shadow, CHAR_FILLED_HU);
}

void ht_button::getdata(void *buf)
{
	((ht_button_data*)buf)->pressed=pressed;
}

void ht_button::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
			if (msg->data_int==shortcut) {
				push();
				dirtyview();
				clearmsg(msg);
				return;
			}
		}
	} else if (msg->type==htmt_nothing) {
		if (msg->msg==htm_keypressed) {
			switch (msg->data_int) {
				case K_Return:
				case K_Space:
					push();
					dirtyview();
					clearmsg(msg);
					return;
			}
		}
	}
	ht_view::handlemsg(msg);
}

void ht_button::push()
{
/* FIXME: wont work for encapsulated buttons... */
/* FIXME: (thats why I hacked this now...) */
	app->sendmsg(htm_button_pressed, value);
	pressed=1;
}

/*
 *	ht_listbox
 */

void ht_listbox::init(bounds *b)
{
	ht_view::init(b, VO_SELECTABLE | VO_OWNBUFFER, 0);

	bounds c=*b;
	c.x=c.w-1;
	c.y=0;
	c.w=1;
	scrollbar=new ht_scrollbar();
	scrollbar->init(&c, &windialog_color_scrollbar);

	pos = 0;
	cursor = 0;
	e_top = getfirst();
	e_cursor = e_top;
	visible_height = 0;
	x = 0;
	clear_quickfind();
	update();
}

void	ht_listbox::done()
{
	ht_view::done();
}

void ht_listbox::adjust_scrollbar()
{
	if (count > size.h) {
		scrollbar->enable();
		if (count > 1) scrollbar->setpos(pos * 100 / (count-1));
	} else {
		scrollbar->disable();
	}
}

int  ht_listbox::calc_count()
{
	// abstract
	return 0;
}

void ht_listbox::clear_quickfind()
{
	quickfinder[0] = 0;
	qpos = quickfinder;
	update_cursor();
}

int  ht_listbox::cursor_adjust()
{
	return 0;
}

int  ht_listbox::cursor_up(int n)
{
	void *tmp;
	int  i = 0;

	while (n--) {
		tmp = getprev(e_cursor);
		if (!tmp) break;
		if (e_cursor==e_top) {
			e_top = tmp;
		} else {
			cursor--;
		}
		e_cursor = tmp;
		pos--;
		i++;
	}
	return i;
}

int  ht_listbox::cursor_down(int n)
{
	void *tmp;
	int  i = 0;

	while (n--) {
		tmp = getnext(e_cursor);
		if (!tmp) break;
		if (cursor+1>=visible_height) {
			e_top = getnext(e_top);
		} else {
			cursor++;
		}
		pos++;
		e_cursor = tmp;
		i++;
	}
	return i;
}

int  ht_listbox::datasize()
{
	return 8;
}

void ht_listbox::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_listbox::draw()
{
	int fc = focused ? windialog_color_list_focused_unselected : windialog_color_list_unfocused_unselected;

	clear(fc);

	void *entry = e_top;

	int i=0;
	int cols = num_cols();
	while ((entry) && (i < visible_height)) {
		int c=(i==cursor) ? (focused ? windialog_color_list_focused_selected : windialog_color_list_unfocused_selected) : fc;
		if (i==cursor) {
			fill(0, i, size.w, 1, c, ' ');
		}
		int X = -x;
		for (int j=0; j<cols; j++) {
			char *s = getstr(j, entry);
			int slen = strlen(s);
			if (s) {
				if (X > 0) {
					lprint(X, i, c, size.w, s);
				} else {
					if (slen > -X) lprint(0, i, c, size.w, &s[-X]);
				}
			}
			X += slen;
			if (j+1<cols) {
				printchar(X++, i, c, ' ');
				printchar(X++, i, c, CHAR_LINEV);
				printchar(X++, i, c, ' ');
			}
		}
		entry = getnext(entry);
		i++;
	}
}

void ht_listbox::getdata(void *buf)
{
	((ht_listbox_data*)buf)->top = e_top;
	((ht_listbox_data*)buf)->cursor = e_cursor;
}

void *ht_listbox::getfirst()
{
	// abstract
	return NULL;
}

void *ht_listbox::getlast()
{
	// abstract
	return NULL;
}

void *ht_listbox::getnext(void *entry)
{
	// abstract
	return NULL;
}

void *ht_listbox::getnth(int n)
{
	// abstract
	return NULL;
}

void *ht_listbox::getprev(void *entry)
{
	// abstract
	return NULL;
}

char *ht_listbox::getstr(int col, void *entry)
{
	// abstract
	return NULL;
}

void ht_listbox::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Control_PageUp:
			case K_Home:
				clear_quickfind();
				pos = cursor = 0;
				e_top = e_cursor = getfirst();
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Control_PageDown:
			case K_End: {
				clear_quickfind();
				cursor = 0;
				pos = count;
				e_cursor = e_top = getlast();
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			}
			case K_PageUp:
				clear_quickfind();
				cursor_up(visible_height-1);
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_PageDown: {
				clear_quickfind();
				cursor_down(visible_height-1);
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			}
			case K_Up:
				clear_quickfind();
				cursor_up(1);
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Down:
				clear_quickfind();
				cursor_down(1);
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Left:
				if (x > 0) x--;
				update_cursor();
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Right:
				x++;
				update_cursor();
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_BackSpace: {
				if (qpos > quickfinder) {
					*(--qpos) = 0;
					void *a = quickfind(quickfinder);
					if (a) {
						e_cursor = a;
						e_top = a;
						cursor = 0;
						update_cursor();
						dirtyview();
						clearmsg(msg);
					}
				}
				return;
			}
			default: {
				if (msg->data_int < 0x7f) {
					*(qpos++) = msg->data_int;
					*qpos = 0;
					void *a = quickfind(quickfinder);
					if (a) {
						e_cursor = a;
						e_top = a;
						pos = cursor = 0;
						update_cursor();
						dirtyview();
						clearmsg(msg);
					} else {
						*(--qpos) = 0;
					}
				}
			}
		}
	}
	ht_view::handlemsg(msg);
}

int	ht_listbox::num_cols()
{
	return 1;
}

void *ht_listbox::quickfind(char *s)
{
	// abstract
	return NULL;
}

void ht_listbox::redraw()
{
	ht_view::redraw();
	scrollbar->relocate_to(this);
	scrollbar->redraw();
	scrollbar->unrelocate_to(this);
}

bool ht_listbox::seek(int index)
{
//	if (index < count) {
		cursor = index;
//		pos = 0;
		if (cursor > visible_height-1) cursor = visible_height-1;
		void *e = getfirst();
		for (int i=0; i<cursor; i++) {
			e = getnext(e);
		}
		e_cursor = e;
		adjust_scrollbar();
		return true;
//	}
	return false;
}

void ht_listbox::select_entry(void *entry)
{
	e_cursor = e_top = entry;
	cursor = 0;
}

void ht_listbox::setdata(void *buf)
{
	e_top = ((ht_listbox_data*)buf)->top;
	e_cursor = ((ht_listbox_data*)buf)->cursor;
	update();
}

/*
 *	must be called if data changes
 */
void ht_listbox::update()
{
	if (!e_top) {
		e_top = getfirst();
		if (!e_cursor) e_cursor = e_top;
	}
	count = calc_count();
	visible_height=MIN(size.h, count);
	void *entry = e_top;
	cursor = 0;
	while ((entry) && (cursor<visible_height)) {
		if (entry == e_cursor) goto ok;
		entry = getnext(entry);
		cursor++;
	}
	cursor = 0;
	e_top = e_cursor;
	count = calc_count();
	visible_height=MIN(size.h, count);
ok:
	adjust_scrollbar();
	dirtyview();
}

void ht_listbox::update_cursor()
{
	if (quickfinder[0]) {
		setcursor((qpos-quickfinder)+cursor_adjust()-x, cursor);
	} else {
		hidecursor();
	}
}

/*
 *	ht_text_listbox
 */

void	ht_text_listbox::init(bounds *b, int Cols, int Keycol)
{
	cols = Cols;
	keycol = Keycol;
	first = last = NULL;
	count = 0;
	widths = (int *) calloc(sizeof(int)*cols, 1);
	return_str = (char *) malloc(1024);
	Cursor_adjust = 0;
	ht_listbox::init(b);
}

void ht_text_listbox::done()
{
	ht_listbox::done();
	free(widths);
	free(return_str);
}

int	ht_text_listbox::calc_count()
{
	return count;
}

int  ht_text_listbox::cursor_adjust()
{
	return Cursor_adjust;
}

void *ht_text_listbox::getfirst()
{
	return first;
}

void *ht_text_listbox::getlast()
{
	return last;
}

void *ht_text_listbox::getnext(void *entry)
{
	if (!entry) return NULL;
	return ((ht_text_listbox_item *)entry)->next;
}

void *ht_text_listbox::getprev(void *entry)
{
	if (!entry) return NULL;
	return ((ht_text_listbox_item *)entry)->prev;
}

char *ht_text_listbox::getstr(int col, void *entry)
{
	memset(return_str, ' ', widths[col]);
	return_str[widths[col]] = 0;
	if (entry || (col < cols)) {
		memmove(return_str, ((ht_text_listbox_item *)entry)->data[col],
			   strlen(((ht_text_listbox_item *)entry)->data[col]));
	}
	return return_str;
}

void ht_text_listbox::insert_str(int id, char *str, ...)
{
	ht_text_listbox_item *item = (ht_text_listbox_item *)malloc(sizeof(ht_text_listbox_item )+sizeof(char *)*cols);
	item->next = NULL;
	item->prev = last;
	item->id = id;
	char **str2 = &str;
	for (int i=0; i<cols; i++) {
		int slen = strlen(*str2);
		if (slen > widths[i]) {
			widths[i] = slen;
		}

		item->data[i] = strdup(*(str2++));
	}
	if (first) {
		last->next = item;
	} else {
		first = item;
	}
	last = item;
	count++;
}

int	ht_text_listbox::num_cols()
{
	return cols;
}

void *ht_text_listbox::quickfind(char *s)
{
//	fprintf(stdout, "quickfind: %s\n", s);
	ht_text_listbox_item *item = first;
	int slen = strlen(s);
	while (item && (strncmp(item->data[keycol], s, slen)!=0)) {
		item = item->next;
	}
	return item;
}

void ht_text_listbox::update()
{
	ht_listbox::update();
	Cursor_adjust = 0;
	for (int i=0; i<keycol; i++) {
		Cursor_adjust+=widths[i]+3;
	}
}

/*
 *	CLASS ht_listpopup
 */

#define LISTPOPUP_HEIGHT 5

void ht_listpopup::init(bounds *b)
{
	ht_view::init(b, VO_SELECTABLE | VO_OWNBUFFER, "some listbox");
	VIEW_DEBUG_NAME("ht_listbox");

	bounds c=*b;
	c.x=c.w-1;
	c.y=2;
	c.w=1;
	scrollbar=new ht_scrollbar();
	scrollbar->init(&c, &windialog_color_scrollbar);
	scrollbar->disable();

	list=new ht_string_collection();
	pos=0;
	cursor=0;
	vis_h_size=0;
	adjust_scrollbar();

	disable_buffering();
	expanded=0;
}

void ht_listpopup::done()
{
	delete list;
	scrollbar->done();
	delete scrollbar;
	ht_view::done();
}

void ht_listpopup::adjust_scrollbar()
{
	int c=list->count();
	scrollbar->disable();
	if (c>size.h-2) {
		if (expanded) scrollbar->enable();
		if (c>1) scrollbar->setpos((pos+cursor)*100/(c-1));
	}
}

void ht_listpopup::collapse()
{
	if (expanded) {
		bounds b;
		getbounds(&b);
		b.h=1;
		setbounds(&b);
		expanded=0;
		vis_h_size=1;
		if (scrollbar->enabled) {
			scrollbar->disable();
		}
		dirtyview();
	}
}

int ht_listpopup::datasize()
{
	return sizeof (ht_listpopup_data);
}

void ht_listpopup::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_listpopup::draw()
{
	int c=focused ? windialog_color_list_focused_selected : windialog_color_list_unfocused_selected;
	fill(0, 0, size.w, 1, c, ' ');
	lprint(0, 0, c, size.w-1, list->get(pos+cursor));
	int fc=focused ? windialog_color_list_focused_unselected : windialog_color_list_unfocused_unselected;
	printchar(size.w-1, 0, fc, CHAR_ARROW_DOWN);
	if (expanded) {
		fill(0, 1, size.w, 1, fc, CHAR_LINEH);
		fill(0, 2, size.w, vis_h_size, fc, ' ');
		for (int i=0; i<vis_h_size; i++) {
			int c=fc;
			if (i==cursor) {
				c=windialog_color_list_focused_selected;
				fill(0, i+2, size.w, 1, c, ' ');
			}
			lprint(0, i+2, c, size.w, list->get(pos+i));
		}
	}
}


void ht_listpopup::expand()
{
	if (!expanded) {
		bounds b;
		getbounds(&b);
		b.h=MIN(LISTPOPUP_HEIGHT, list->count())+2;
		setbounds(&b);
		expanded=1;
		vis_h_size=MIN(size.h-2, list->count());
		dirtyview();
		if ((!(scrollbar->enabled)) && (list->count()>vis_h_size)) {
			bounds b;
			scrollbar->getbounds(&b);
			b.h=vis_h_size;
			scrollbar->setbounds(&b);
			scrollbar->enable();
		}
	}
}

void ht_listpopup::getdata(void *buf)
{
	((ht_listpopup_data*)buf)->list=list;
	((ht_listpopup_data*)buf)->index=pos+cursor;
}

void ht_listpopup::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Control_PageUp:
			case K_Home:
				expand();
				pos=0;
				cursor=0;
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Control_PageDown:
			case K_End: {
				expand();
				pos=0;
				int c=list->count();
				cursor=c-1;
				if (cursor>=vis_h_size) {
					pos+=cursor-vis_h_size+1;
					cursor=vis_h_size-1;
				}
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			}
			case K_PageUp:
				expand();
				cursor-=vis_h_size;
				if (cursor<0) {
					pos+=cursor;
					cursor=0;
					if (pos<0) pos=0;
				}
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_PageDown: {
				expand();
				int c=list->count();
				cursor+=vis_h_size;
				if (cursor>=vis_h_size) {
					pos+=cursor-vis_h_size+1;
					cursor=vis_h_size-1;
					if (pos+cursor>=c) pos=c-cursor-1;
				}
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			}
			case K_Up:
				expand();
				if (cursor) {
					cursor--;
				} else {
					if (pos) pos--;
				}
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Down:
				expand();
				if (cursor<vis_h_size-1) {
					if (cursor<list->count()-1) cursor++;
				} else {
					if (pos+vis_h_size<list->count()) pos++;
				}
				dirtyview();
				clearmsg(msg);
				adjust_scrollbar();
				return;
			case K_Return:
				if (expanded) {
					collapse();
					dirtyview();
					clearmsg(msg);
					return;
				}
				break;
		}
	}
	ht_view::handlemsg(msg);
}

void ht_listpopup::insertstring(char *str)
{
	list->insert(str);
	int c=list->count();
	vis_h_size=MIN(size.h, c);
}

void ht_listpopup::redraw()
{
	ht_view::redraw();
	scrollbar->relocate_to(this);
	scrollbar->redraw();
	scrollbar->unrelocate_to(this);
}

void ht_listpopup::releasefocus()
{
	collapse();
	ht_view::releasefocus();
}

int ht_listpopup::seek(int index)
{
	int c=list->count();
	if (index<c) {
		int h=MIN(LISTPOPUP_HEIGHT, c);
		cursor=index;
		pos=0;
		if (cursor>h-1) {
			pos=cursor+1-h;
			cursor=h-1;
		}
		adjust_scrollbar();
		return 1;
	}
	return 0;
}

void ht_listpopup::setdata(void *buf)
{
	seek(((ht_listpopup_data*)buf)->index);
}

/*
 *	CLASS ht_statictext
 */

#define STATICTEXT_MAX_AUTO_LINE_BREAK 15

void ht_statictext_align(ht_statictext_linedesc *d, int align, int *lastalign, int w)
{
again:
	switch (align) {
		case align_center:
			d->ofs=(w-d->len)/2;
			break;
		case align_right:
			d->ofs=w-d->len;
			break;
		case align_custom:
			if (d->text[0]==ALIGN_CHAR_ESCAPE) {
				d->text++;
				d->len--;
				switch (d->text[0]) {
					case ALIGN_CHAR_LEFT:
						d->ofs=0;
						*lastalign=align_left;
						break;
					case ALIGN_CHAR_CENTER:
						d->ofs=(w-(d->len-1))/2;
						*lastalign=align_center;
						break;
					case ALIGN_CHAR_RIGHT:
						d->ofs=w-(d->len-1);
						*lastalign=align_right;
						break;
					default:
						align=*lastalign;
						goto again;
				}
				d->text++;
				d->len--;
			} else {
				align=*lastalign;
				goto again;
			}
			break;
		default:
			d->ofs=0;
			break;
	}
}

void ht_statictext::init(bounds *b, char *_text, int _align)
{
	ht_view::init(b, VO_OWNBUFFER | VO_RESIZE, "some statictext");
	VIEW_DEBUG_NAME("ht_statictext");
	growmode_x=GROW_EDP;
	growmode_y=GROW_EDP;
	align=_align;
	text=strdup(_text);
}

void ht_statictext::done()
{
	if (text) delete text;
	ht_view::done();
}

void ht_statictext::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_statictext::draw()
{
/* format string... */	
	ht_statictext_linedesc *d=(ht_statictext_linedesc *)malloc(sizeof (ht_statictext_linedesc)*size.h), *e=d;
	char *t=gettext(), *n;
	int l=strlen(t);
	int lastalign=align_left;
	int c=0;
	int is_custom_flag;
	while ((*t) && (c<size.h) && (l)) {
		char *lastbreakable=0;
		char *lastgoodbreakable=0;
		int get_lastgoodbreakable=1;
		if ((*t==ALIGN_CHAR_ESCAPE) && (align==align_custom)) {
			t+=2;
			l-=2;
			is_custom_flag=1;
		} else is_custom_flag=0;
		d->len=l;
		d->text=t;
		int i=0;
		while (1) {
			if (i>=size.w+1) {
				if ((!lastbreakable) || (i-(lastbreakable-t)>STATICTEXT_MAX_AUTO_LINE_BREAK)) {
/* force line break... */
					d->len=MIN(lastgoodbreakable-t, size.w);
					n=d->text+MIN(lastbreakable-t, d->len);
				} else {
/* automatic separator line break */
					d->len=lastgoodbreakable-t;
					n=d->text+(lastbreakable-t);
				}
				break;
			}
			if ((i>=l) || (!t[i])) {
				d->len=i;
				n=d->text+d->len;
				break;
			}
			char *isbreakable=strchr(" ,.?!-=", *(t+i));
			if (isbreakable) {
				lastbreakable=t+i+1;
				if (get_lastgoodbreakable) {
					lastgoodbreakable=lastbreakable-1;
					get_lastgoodbreakable=0;
				}
			} else if (*(t+i)=='\n') {
				d->len=i;
				n=d->text+d->len+1;
				break;
			} else {
				get_lastgoodbreakable=1;
			}
			i++;
		}
		if (is_custom_flag) {
			d->len+=2;
			d->text-=2;
		}
		ht_statictext_align(d, align, &lastalign, size.w);
		d++;
		c++;
		t=n;
		while (*t && (*t!='\e') && (*t!='\n') && (*t<=32)) t++;
		l=strlen(t);
	}
/**/
	clear(windialog_color_body);
	d=e;
	for (int i=0; i<c; i++) {
		lprint(d->ofs, i, windialog_color_body, d->len, d->text);
		d++;
	}
	free(e);
}

char *ht_statictext::gettext()
{
	return text;
}

void ht_statictext::settext(char *_text)
{
	if (text) delete text;
	text=strdup(_text);
}

/*
 *	CLASS ht_listbox_ptr
 */

ht_listbox_ptr::ht_listbox_ptr(ht_listbox *_listbox)
{
	listbox=_listbox;
}

ht_listbox_ptr::~ht_listbox_ptr()
{
}

/*
 *	CLASS ht_label
 */

void ht_label::init(bounds *b, char *_text, ht_view *_connected)
{
	ht_view::init(b, VO_POSTPROCESS, 0);
	text=strdup(_text);
	magicchar=strchr(text, '~');
	if (magicchar) {
		int l=strlen(text);
		memmove(magicchar, magicchar+1, l-(magicchar-text));
	}
	connected=_connected;
	if (magicchar) {
		shortcut=htkey2alt(*magicchar);
	} else shortcut=-1;
}

void ht_label::done()
{
	if (text) delete text;
	ht_view::done();
}

void ht_label::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_label::draw()
{
	lprint(0, 0, windialog_color_body, size.w, text);
	if (magicchar) printchar(magicchar-text, 0, windialog_color_shortcut, *magicchar);
}

void ht_label::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
			if ((shortcut!=-1) && (msg->data_int==shortcut)) {
				app->focus(connected);
				dirtyview();
				clearmsg(msg);
				return;
			}
		}
	} else ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_progress_indicator_body
 */

void ht_progress_indicator_body::init(bounds *b, char *_format)
{
	ht_view::init(b, VO_OWNBUFFER | VO_SELECTABLE, 0);
	VIEW_DEBUG_NAME("ht_progress_indicator_body");
	indicator=0;
	format=_format;
}

void ht_progress_indicator_body::done()
{
	ht_view::done();
}

void ht_progress_indicator_body::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_progress_indicator_body::draw()
{
	char buf[1024];		/* possible buffer overflow */
	clear(windialog_color_body);
	sprintf(buf, format, indicator);
	print((size.w-strlen(buf))/2, size.h/2-1, windialog_color_body, buf);
	fill(size.w/4, size.h/2, size.w/2, 1, windialog_color_body, CHAR_FILLED_M);
	fill(size.w/4, size.h/2, size.w/2*indicator/100, 1, windialog_color_body, CHAR_FILLED_F);
}

int ht_progress_indicator_body::datasize()
{
	return sizeof (ht_progress_indicator_data);
}

void ht_progress_indicator_body::setdata(void *buf)
{
	indicator=((ht_progress_indicator_data*)buf)->indicator;
	dirtyview();
}

/*
 *	CLASS ht_progress_indicator
 */

void ht_progress_indicator::init(bounds *b, char *format)
{
	ht_window::init(b, "Progress", 0, FS_TITLE);
	VIEW_DEBUG_NAME("ht_progress_indicator");
	bounds c;
	frame->getclientarea(&c);
	c.x=0;
	c.y=0;
	body=new ht_progress_indicator_body();
	body->init(&c, format);
	insert(body);
	setpalette(dialog_palette_str);
}

void ht_progress_indicator::done()
{
	ht_window::done();
}

int ht_progress_indicator::datasize()
{
	return body->datasize();
}

void ht_progress_indicator::setdata(void *buf)
{
	body->setdata(buf);
}

/*
 *	CLASS ht_color_block
 */

int vcs[16]={VC_BLACK, VC_BLUE, VC_GREEN, VC_CYAN, VC_RED, VC_MAGENTA, VC_YELLOW, VC_WHITE,   VC_LIGHT(VC_BLACK), VC_LIGHT(VC_BLUE), VC_LIGHT(VC_GREEN), VC_LIGHT(VC_CYAN), VC_LIGHT(VC_RED), VC_LIGHT(VC_MAGENTA), VC_LIGHT(VC_YELLOW), VC_LIGHT(VC_WHITE)};

void ht_color_block::init(bounds *b, int selected, int _flags)
{
	ht_view::init(b, VO_OWNBUFFER | VO_SELECTABLE, 0);
	VIEW_DEBUG_NAME("ht_color_block");
	ht_color_block_data d;
	d.color=selected;
	setdata(&d);
	if (_flags & cf_light) colors=16; else colors=8;
	flags=_flags;
}

void ht_color_block::done()
{
	ht_view::done();
}

int ht_color_block::datasize()
{
	return sizeof (ht_color_block_data);
}

void ht_color_block::defaultpalette()
{
	pal_name=dialog_palette_str;
}

void ht_color_block::draw()
{
	clear(windialog_color_body);
	dword cursor=VCP(focused ? VC_LIGHT(VC_WHITE) : VC_BLACK, VC_TRANSPARENT);
	for (int i=0; i<colors; i++) {
		printchar((i%4)*3+1, i/4, VCP(vcs[i], VC_TRANSPARENT), CHAR_FILLED_F);
		printchar((i%4)*3+2, i/4, VCP(vcs[i], VC_BLACK), CHAR_FILLED_M);
		if (i==color) {
			printchar((i%4)*3, i/4, cursor, '>');
			printchar((i%4)*3+3, i/4, cursor, '<');
		}
	}
	if (flags & cf_transparent) {
		print(1, (colors==8) ? 2 : 4, VCP(VC_BLACK, VC_TRANSPARENT), "transparent");
		if (color==-1) {
			printchar(0, (colors==8) ? 2 : 4, cursor, '>');
			printchar(12, (colors==8) ? 2 : 4, cursor, '<');
		}
	}
}

void ht_color_block::getdata(void *buf)
{
	((ht_color_block_data*)buf)->color=(color==-1) ? VC_TRANSPARENT : vcs[color];
}

void ht_color_block::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
				if (color==-1) color=(flags & cf_light) ? 15 : 7; else
					if (color%4-1>=0) color--;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right:
				if (color==-1) color=(flags & cf_light) ? 15 : 7; else
					if (color%4+1<=3) color++;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Up:
				if (color==-1) color=(flags & cf_light) ? 15 : 7; else
					if (color-4>=0) color-=4;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
				if (color!=-1)
					if (color+4<colors) color+=4; else color=-1;
				dirtyview();
				clearmsg(msg);
				return;
		}
	}
	ht_view::handlemsg(msg);
}

void ht_color_block::setdata(void *buf)
{
	int c=((ht_color_block_data*)buf)->color;
	if (c==VC_TRANSPARENT) color=-1; else {
		for (int i=0; i<16; i++) if (vcs[i]==c) {
			color=i;
			break;
		}
	}
	dirtyview();
}

/*
 *	CLASS ht_color_manager
 */

void ht_color_manager::init(bounds *b, palette_group *_pal_grp)
{
	ht_dialog::init(b, "colors", FS_KILLER | FS_TITLE);
	VIEW_DEBUG_NAME("ht_color_manager");

	ipalette *i;

	orig_pal_grp=_pal_grp;
	pal_grp=orig_pal_grp->copy();

	bounds c;
/* group */
	c.x=1;
	c.y=1;
	c.w=25;
	c.h=5;
	grp=new ht_text_listbox();
	grp->init(&c);
	i=pal_grp->enum_first();
	int grp_it=0;
	while (i) {
		grp->insert_str(grp_it++, i->name);
		i=pal_grp->enum_next();
	}
	grp->update();
	insert(grp);
/* group label */
	c.x=1;
	c.y=0;
	c.w=10;
	c.h=2;
	ht_label *glabel=new ht_label();
	glabel->init(&c, "~group", grp);
	insert(glabel);

/* items */
	c.x=1;
	c.y=8;
	c.w=25;
	c.h=5;
	items=new ht_xgroup();
	items->init(&c, VO_SELECTABLE, "items");
	c.x=0;
	c.y=0;
	group_to_items=new ht_collection();
	i=pal_grp->enum_first();
	while (i) {
		palette *p=i->pal;
		ht_text_listbox *itemlist=new ht_text_listbox();
		itemlist->init(&c);
		group_to_items->insert(new ht_listbox_ptr(itemlist));
		for (int k=0; k<i->entries; k++) {
			itemlist->insert_str(k, p->name);
			p++;
		}
		itemlist->update();
		items->insert(itemlist);
		i=pal_grp->enum_next();
	}
	insert(items);
/* item label */
	c.x=1;
	c.y=7;
	c.w=10;
	c.h=1;
	ht_label *ilabel=new ht_label();
	ilabel->init(&c, "~item", items);
	insert(ilabel);

/* foreground colors */
	c.x=29;
	c.y=1;
	c.w=13;
	c.h=5;
	fg_colors=new ht_color_block();
	fg_colors->init(&c, 0, cf_light | cf_transparent);
	insert(fg_colors);
/* foreground label */
	c.x=30;
	c.y=0;
	c.w=10;
	c.h=1;
	ht_label *fglabel=new ht_label();
	fglabel->init(&c, "~foreground", fg_colors);
	insert(fglabel);

/* background colors */
	c.x=29;
	c.y=8;
	c.w=13;
	c.h=5;
	bg_colors=new ht_color_block();
	bg_colors->init(&c, 0, cf_light | cf_transparent);
	insert(bg_colors);
/* background label */
	c.x=30;
	c.y=7;
	c.w=10;
	c.h=1;
	ht_label *bglabel=new ht_label();
	bglabel->init(&c, "~background", bg_colors);
	insert(bglabel);

/* select the current item */
	sendmsg(htm_keypressed, K_Up);	/* FIXME: this a hack ! */
}

void ht_color_manager::done()
{
	delete group_to_items;
	pal_grp->done();
	delete pal_grp;
	ht_dialog::done();
}

int ht_color_manager::datasize()
{
	return sizeof (ht_color_manager_data);
}

void ht_color_manager::getdata(void *buf)
{
	((ht_color_manager_data*)buf)->pal_grp=pal_grp;
}

void ht_color_manager::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {

		ht_dialog::handlemsg(msg);

		if (grp->isviewdirty() || items->getselected()->isviewdirty()) {
			ht_view *cur=items->group->current;
			ht_listbox_data data;
			grp->getdata(&data);
			int idx=((ht_text_listbox_item*)data.cursor)->id;
			ht_listbox_ptr *p=(ht_listbox_ptr*)group_to_items->get(idx);
			ht_text_listbox_item *s=(ht_text_listbox_item*)data.cursor;
			palette *q=pal_grp->get(s->data[0]);
			if (q) {
				p->listbox->getdata(&data);
				char *name;
				int color;
				idx=((ht_text_listbox_item*)data.cursor)->id;
				if (pal_grp->get_entry(q, idx, &name, &color)) {
					ht_color_block_data d;
					d.color=VCP_FOREGROUND(color);
					fg_colors->setdata(&d);
					d.color=VCP_BACKGROUND(color);
					bg_colors->setdata(&d);
				}
			}
			if (grp->isviewdirty()) {
				focus(p->listbox);
				items->group->focus(cur);
			}
		} else if (fg_colors->isviewdirty() || bg_colors->isviewdirty()) {
			ht_listbox_data data;
			grp->getdata(&data);
			int idx=((ht_text_listbox_item*)data.cursor)->id;
			ht_listbox_ptr *p=(ht_listbox_ptr*)group_to_items->get(idx);
			ht_text_listbox_item *s=(ht_text_listbox_item*)data.cursor;
			palette *q=pal_grp->get(s->data[0]);
			if (q) {
				p->listbox->getdata(&data);
				char *name;
				int fg, bg;
				int color;
				idx=((ht_text_listbox_item*)data.cursor)->id;
				if (pal_grp->get_entry(q, idx, &name, &color)) {
					ht_color_block_data d;
					fg_colors->getdata(&d);
					fg=d.color;
					bg_colors->getdata(&d);
					bg=d.color;
					pal_grp->set_entry(q, idx, name, VCP(fg, bg));
				}
			}
		}
	} else {
		ht_dialog::handlemsg(msg);
	}
}

/*
 *	CLASS ht_search_dialog
 */

/* STATIC! */
int ht_search_dialog_lasttype=-1;
/**/

void ht_search_dialog::init(bounds *b)
{
	ht_dialog::init(b, "search", FS_KILLER | FS_TITLE);
	VIEW_DEBUG_NAME("ht_search_dialog");

	modecount=0;

	bounds c;
	c.x=1;
	c.y=1;
	c.w=20;
	c.h=1;

	searchmodes=new ht_listpopup();
	searchmodes->init(&c);
	insert(searchmodes);

	c.x=1;
	c.y=0;
	c.w=4;
	c.h=1;
	ht_label *mlabel=new ht_label();
	mlabel->init(&c, "~mode", searchmodes);
	insert(mlabel);

	c.x=1;
	c.y=3;
	c.w=size.w-4;
	c.h=size.h-4;
	modes=new ht_xgroup();
	modes->init(&c, VO_SELECTABLE, "modes");
	insert(modes);
}

void ht_search_dialog::done()
{
//     ht_search_dialog_lasttype=mode[modeidx].id;
	ht_dialog::done();
}

int ht_search_dialog::datasize()
{
	return sizeof (ht_search_dialog_data);
}

int ht_search_dialog::findmode(int id)
{
	for (int i=0; i<modecount; i++) {
		if (mode[i].id==id) {
			return i;
		}
	}
	return -1;
}

void ht_search_dialog::getdata(void *buf)
{
	((ht_search_dialog_data*)buf)->id=mode[modeidx].id;
	((ht_search_dialog_data*)buf)->view=mode[modeidx].view;
}

void ht_search_dialog::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		ht_dialog::handlemsg(msg);
		ht_listpopup_data data;
		searchmodes->getdata(&data);
		if (data.index!=modeidx) {
			modeidx=data.index;
			selectmode();
		}
	} else ht_dialog::handlemsg(msg);
}

void ht_search_dialog::insertmode(int id, char *desc, ht_view *v)
{
	if (modecount<MAX_SEARCH_DIALOG_MODES) {
		modes->insert(v);
		searchmodes->insertstring(desc);
		mode[modecount].id=id;
		mode[modecount].view=v;
		modecount++;
		focus(modes);
	}
}

void ht_search_dialog::selectmode()
{
	ht_listpopup_data d;
	d.index=modeidx;
	searchmodes->setdata(&d);
	sendmsg(htm_dirtyview, 0);
	modes->select(mode[modeidx].view);
}

