/* 
 *	HT Editor
 *	htdialog.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTDIALOG_H__
#define __HTDIALOG_H__

#include "htdebug.h"
#include "htobj.h"

/*
 *	CLASS ht_dialog
 */

#define ds_normal			0
#define ds_term_ok			1
#define ds_term_cancel		2

#define button_cancel		0
#define button_ok			1
#define button_yes            button_ok
#define button_no			2

class ht_dialog: public ht_window {
private:
	int state;
	int return_val;
public:
			void	init(bounds *b, char *desc, int framestyle);
	virtual	void	done();
/* overwritten */
	virtual	int alone();
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	int getstate(int *return_val);
	virtual	void setstate(int state, int return_val);
};

/*
 *	CLASS ht_cluster
 */

class ht_cluster: public ht_view {
protected:
			ht_string_collection *strings;
			int sel;
			int scount;
			int shortcuts[32];
public:
			void	init(bounds *b, ht_string_collection *strings);
	virtual	void	done();
	virtual	void defaultpalette();
};

/*
 *	CLASS ht_checkboxes
 */

struct ht_checkboxes_data {
	dword state;
};

class ht_checkboxes: public ht_cluster {
protected:
			dword state;
public:
			void	init(bounds *b, ht_string_collection *strings);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void getdata(void *buf);
	virtual	void setdata(void *buf);
};

/*
 *	CLASS ht_radioboxes
 */

struct ht_radioboxes_data {
	dword sel;
};

class ht_radioboxes: public ht_cluster {
public:
			void	init(bounds *b, ht_string_collection *strings);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void getdata(void *buf);
};

/*
 *	CLASS ht_inputfield
 */

class ht_inputfield;

struct ht_inputfield_data {
	int isattached;
	union {
		struct {
			ht_inputfield *attachedto;
		} attached;
		struct {
			int textlen;
			int maxtextlen;
			byte *text;
		} pure;
	};
};

class ht_inputfield: public ht_view {
protected:
			byte **text, *textv;
			byte **curchar, *curcharv;
			byte **selstart, *selstartv;
			byte **selend, *selendv;
			int *textlen, textlenv;
			int *maxtextlen, maxtextlenv;
			int insert;
			int ofs;
			ht_inputfield *attachedto;
			ht_collection *history;

			void freebuf();
			int insertbyte(byte *pos, byte b);
			void select_add(byte *start, byte *end);
public:
			void	init(bounds *b, int maxtextlen, ht_collection *history=0);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void defaultpalette();
	virtual	void getdata(void *buf);
	virtual	void setdata(void *buf);
/* new */
			void attach(ht_inputfield *inputfield);
			void query(byte ***curchar, byte ***text, byte ***selstart, byte ***selend, int **textlen, int **maxtextlen);
};

/*
 *	CLASS ht_strinputfield
 */

#define ht_strinputfield_data ht_inputfield_data

class ht_strinputfield: public ht_inputfield {
protected:
	int is_virgin;			/* untouched except for cursor keys */
			void correct_viewpoint();
public:
			void	init(bounds *b, int maxtextlen, ht_collection *history=0);
	virtual	void	done();
/* overwritten */
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void receivefocus();
/* new */
			void setbyte(byte a);
};

/*
 *	CLASS ht_hexinputfield
 */

#define ht_hexinputfield_data ht_inputfield_data

class ht_hexinputfield: public ht_inputfield {
protected:
			int nib;

			void correct_viewpoint();
public:
			void	init(bounds *b, int maxtextlen);
	virtual	void	done();
/* overwritten */
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
	virtual	void receivefocus();
/* new */
			void setnibble(byte a);
};

/*
 *	CLASS ht_button
 */

struct ht_button_data {
	int pressed;
};

class ht_button: public ht_view {
protected:
	int value;
	int pressed;
	char *text;
	char *magicchar;
	int shortcut;
public:
			void	init(bounds *b, char *text, int value);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void defaultpalette();
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	void push();
};

/*
 *	ht_listbox
 */

struct ht_listbox_data {
	void *top;
	void *cursor;
};

class ht_listbox: public ht_view {
public:
	int cursor, pos, count;
	int visible_height;
	void *e_top, *e_cursor;
	ht_scrollbar *scrollbar;
	int x;
	char     quickfinder[100];
	char     *qpos;

			void	init(bounds *b);
	virtual	void	done();
			void adjust_scrollbar();
	virtual   int  calc_count();                                // abstract
			void clear_quickfind();
	virtual   int  cursor_adjust();						// abstract
			int  cursor_up(int n);
			int  cursor_down(int n);
	virtual	int	datasize();
	virtual	void defaultpalette();
	virtual	void draw();
	virtual	void getdata(void *buf);
	virtual   void *getfirst();							// abstract
	virtual   void *getlast();							// abstract
	virtual   void *getnext(void *entry);					// abstract
	virtual   void *getnth(int n);						// abstract
	virtual   void *getprev(void *entry);					// abstract
	virtual   char *getstr(int col, void *entry);			// abstract
	virtual	void handlemsg(htmsg *msg);
	virtual   int	num_cols();							// stub
	virtual	void *quickfind(char *s);					// abstract
	virtual	void redraw();
	virtual	bool	seek(int index);
	virtual   void select_entry(void *entry);
	virtual	void setdata(void *buf);
	virtual	void update();
			void update_cursor();
};

/*
 *	CLASS ht_text_listbox
 */

struct ht_text_listbox_item {
	ht_text_listbox_item	*next, *prev;
	int				id;
	char				*data[0];
};

class ht_text_listbox: public ht_listbox {
public:
	int				cols, keycol, count;
	ht_text_listbox_item	*first, *last;
	int				*widths;
	char				*return_str;
	int				Cursor_adjust;
			void	init(bounds *b, int Cols=1, int Keycol=0);
	virtual	void	done();
	virtual   int  calc_count();
	virtual   int	cursor_adjust();
	virtual   void	*getfirst();
	virtual   void	*getlast();
	virtual   void	*getnext(void *entry);
	virtual   void	*getprev(void *entry);
	virtual   char	*getstr(int col, void *entry);
	virtual   int	num_cols();
	virtual	void	*quickfind(char *s);
	virtual   void update();
/* new */
			void	insert_str(int id, char *str, ...);
};

/*
 *	CLASS ht_listpopup
 */

struct ht_listpopup_data {
	ht_string_collection *list;
	int index;
};

class ht_listpopup: public ht_view {
protected:
	ht_string_collection *list;
	int pos, cursor;
	int vis_h_size;
	ht_scrollbar *scrollbar;
	int local_scrollbar;
	int expanded;

/* overwritten */
	virtual	void adjust_scrollbar();
public:
			void	init(bounds *b);
	virtual	void	done();
/* overwritten */
			void collapse();
	virtual	int datasize();
	virtual	void defaultpalette();
	virtual	void draw();
			void expand();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
	virtual	void insertstring(char *str);
	virtual	void redraw();
	virtual	void releasefocus();
	virtual	int seek(int index);
	virtual	void setdata(void *buf);
};

/*
 *	CLASS ht_statictext
 */

#define align_left		0
#define align_center	1
#define align_right		2
#define align_custom	3

#define ALIGN_CHAR_ESCAPE	'\e'
#define ALIGN_CHAR_LEFT		'l'
#define ALIGN_CHAR_CENTER	'c'
#define ALIGN_CHAR_RIGHT		'r'
 
struct ht_statictext_linedesc {
	int ofs;
	int len;
	char *text;
};

class ht_statictext: public ht_view {
	char *text;
	int align;
public:
			void	init(bounds *b, char *text, int align);
	virtual	void	done();
/* overwritten */
	virtual	void defaultpalette();
	virtual 	void draw();
/* new */
	virtual	void settext(char *text);
	virtual	char *gettext();
};

/*
 *	CLASS ht_label
 */

class ht_label: public ht_view {
protected:
	ht_view *connected;
	char *text;
	char *magicchar;
	int shortcut;
public:
			void	init(bounds *b, char *text, ht_view *connected);
	virtual	void	done();
/* overwritten */
	virtual	void defaultpalette();
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_progress_indicator_body
 */

class ht_progress_indicator_body: public ht_view {
protected:
	char *format;
	int indicator;
public:
			void	init(bounds *b, char *format);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void defaultpalette();
	virtual	void draw();
	virtual	void setdata(void *buf);
};

/*
 *	CLASS ht_progress_indicator
 */

struct ht_progress_indicator_data {
	int indicator;
};

class ht_progress_indicator: public ht_window {
protected:
	ht_progress_indicator_body *body;
public:
			void	init(bounds *b, char *format);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void setdata(void *buf);
};

/*
 *	CLASS ht_color_block
 */

struct ht_color_block_data {
	int color;
};

#define cf_light		1
#define cf_transparent   2

class ht_color_block: public ht_view {
protected:
	int color;
	int colors;
	int flags;
public:
			void	init(bounds *b, int selected, int flags);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void defaultpalette();
	virtual	void draw();
	virtual	void getdata(void *buf);
	virtual	void setdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_listbox_ptr
 */

class ht_listbox_ptr: public ht_data {
public:
	ht_listbox *listbox;

	ht_listbox_ptr(ht_listbox *listbox);
	~ht_listbox_ptr();
};

/*
 *	CLASS ht_color_manager
 */

struct ht_color_manager_data {
	palette_group *pal_grp;
};

class ht_color_manager: public ht_dialog {
protected:
	ht_text_listbox *grp;
	ht_xgroup *items;
	ht_collection *group_to_items;
	ht_color_block *fg_colors;
	ht_color_block *bg_colors;
	palette_group *orig_pal_grp;
	palette_group *pal_grp;
public:
			void	init(bounds *b, palette_group *pal);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_search_dialog
 */

struct ht_search_dialog_data {
	int id;
	ht_view *view;
};

struct ht_search_dialog_mode {
	int id;
	ht_view *view;
};

#define MAX_SEARCH_DIALOG_MODES 16

class ht_search_dialog: public ht_dialog {
protected:
	ht_listpopup *searchmodes;
	ht_xgroup *modes;

	int modeidx;
	ht_search_dialog_mode mode[MAX_SEARCH_DIALOG_MODES];
	int modecount;

	int findmode(int id);
	void selectmode();
public:
			void	init(bounds *b);
	virtual	void	done();
/* overwritten */
	virtual	int datasize();
	virtual	void getdata(void *buf);
	virtual	void handlemsg(htmsg *msg);
/* new */
			void insertmode(int id, char *desc, ht_view *v);
};

/*
 *	misc
 */

int execute_dialog(ht_dialog *dialog);
int execute_dialog_modal(ht_dialog *dialog);

#endif /* !__HTDIALOG_H__ */

