/* 
 *	HT Editor
 *	htiobox.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htatom.h"
#include "htctrl.h"
#include "htdialog.h"
#include "htiobox.h"
#include "tools.h"

#include <stdlib.h>
#include <string.h>

int imsgbox(bounds *b, int buttonmask, char *title, int modal, int align, char *buf)
{
	ht_dialog *dialog=new ht_dialog();
	dialog->init(b, title, FS_KILLER | FS_TITLE);

	bounds c;

	c.x=1;
	c.y=0;
	c.w=b->w-4;
	c.h=b->h-4;
	ht_statictext *text=new ht_statictext();
	text->init(&c, buf, align);
	dialog->insert(text);

	int buttons=0;
	if (buttonmask & msgbox_ok) buttons++;
	if (buttonmask & msgbox_cancel) buttons++;
	if (buttonmask & msgbox_yes) buttons++;
	if (buttonmask & msgbox_no) buttons++;

	int pos=0;
	if (buttonmask & msgbox_ok) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-4;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button();
		button->init(&c, "O~K", button_ok);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_yes) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-4;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button();
		button->init(&c, "~Yes", button_yes);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_no) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-4;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button();
		button->init(&c, "~No", button_no);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_cancel) {
		c.x=b->w/2-5*(buttons-2*pos);
		c.y=b->h-4;
		c.w=9;
		c.h=2;
		ht_button *button=new ht_button();
		button->init(&c, "Cancel", button_cancel);
		dialog->insert(button);
		pos++;
	}
	if (!buttons) {
		bounds x;
		text->getbounds(&x);
		x.h+=2;
		text->setbounds(&x);
	}

	int r;
	if (modal) r=execute_dialog_modal(dialog); else r=execute_dialog(dialog);
	dialog->done();
	delete dialog;
	return r;
}

int msgbox(int buttonmask, char *title, int modal, int align, char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	int strl=strlen(buf);

	bounds b;
	app->getbounds(&b);
	b.w=45;
	b.h=strl/(b.w-4)+6;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	return imsgbox(&b, buttonmask, title, modal, align, buf);
}

int msgboxrect(bounds *b, int buttonmask, char *title, int modal, int align, char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);

	return imsgbox(b, buttonmask, title, modal, align, buf);
}

int inputbox(char *title, char *Label, char *result, int limit, dword histid)
{
	bounds b;
	app->getbounds(&b);
	b.x = (b.w - 60) / 2,
	b.y = (b.h - 8) / 2;
	b.w = 60;
	b.h = 8;
	return inputboxrect(&b, title, Label, result, limit, histid);
}

int inputboxrect(bounds *b, char *title, char *Label, char *result, int limit, dword histid)
{
	ht_dialog *dialog=new ht_dialog();
	dialog->init(b, title, FS_KILLER | FS_TITLE);

	bounds  b2;
	b2.x = 3 + strlen(Label);
	b2.y = 1;
	b2.w = b->w - 3 - b2.x;
	b2.h = 1;

	ht_collection *hist=0;
	if (histid) hist=(ht_collection*)find_atom(histid);
	ht_strinputfield *input=new ht_strinputfield();
	input->init(&b2, limit, hist);
	ht_inputfield_data d;
	d.isattached=0;
	d.pure.text=(byte *)result;
	d.pure.textlen=strlen(result);
	d.pure.maxtextlen=limit;
	input->setdata(&d);
	dialog->insert(input);

	if (Label) {
		b2.x = 1;
		b2.y = 1;
		b2.w = 3 + strlen(Label) - b2.x;
		b2.h = 1;

		ht_statictext *lab=new ht_statictext();
		lab->init(&b2, Label, 0);
		dialog->insert(lab);
	}

	b2.x = b->w - 25;
	b2.y = b->h - 5;
	b2.w = 10;
	b2.h = 2;

	ht_button *bok=new ht_button();
	bok->init(&b2, "O~k", button_ok);
	dialog->insert(bok);

	b2.x += 12;

	ht_button *bcancel=new ht_button();
	bcancel->init(&b2, "Cancel", button_cancel);
	dialog->insert(bcancel);

	if (execute_dialog(dialog)) {
		ht_inputfield_data *data=(ht_inputfield_data*)malloc(input->datasize());
		input->getdata(data);
		memmove(result, data->pure.text, data->pure.textlen);
		result[data->pure.textlen] = 0;
		if (hist) {
			ht_data_ptr *d=new ht_data_ptr(data);
			hist->insert(d);
		} else {
			delete data;
		}
		dialog->done();
		delete dialog;
		return true;
	}
	dialog->done();
	delete dialog;
	return false;
}

void get_std_progress_indicator_metrics(bounds *b)
{
	app->getbounds(b);
	b->w=b->w*4/5;
	b->h=6;
	b->x=(screen->size.w-b->w)/2;
	b->y=(screen->size.h-b->h)/2;
}

