/* 
 *	HT Editor
 *	htleimg.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htdebug.h"
#include "htnewexe.h"
#include "htle.h"
#include "htleimg.h"
#include "htstring.h"
#include "htviewer.h"
#include "viewers.h"
#include "x86dis.h"

#include "lestruct.h"

#include <string.h>

class x86dis_le: public x86dis {
public:
	x86dis_le(int opsize, int addrsize);
	virtual ~x86dis_le();

	virtual dis_insn *decode(byte *code, byte maxlen, CPU_ADDR addr);
};

/**/

ht_view *htleimage_init(bounds *b, ht_stream *file, ht_group *group)
{
	ht_group *le_group=group;
	while (strcmp(le_group->desc, DESC_LE)) le_group=le_group->group;
	ht_le_shared_data *le_shared=(ht_le_shared_data *)((ht_le*)le_group)->shared_data;

	ht_le_page_stream *mfile=new ht_le_page_stream();
	mfile->init(file, 1, &le_shared->pagemap, le_shared->pagemap.count,
		le_shared->hdr.pagesize);

	ht_uviewer *v=new ht_uviewer();
	v->init(b, "le/image", mfile, VC_EDIT | VC_GOTO);

	disassembler *disasm=new x86dis_le(X86_OPSIZE32, X86_ADDRSIZE32);

	for (int i=0; i<le_shared->objmap.count; i++) {
		hts_mask *m=new hts_mask();
		m->init(mfile);
		char t[64];
		char n[5];
		memmove(n, le_shared->objmap.header[i].name, 4);
		n[4]=0;
		sprintf(t, "section %d: %s (psize=%08lx, vsize=%08lx)", i, n, le_shared->objmap.psize[i], le_shared->objmap.vsize[i]);
		m->add_mask(t);
		v->insertsub(m);

		hts_unasm *d=new hts_unasm();
		d->init(mfile, (le_shared->objmap.header[i].page_map_index-1)*
			le_shared->hdr.pagesize, le_shared->objmap.vsize[i], disasm,
			X86DIS_STYLE_OPTIMIZE_ADDR);
		v->insertsub(d);
	}

	return v;
}

viewer_if htleimage_if = {
	htleimage_init,
	0
};

/*
 *	CLASS ht_le_page_stream
 */

void ht_le_page_stream::init(ht_stream *stream, int own_stream, ht_le_pagemap *_pagemap, dword _pagemapsize, dword _page_size)
{
	ht_layer_stream::init(stream, own_stream);
	pagemap=_pagemap;
	pagemapsize=_pagemapsize;
	page_size=_page_size;
	ofs=0;
}

void ht_le_page_stream::done()
{
	ht_layer_stream::done();
}

int ht_le_page_stream::isdirty(dword offset, int range)
{
	dword mofs, msize;
	while (range) {
		dword s=range;
		if (!map_ofs(offset, &mofs, &msize)) break;
		if (s>msize) s=msize;
		if (stream->isdirty(mofs, s)) return 1;
		range-=s;
		ofs+=s;
	}
	return 0;
}

int ht_le_page_stream::map_ofs(dword qofs, dword *offset, dword *maxsize)
{
	dword i=qofs/page_size, j=qofs % page_size;
	if (i<pagemapsize) {
		if (j<pagemap->vsize[i]) {
			*offset=pagemap->offset[i]+j;
			*maxsize=pagemap->vsize[i]-j;
			return 1;
		}
	}
	return 0;
}

dword ht_le_page_stream::read(void *buf, dword size)
{
	dword mofs, msize;
	int c=0;
	while (size) {
		dword s=size;
		if (!map_ofs(ofs, &mofs, &msize)) break;
		if (s>msize) s=msize;
		stream->seek(mofs);
		s=stream->read(buf, s);
		if (!s) break;
		((byte*)buf)+=s;
		size-=s;
		c+=s;
		ofs+=s;
	}
	return c;
}

int ht_le_page_stream::seek(dword offset)
{
	ofs=offset;
	return 1;
}

dword ht_le_page_stream::tell()
{
	return ofs;
}

dword ht_le_page_stream::write(void *buf, dword size)
{
	dword mofs, msize;
	int c=0;
	while (size) {
		dword s=size;
		if (!map_ofs(ofs, &mofs, &msize)) break;
		if (s>msize) s=msize;
		stream->seek(mofs);
		((byte*)buf)+=stream->write(buf, s);
		size-=s;
		c+=s;
		ofs+=s;
	}
	return c;
}

/*
 *	CLASS x86dis_le
 */

x86dis_le::x86dis_le(int opsize, int addrsize)
: x86dis(opsize, addrsize)
{
}

x86dis_le::~x86dis_le()
{
}

dis_insn *x86dis_le::decode(byte *code, byte maxlen, CPU_ADDR addr)
{
	if ((maxlen>=6) && (code[0]==0xcd) && (code[1]==0x20)) {
		insn.name="VxDCall";
		insn.size=6;
		insn.op[0].type=X86_OPTYPE_IMM;
		insn.op[0].size=2;
		insn.op[0].imm=*(word*)(code+2);
		insn.op[1].type=X86_OPTYPE_IMM;
		insn.op[1].size=2;
		insn.op[1].imm=*(word*)(code+4);
		insn.op[2].type=X86_OPTYPE_EMPTY;
		insn.lockprefix=X86DIS_PREFIX_NO;
		insn.repprefix=X86DIS_PREFIX_NO;
		insn.segprefix=X86DIS_PREFIX_NO;
		return &insn;
	}
	return x86dis::decode(code, maxlen, addr);
}

