/* 
 *	HT Editor
 *	htmenu.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htcolor.h"
#include "htctrl.h"
#include "htkeyb.h"
#include "htmenu.h"
#include "htstring.h"
#include "tools.h"

#include <string.h>

/* menu palette */

#define menu_palette_str				"menu"
#define menu_palette_colors			6
#define menu_coloridx_normal			0
#define menu_coloridx_normal_shortcut	1
#define menu_coloridx_selected		2
#define menu_coloridx_selected_shortcut	3
#define menu_coloridx_disabled		4
#define menu_coloridx_disabled_selected	5
#define menu_color_normal			pal[menu_coloridx_normal].value
#define menu_color_normal_shortcut		pal[menu_coloridx_normal_shortcut].value
#define menu_color_selected			pal[menu_coloridx_selected].value
#define menu_color_selected_shortcut	pal[menu_coloridx_selected_shortcut].value
#define menu_color_disabled			pal[menu_coloridx_disabled].value
#define menu_color_disabled_selected	pal[menu_coloridx_disabled_selected].value

#define button_left		100
#define button_right	101

char *shortcut_str(char *str)
{
	while (*str) {
		if (*str=='~') {
			memmove(str, str+1, strlen(str));
			return str;
		}
		str++;
	}
	return 0;
}

/*
 *	CLASS ht_context_menu
 */

void ht_context_menu::init(char *_name)
{
	_name=strdup(_name);
	shortcut=shortcut_str(_name);
	name=_name;
	width=0;
}

void ht_context_menu::done()
{
	delete name;
}

int ht_context_menu::count()
{
	return 0;
}

ht_context_menu_entry *ht_context_menu::enum_entry_first()
{
	return 0;
}

ht_context_menu_entry *ht_context_menu::enum_entry_next()
{
	return 0;
}

ht_context_menu_entry *ht_context_menu::get_entry(int n)
{
	ht_context_menu_entry *e=enum_entry_first();
	while (n--) {
		if (!e) break;
		e=enum_entry_next();
	}
	return e;
}

/*
 *	CLASS ht_static_context_menu
 */

void ht_static_context_menu::init(char *name)
{
	ht_context_menu::init(name);
	context_menu_entry=new ht_collection();
}

void ht_static_context_menu::done()
{
	delete context_menu_entry;
	ht_context_menu::done();
}

int ht_static_context_menu::count()
{
	return context_menu_entry->count();
}

ht_context_menu_entry *ht_static_context_menu::enum_entry_first()
{
	enum_idx=0;
	ht_context_menu_entry *e=(ht_context_menu_entry*)context_menu_entry->get(enum_idx);
	return e;
}

ht_context_menu_entry *ht_static_context_menu::enum_entry_next()
{
	enum_idx++;
	ht_context_menu_entry *e=(ht_context_menu_entry*)context_menu_entry->get(enum_idx);
	return e;
}

int ht_static_context_menu::insert_entry(char *_name, char *_comment, int _command, int _key, int _active)
{
	char *name=strdup(_name), *comment, *shortcut=0, *n=name;
	if (_comment) comment=strdup(_comment); else comment=0;
	while (*n) {
		if (*n=='~') {
			memmove(n, n+1, strlen(n));
			shortcut=n;
			break;
		}
		n++;
	}
	int l=strlen(name), cl=comment ? strlen(comment)+2 : 0;
	width=MAX(width, l+cl);
	ht_context_menu_entry *entry=new ht_context_menu_entry();
	entry->name=name;
	entry->shortcut=shortcut;
	entry->comment=comment;
	entry->command=_command;
	entry->key=_key;
	entry->active=_active;
	context_menu_entry->insert(entry);
	return 0;
}

int ht_static_context_menu::insert_separator()
{
	ht_context_menu_entry *entry=new ht_context_menu_entry();
	entry->name=0;
	entry->shortcut=0;
	entry->command=0;
	entry->active=0;
	context_menu_entry->insert(entry);
	return 0;
}

/*
 *	CLASS ht_menu
 */

void ht_menu::init(bounds *b)
{
	ht_view::init(b, VO_OWNBUFFER | VO_POSTPROCESS, "menu");
	VIEW_DEBUG_NAME("ht_menu");

	menu=new ht_collection();
	lastmenux=1;
	curmenu=-1;
}

void ht_menu::done()
{
	ht_view::done();
}

void ht_menu::defaultpalette()
{
	pal_name=menu_palette_str;
}

void ht_menu::draw()
{
	clear(menu_color_normal);
	int c=menu->count();
	for (int i=0; i<c; i++) {
		ht_context_menu *c=(ht_context_menu*)menu->get(i);
		if (i==curmenu) {
			printchar(c->xpos-1, 0, menu_color_selected, ' ');
			print(c->xpos, 0, menu_color_selected, c->name);
			printchar(c->xpos+strlen(c->name), 0, menu_color_selected, ' ');
			if (c->shortcut) printchar(c->xpos+(c->shortcut-c->name), 0, menu_color_selected_shortcut, *c->shortcut);
		} else {
			print(c->xpos, 0, menu_color_normal, c->name);
			if (c->shortcut) printchar(c->xpos+(c->shortcut-c->name), 0, menu_color_normal_shortcut, *c->shortcut);
		}
	}
}

void ht_menu::execute_menu(int i)
{
	curmenu=i;
	int cur_entry=0;
	curmenu=i;
	int term=0;
	dirtyview();
	do {
		ht_context_menu *m=(ht_context_menu*)menu->get(curmenu);
		bounds b;
		b.x=m->xpos;
		b.y=1;
		b.w=m->width+4;
		b.h=m->count()+2;
		ht_menu_window *d=new ht_menu_window();
		d->init(&b, m);
		ht_context_menu_window_data a;
		a.selected=cur_entry;
		d->setdata(&a);
		b.x=0;
		b.y=0;
		ht_frame *frame=new ht_menu_frame();
		frame->init(&b, "test", -1, 0);
		d->setframe(frame);
		int r=execute_dialog(d);
		switch (r) {
			case button_left:
				curmenu--;
				if (curmenu<0) curmenu=menu->count()-1;
				cur_entry=0;
				dirtyview();
				break;
			case button_right:
				curmenu++;
				if (curmenu>menu->count()-1) curmenu=0;
				cur_entry=0;
				dirtyview();
				break;
			case button_ok: {
				ht_context_menu_window_data a;
				d->getdata(&a);
				cur_entry=a.selected;
				ht_context_menu_entry *e=m->get_entry(a.selected);
				if (e->active) {
					term=1;
					curmenu=-1;
					dirtyview();
					app->sendmsg(htm_command, e->command);
				}
				break;
			}
			default:
				term=1;
		}
		d->done();
		delete d;
	} while (!term);
	curmenu=-1;
	dirtyview();
}

void ht_menu::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_postprocess) {
		if (msg->msg==htm_keypressed) {
/* shortcuts */
			int k=htalt2key(msg->data_int);
			if (k!=-1) {
				int c=menu->count();
				for (int i=0; i<c; i++) {
					ht_context_menu *m=(ht_context_menu*)menu->get(i);
					int s=*m->shortcut;
					if ((s>='A') && (s<='Z')) s+='a'-'A';
					if (s==k) {
						execute_menu(i);
						return;
					}
				}
			}
/* keys associated with menu entries */
			int c=menu->count();
			for (int i=0; i<c; i++) {
				ht_context_menu *a=(ht_context_menu*)menu->get(i);
				int t=a->count();
				ht_context_menu_entry *e=a->enum_entry_first();
				for (int j=0; j<t; j++) {
					if ((e->key) && (msg->data_int==e->key)) {
						app->sendmsg(htm_command, e->command);
						clearmsg(msg);
						return;
					}
					e=a->enum_entry_next();
				}
			}
		}
	}
	ht_view::handlemsg(msg);
}

int ht_menu::insert_menu(ht_context_menu *m)
{
	int namelen=strlen(m->name);

	m->xpos=lastmenux;
	menu->insert(m);

	lastmenux+=namelen+1;
	return 1;
}

/*
 *	CLASS ht_menu_frame
 */

void ht_menu_frame::init(bounds *b, char *desc, int number, int style)
{
	ht_frame::init(b, desc, number, style);
	VIEW_DEBUG_NAME("ht_menu_frame");
}

void ht_menu_frame::done()
{
	ht_frame::done();
}

void ht_menu_frame::defaultpalette()
{
	pal_name=menu_palette_str;
}

int ht_menu_frame::getcurcol_normal()
{
	switch (framestate) {
		case FST_FOCUSED:
			return menu_color_normal;
		case FST_UNFOCUSED:
			return menu_color_normal;
		case FST_MOVE:
			return menu_color_normal;
	}
	return 0;
}

int ht_menu_frame::getcurcol_killer()
{
	return 0;
}

/*
 *	CLASS ht_context_menu_window
 */

void ht_context_menu_window::init(bounds *b, ht_context_menu *menu)
{
	ht_dialog::init(b, 0, 0);
	VIEW_DEBUG_NAME("ht_context_menu_window");
	bounds c=*b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	body=new ht_context_menu_window_body();
	body->init(&c, menu);
	insert(body);
}

void ht_context_menu_window::done()
{
	ht_dialog::done();
}

void ht_context_menu_window::getdata(void *buf)
{
	body->getdata(buf);
}

void ht_context_menu_window::setdata(void *buf)
{
	body->setdata(buf);
}

/*
 *	CLASS ht_context_menu_window_body
 */

void ht_context_menu_window_body::init(bounds *b, ht_context_menu *_menu)
{
	ht_view::init(b, VO_OWNBUFFER, 0);
	VIEW_DEBUG_NAME("ht_context_menu_window_body");
	context_menu=_menu;
	selected=next_selectable(-1);
}

void ht_context_menu_window_body::done()
{
	ht_view::done();
}

void ht_context_menu_window_body::defaultpalette()
{
	pal_name=menu_palette_str;
}

void ht_context_menu_window_body::draw()
{
	clear(menu_color_normal);
	int c=context_menu->count();
	if (c>size.h) c=size.h;
	ht_context_menu_entry *e=context_menu->enum_entry_first();
	for (int i=0; i<c; i++) {
		if (e->name) {
			if (i==selected) {
				if (e->active) {
					fill(0, i, size.w, 1, menu_color_selected, ' ');
					lprint(1, i, menu_color_selected, size.w, e->name);
					if (e->comment) lprint(size.w-1-strlen(e->comment), i, menu_color_selected, size.w, e->comment);
					if (e->shortcut) printchar(e->shortcut-e->name+1, i, menu_color_selected_shortcut, *e->shortcut);
				} else {
					fill(0, i, size.w, 1, menu_color_disabled_selected, ' ');
					lprint(1, i, menu_color_disabled_selected, size.w, e->name);
					if (e->comment) lprint(size.w-1-strlen(e->comment), i, menu_color_disabled_selected, size.w, e->comment);
				}
			} else {
				if (e->active) {
					lprint(1, i, menu_color_normal, size.w, e->name);
					if (e->comment) lprint(size.w-1-strlen(e->comment), i, menu_color_normal, size.w, e->comment);
					if (e->shortcut) printchar(e->shortcut-e->name+1, i, menu_color_normal_shortcut, *e->shortcut);
				} else {
					lprint(1, i, menu_color_disabled, size.w, e->name);
					if (e->comment) lprint(size.w-1-strlen(e->comment), i, menu_color_disabled, size.w, e->comment);
				}
			}
		} else {
			fill(0, i, size.w, 1, menu_color_normal, CHAR_LINEH);
		}
		e=context_menu->enum_entry_next();
	}
}

void ht_context_menu_window_body::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Up:
				selected=prev_selectable(selected);
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
				selected=next_selectable(selected);
				dirtyview();
				clearmsg(msg);
				return;
		}
/* shortcuts */
		int k=msg->data_int;
//		int k=htalt2key(msg->data_int);
//		if (k!=-1) {
			int c=context_menu->count();
			ht_context_menu_entry *e=context_menu->enum_entry_first();
			for (int i=0; i<c; i++) {
				if (e->shortcut) {
					int s=*e->shortcut;
					if ((s>='A') && (s<='Z')) s+='a'-'A';
					if (s==k) {
						selected=i;
						((ht_dialog*)group->group)->sendmsg(htm_button_pressed, button_ok);
						dirtyview();
						clearmsg(msg);
						return;
					}
				}					
				e=context_menu->enum_entry_next();
			}
//		}
	}
	ht_view::handlemsg(msg);
}

int ht_context_menu_window_body::next_selectable(int to)
{
	int s=to+1;
	int c=context_menu->count();
	if (s>c-1) s=0;
	while (s!=to) {
		ht_context_menu_entry *e=context_menu->get_entry(s);
		if (e->name) return s;
		s++;
		if (s>c-1) s=0;
	}
	return to;
}

int ht_context_menu_window_body::prev_selectable(int to)
{
	int s=to-1;
	int c=context_menu->count();
	if (s<0) s=c-1;
	while (s!=to) {
		ht_context_menu_entry *e=context_menu->get_entry(s);
		if (e->name) return s;
		s--;
		if (s<0) s=c-1;
	}
	return to;
}

void ht_context_menu_window_body::getdata(void *buf)
{
	((ht_context_menu_window_data*)buf)->selected=selected;
}

void ht_context_menu_window_body::setdata(void *buf)
{
	selected=((ht_context_menu_window_data*)buf)->selected;
}

/*
 *	CLASS ht_menu_window
 */

void ht_menu_window::init(bounds *b, ht_context_menu *menu)
{
	ht_dialog::init(b, 0, 0);
	VIEW_DEBUG_NAME("ht_menu_window");

	bounds c=*b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	body=new ht_menu_window_body();
	body->init(&c, menu);
	insert(body);
}

void ht_menu_window::done()
{
	ht_dialog::done();
}

void ht_menu_window::getdata(void *buf)
{
	body->getdata(buf);
}

void ht_menu_window::setdata(void *buf)
{
	body->setdata(buf);
}

/*
 *	CLASS ht_menu_window_body
 */

void ht_menu_window_body::init(bounds *b, ht_context_menu *menu)
{
	ht_context_menu_window_body::init(b, menu);
	VIEW_DEBUG_NAME("ht_menu_window_body");
}

void ht_menu_window_body::done()
{
	ht_context_menu_window_body::done();
}

void ht_menu_window_body::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
				((ht_dialog*)group->group)->setstate(ds_term_ok, button_left);
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right:
				((ht_dialog*)group->group)->setstate(ds_term_ok, button_right);
				dirtyview();
				clearmsg(msg);
				return;
		}
	}
	ht_context_menu_window_body::handlemsg(msg);
}

/*
 *	INIT
 */

int init_menu()
{
	palette *menu_palette=pal_grp->create(menu_palette_str, menu_palette_colors, PCLASS_MENU);
	pal_grp->set_entry(menu_palette, menu_coloridx_normal, "normal", VCP(VC_WHITE, VC_BLACK));
	pal_grp->set_entry(menu_palette, menu_coloridx_normal_shortcut, "shortcut normal", VCP(VC_RED, VC_TRANSPARENT));
	pal_grp->set_entry(menu_palette, menu_coloridx_selected, "selected", VCP(VC_BLACK, VC_GREEN));
	pal_grp->set_entry(menu_palette, menu_coloridx_selected_shortcut, "shortcut selected", VCP(VC_RED, VC_TRANSPARENT));
	pal_grp->set_entry(menu_palette, menu_coloridx_disabled, "disabled", VCP(VC_LIGHT(VC_BLACK), VC_TRANSPARENT));
	pal_grp->set_entry(menu_palette, menu_coloridx_disabled_selected, "disabled selected", VCP(VC_LIGHT(VC_BLACK), VC_GREEN));
	return 1;
}

/*
 *	DONE
 */

void done_menu()
{
}
