/*
 *	HT Editor
 *	htobj.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htcfg.h"
#include "htcolor.h"
#include "htctrl.h"
#include "htkeyb.h"
#include "htobj.h"
#include "htstring.h"
#include "tools.h"

#include <string.h>
#include <stdlib.h>

void bounds_and(bounds *a, bounds *b)
{
	if (b->x>a->x) {
		a->w-=b->x-a->x;
		a->x=b->x;
	}
	if (b->y>a->y) {
		a->h-=b->y-a->y;
		a->y=b->y;
	}
	if (a->x+a->w>b->x+b->w) a->w-=a->x+a->w-b->x-b->w;
	if (a->y+a->h>b->y+b->h) a->h-=a->y+a->h-b->y-b->h;
	if (a->w<0) a->w=0;
	if (a->h<0) a->h=0;
}

/*
 *	CLASS ht_view
 */

void ht_view::init(bounds *b, int _options, char *_desc)
{
	object::init();
	VIEW_DEBUG_NAME("ht_view");
	desc=strdup(_desc);
	group=0;
	focused=0;
	browse_idx=0;
	view_is_dirty=1;
	size.x=0;
	size.y=0;
	size.w=0;
	size.h=0;
	prev=0;
	next=0;
	setoptions(_options);
	buf=0;
	enabled=1;
	growmode_x=GROW_CSLT;
	growmode_y=GROW_CSLT;
	if (options & VO_OWNBUFFER) {
		buf=new drawbuf(&size);
		enable_buffering();
	} else {
		buf=screen;
		disable_buffering();
	}
	csize=*b;
	setbounds(b);
	defaultpalette();
	ipalette *p=pal_grp->iget(pal_name);
	pal_class=p->palclass;
	setpalette(pal_name);
}

void ht_view::done()
{
	if (desc) delete desc;
	object::done();
}

int ht_view::alone()
{
	return (group && group->isalone(this));
}

int ht_view::childcount()
{
	return 1;
}

void ht_view::cleanview()
{
	view_is_dirty=0;
}

void ht_view::clear(int c)
{
	buf->fill(vsize.x, vsize.y, vsize.w, vsize.h, c, ' ');
}

void ht_view::clearmsg(htmsg *msg)
{
	msg->msg=htm_nothing;
	msg->type=htmt_nothing;
}

void ht_view::clipbounds(bounds *b)
{
	bounds c;
	getclientarea(&c);
	bounds_and(b, &c);
	bounds_and(b, &vsize);
	if (group) group->clipbounds(b);
}

int ht_view::countselectables()
{
	return (options & VO_SELECTABLE) ? 1 : 0;
}

int ht_view::datasize()
{
	return 0;
}

void ht_view::defaultpalette()
{
	pal_name=window_palette_str;
}

void ht_view::dirtyview()
{
	view_is_dirty=1;
}

void ht_view::disable()
{
	enabled=0;
}

void ht_view::disable_buffering()
{
	if (options & VO_OWNBUFFER) {
		if (buf) delete buf;
		buf=screen;
		setoptions(options&(~VO_OWNBUFFER));
	}
}

STUB void ht_view::draw()
{
}

void ht_view::enable()
{
	enabled=1;
}

void ht_view::enable_buffering()
{
	if (!(options & VO_OWNBUFFER)) {
		buf=new drawbuf(&size);
		setoptions(options | VO_OWNBUFFER);
	}
}

void ht_view::fill(int x, int y, int w, int h, int c, int chr)
{
	bounds b;
	b.x=size.x+x;
	b.y=size.y+y;
	b.w=w;
	b.h=h;
	bounds_and(&b, &vsize);
	buf->fill(b.x, b.y, b.w, b.h, c, chr);
}

int ht_view::focus(ht_view *view)
{
	if (view==this) {
		if (!focused) receivefocus();
		return 1;
	}
	return 0;
}

ABSTRACT void ht_view::getdata(void *buf)
{
	abstract();
}

void ht_view::getbounds(bounds *b)
{
	*b=size;
}

void ht_view::getclientarea(bounds *b)
{
	*b=size;
}

void ht_view::getclientcarea(bounds *b)
{
	*b=csize;
}

ht_view *ht_view::getselected()
{
	return this;
}

STUB ht_view *ht_view::getfirstchild()
{
	return 0;
}

int ht_view::getnumber()
{
	return -1;
}

void ht_view::handlemsg(htmsg *msg)
{
	switch (msg->msg) {
		case htm_draw:
			redraw();
			break;
		case htm_dirtyview:
			dirtyview();
			if (msg->type & htmt_broadcast==0) clearmsg(msg);
			break;
	}
}

void ht_view::hidecursor()
{
	screen->hidecursor();
}

int ht_view::isalone(ht_view *view)
{
	return (view==this) && (countselectables()==1);
}

int ht_view::isviewdirty()
{
	return view_is_dirty;
}

int ht_view::lprint(int x, int y, int c, int l, char *text)
{
	if ((size.y+y>=vsize.y) && (size.y+y<vsize.y+vsize.h)) {
		if (size.x+x+l>vsize.x+vsize.w) l=vsize.x+vsize.w-size.x-x;
		if (size.x+x-vsize.x<0) {
			int kqx=-size.x-x+vsize.x;
			for (int i=0; i<kqx; i++) {
				if (!*text) return 0;
				text++;
				x++;
				l--;
			}
		}
		return (l>0) ? buf->lprint(size.x+x, size.y+y, c, l, text) : 0;
	}
	return 0;
}

int ht_view::lprintw(int x, int y, int c, int l, int *text)
{
	if ((size.y+y>=vsize.y) && (size.y+y<vsize.y+vsize.h)) {
		if (size.x+x+l>vsize.x+vsize.w) l=vsize.x+vsize.w-size.x-x;
		if (size.x+x-vsize.x<0) {
			int kqx=-size.x-x+vsize.x;
			for (int i=0; i<kqx; i++) {
				if (!*text) return 0;
				text++;
				x++;
				l--;
			}
		}
		return (l>0) ? buf->lprintw(size.x+x, size.y+y, c, l, text) : 0;
	}
	return 0;
}

void ht_view::move(int rx, int ry)
{
	size.x+=rx;
	size.y+=ry;
	buf->rmove(rx, ry);
	vsize=size;
	if (group) group->clipbounds(&vsize);
}

int ht_view::pointvisible(int x, int y)
{
	return ((x>=vsize.x) && (y>=vsize.y) && (x<vsize.x+vsize.w) && (y<vsize.y+vsize.h));
}

int ht_view::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return print(x, y, c, buf);
}

int ht_view::print(int x, int y, int c, char *text)
{
	if ((size.y+y>=vsize.y) && (size.y+y<vsize.y+vsize.h)) {
		int l=vsize.x+vsize.w-x-size.x;
		if (size.x+x-vsize.x<0) {
			int kqx=-size.x-x+vsize.x;
			for (int i=0; i<kqx; i++) {
				if (!*text) return 0;
				text++;
				x++;
				l--;
			}
		}
		return (l>0) ? buf->lprint(size.x+x, size.y+y, c, l, text) : 0;
	}
	return 0;
}

int ht_view::printw(int x, int y, int c, int *text)
{
	if ((size.y+y>=vsize.y) && (size.y+y<vsize.y+vsize.h)) {
		int l=vsize.x+vsize.w-x-size.x;
		if (size.x+x-vsize.x<0) {
			int kqx=-size.x-x+vsize.x;
			for (int i=0; i<kqx; i++) {
				if (!*text) return 0;
				text++;
				x++;
				l--;
			}
		}
		return (l>0) ? buf->lprintw(size.x+x, size.y+y, c, l, text) : 0;
	}
	return 0;
}

void ht_view::printchar(int x, int y, int c, int ch)
{
	if (pointvisible(size.x+x, size.y+y)) buf->printchar(size.x+x, size.y+y, c, ch);
}

void ht_view::receivefocus()
{
	dirtyview();
	focused=1;
}

void ht_view::redraw()
{
	if (options & VO_OWNBUFFER) {
		if (isviewdirty()) {
			draw();
			cleanview();
		}
		screen->drawbuffer((drawbuf*)buf, size.x, size.y, &vsize);
	} else {
		draw();
		cleanview();
	}
}

void ht_view::resize(int rw, int rh, int first_call)
{
	if (options & VO_RESIZE) {
		switch (growmode_x) {
			case GROW_CSLT:
				break;
			case GROW_CSRB:
				move(-rw, 0);
				break;
			case GROW_EDP:
				if (group) {
					bounds b;
					group->getclientarea(&b);
					size.w=(size.w*b.w+(b.w-rw)/2)/(b.w-rw);
				}
				break;
		}
		switch (growmode_y) {
			case GROW_CSLT:
				break;
			case GROW_CSRB:
				move(0, -rh);
				break;
			case GROW_EDP:
				if (group) {
					bounds b;
					group->getclientarea(&b);
					size.h=(size.h*b.h+(b.h-rh)/2)/(b.h-rh);
				}
				break;
		}
	}
	bounds b;
	getbounds(&b);
	if (group) group->clipbounds(&b);
	setvisualbounds(&b);
	dirtyview();
}

void ht_view::releasefocus()
{
	dirtyview();
	hidecursor();
	focused=0;
}

void ht_view::relocate_to(ht_view *view)
{
	bounds b;
	view->getbounds(&b);
	move(b.x, b.y);
}

int ht_view::select(ht_view *view)
{
	return (view==this);
}

void ht_view::selectfirst()
{
}

void ht_view::selectlast()
{
}

void ht_view::sendmsg(htmsg *msg)
{
	handlemsg(msg);
}

void ht_view::sendmsg(int msg, void *data)
{
	sendmsg(msg, (int)data);
}

void ht_view::sendmsg(int msg, int data)
{
	htmsg m;
	switch (msg) {
		case htm_nothing:
			return;
		case htm_draw:
		case htm_dirtyview:
			m.msg=msg;
			m.type=htmt_broadcast;
			m.data_int=data;
			break;
		default:
			m.type=htmt_nothing;
			m.msg=msg;
			m.data_int=data;
			break;
	}
	sendmsg(&m);
}

void ht_view::setbounds(bounds *b)
{
	size=*b;
	setvisualbounds(&size);
}

void ht_view::setvisualbounds(bounds *b)
{
	vsize=*b;
	if (options & VO_OWNBUFFER) {
		buf->setbounds(b);
	}
}

void ht_view::setcursor(int x, int y)
{
	if (pointvisible(size.x+x, size.y+y)) screen->setcursor(size.x+x, size.y+y); else
		screen->hidecursor();
}

void ht_view::setdata(void *buf)
{
	abstract();
}

void ht_view::setoptions(int _options)
{
	options=_options;
}

void ht_view::setpalette(char *name)
{
	ipalette *p=pal_grp->iget(name);
	if (p->palclass==pal_class) {
		pal=p->pal;
	}
}

void ht_view::unrelocate_to(ht_view *view)
{
	bounds b;
	view->getbounds(&b);
	b.x=-b.x;
	b.y=-b.y;
	move(b.x, b.y);
}

/*
 *	CLASS ht_group
 */

void ht_group::init(bounds *b, int options, char *desc)
{
	first=0;
	current=0;
	last=0;
	ht_view::init(b, options, desc);
	VIEW_DEBUG_NAME("ht_group");
	view_count=0;
	shared_data=0;
	growmode_x=GROW_EDP;
	growmode_y=GROW_EDP;
/* FIXME: debug */
	if (options & VO_OWNBUFFER) {
		WARN("class 'ht_group' should not have own buffer !");
	}
/**/
}

void ht_group::done()
{
	ht_view *a, *b;
	a=first;
	while (a) {
		b=a->next;
		a->done();
		delete a;
		a=b;
	}
	ht_view::done();
}

int ht_group::childcount()
{
	return view_count;
}

int ht_group::countselectables()
{
	int c=0;
	ht_view *v=first;
	while (v) {
		c+=v->countselectables();
		v=v->next;
	}
	return c;
}

STUB int ht_group::focus(ht_view *view)
{
	ht_view *v=first;
	while (v) {
		if (v->focus(view)) {
			releasefocus();
			current=v;
			putontop(v);
			receivefocus();
			return 1;
		}
		v=v->next;
	}
	return ht_view::focus(view);
}

int ht_group::focusnext()
{
	int i=current->browse_idx;
	if (options & VO_SELBOUND) return 1;
	if ((i==view_count-1) && !alone()) {
		return 0;
	}
	while (1) {
		i++;
		if (i>view_count-1) i=0;
		if (i==current->browse_idx) break;
		ht_view *v=get_by_browse_idx(i);
		if (v->options & VO_SELECTABLE) {
			v->selectfirst();
			focus(v);
			return 1;
		}
	}
	return 0;
}

int ht_group::focusprev()
{
	int i=current->browse_idx;
	if (options & VO_SELBOUND) return 1;
	if (!i && !alone()) {
		return 0;
	}
	while (1) {
		i--;
		if (i<0) i=view_count-1;
		if (i==current->browse_idx) break;
		ht_view *v=get_by_browse_idx(i);
		if (v->options & VO_SELECTABLE) {
			v->selectlast();
			focus(v);
			return 1;
		}
	}
	return 0;
}

ht_view *ht_group::get_by_browse_idx(int i)
{
	ht_view *v=first;
	while (v) {
		if (v->browse_idx==i) return v;
		v=v->next;
	}
	return 0;
}

ht_view *ht_group::getselected()
{
	return current->getselected();
}

ht_view *ht_group::getfirstchild()
{
	return first;
}

STUB void ht_group::handlemsg(htmsg *msg)
{
	if (msg->type==htmt_broadcast) {
		ht_view::handlemsg(msg);
		ht_view *v=first;
		while (v) {
			v->handlemsg(msg);
			v=v->next;
		}

		if (msg->msg==htm_draw) {
			msg->type=htmt_alwaysontop;
			v=first;
			while (v) {
				if (v->options & VO_ALWAYSONTOP) {
					v->handlemsg(msg);
				}
				v=v->next;
			}
			msg->type=htmt_broadcast;
		}
	} else if (msg->type==htmt_nothing) {
		int msgtype=msg->type;
		ht_view *v;

		msg->type=htmt_preprocess;
		v=first;
		while (v) {
			if (v->options & VO_PREPROCESS) {
				v->handlemsg(msg);
			}
			v=v->next;
		}

		msg->type=htmt_nothing;
		current->handlemsg(msg);

		msg->type=htmt_postprocess;
		v=first;
		while (v) {
			if (v->options & VO_POSTPROCESS) {
				v->handlemsg(msg);
			}
			v=v->next;
		}

		msg->type=msgtype;
		ht_view::handlemsg(msg);
	} else if (msg->type==htmt_preprocess) {
		ht_view *v;

		v=first;
		while (v) {
			if (v->options & VO_PREPROCESS) {
				v->handlemsg(msg);
			}
			v=v->next;
		}
	} else if (msg->type==htmt_postprocess) {
		ht_view *v;

		v=first;
		while (v) {
			if (v->options & VO_POSTPROCESS) {
				v->handlemsg(msg);
			}
			v=v->next;
		}
	}

	if (((msg->type==htmt_nothing) || (msg->type==htmt_broadcast)) && (msg->msg==htm_keypressed)) {
		switch (msg->data_int) {
			case K_BackTab: {
				if (focusprev()) {
					clearmsg(msg);
					dirtyview();
					return;
				}
				break;
			}
			case K_Tab: {
				if (focusnext()) {
					clearmsg(msg);
					dirtyview();
					return;
				}
				break;
			}
		}
	}
}

int ht_group::load(ht_stream *f)
{
	if (!ht_view::load(f)) return 0;
	dword id;
	while (1) {
		f->read(&id, 4);
		if (!id) break;
		config_object *c=find_config_object(id);
		if (!c) ERROR("unknown config_object (uid=%08x)", id);
		insert((ht_view*)c->loadconstructor(f));
	}
	return 1;
}

void ht_group::move(int rx, int ry)
{
	ht_view::move(rx, ry);
	ht_view *v=first;
	while (v) {
		v->move(rx, ry);
		v=v->next;
	}
}

void ht_group::insert(ht_view *view)
{
	if (current) current->releasefocus();
	if (view->options & VO_PREPROCESS) setoptions(options | VO_PREPROCESS);
	if (view->options & VO_POSTPROCESS) setoptions(options | VO_POSTPROCESS);
	bounds c;
	getclientarea(&c);
	view->move(c.x, c.y);
	if (last) last->next=view;
	view->prev=last;
	view->next=0;
	last=view;
	if (!first) first=view;
	view->group=this;
	view->browse_idx=view_count++;
	if ((!current) || ((current) && (!(current->options & VO_SELECTABLE)) && (view->options & VO_SELECTABLE))) {
		current=view;
	}
	if ((current) && (current->options & VO_SELECTABLE)) {
		if (focused) {
			focus(current);
		} else {
			select(current);
		}
	}
}

int ht_group::isalone(ht_view *view)
{
	ht_view *v=first;
	while (v) {
		if ((v!=view) && (v->countselectables())) return 0;
		v=v->next;
	}
	return 1;
}

int ht_group::isviewdirty()
{
	ht_view *v=first;
	while (v) {
		if (v->isviewdirty()) return 1;
		v=v->next;
	}
	return 0;
}

void ht_group::putontop(ht_view *view)
{
	if (view->next) {
		if (view->prev) view->prev->next=view->next; else first=view->next;
		view->next->prev=view->prev;
		view->prev=last;
		view->next=0;
		last->next=view;
		last=view;
	}
}

void ht_group::receivefocus()
{
	ht_view::receivefocus();
	current->receivefocus();
}

STUB void ht_group::redraw()
{
	ht_view::redraw();
/*	ht_view *v=first;		// _BADBUG_ !!!!!
	while (v) {
		v->redraw();
		v=v->next;
	}*/
}

void ht_group::releasefocus()
{
	ht_view::releasefocus();
	if (current) current->releasefocus();
}

void ht_group::remove(ht_view *view)
{
	bounds c;
	getclientarea(&c);
	view->move(-c.x, -c.y);
	if (view->prev) view->prev->next=view->next;
	if (view->next) view->next->prev=view->prev;
	if (first==view) first=first->next;
	if (last==view) last=last->prev;
	if (current==view) current=last;
}

void ht_group::resize(int rx, int ry, int first_call)
{
	if (first_call) {
		size.w+=rx;
		size.h+=ry;
		bounds b;
		getbounds(&b);
		if (group) group->clipbounds(&b);
		setvisualbounds(&b);
	} else {
		ht_view::resize(rx, ry, 0);
	}

	ht_view *v=first;
	while (v) {
		v->resize(rx, ry, 0);
		v=v->next;
	}
}

STUB int ht_group::select(ht_view *view)
{
	ht_view *v=first;
	while (v) {
		if (v->select(view)) {
			current=v;
			putontop(v);
			return 1;
		}
		v=v->next;
	}
	return ht_view::select(view);
}

void ht_group::selectfirst()
{
	for (int i=0; i<view_count; i++) {
		ht_view *v=first;
		while (v) {
			if ((v->browse_idx==i) && (v->options & VO_SELECTABLE)) {
				select(v);
				return;
			}
			v=v->next;
		}
	}
}

void ht_group::selectlast()
{
	for (int i=view_count-1; i>=0; i--) {
		ht_view *v=first;
		while (v) {
			if ((v->browse_idx==i) && (v->options & VO_SELECTABLE)) {
				select(v);
				return;
			}
			v=v->next;
		}
	}
}

void ht_group::setpalette(char *pal_name)
{
	ht_view *v=first;
	while (v) {
		v->setpalette(pal_name);
		v=v->next;
	}
	ht_view::setpalette(pal_name);
}

void ht_group::store(ht_stream *f)
{
	ht_view::store(f);
	ht_view *v=first;
	while (v) {
		v->store(f);
		v=v->next;
	}
	dword id=0;
	f->write(&id, 4);
}

/*
 *	CLASS ht_xgroup
 */

void ht_xgroup::init(bounds *b, int options, char *desc)
{
	ht_group::init(b, options, desc);
	VIEW_DEBUG_NAME("ht_xgroup");
	first=0;
	current=0;
	last=0;
}

void ht_xgroup::done()
{
	ht_group::done();
}

int ht_xgroup::countselectables()
{
	return current->countselectables();
}

STUB void ht_xgroup::handlemsg(htmsg *msg)
{
/* no broadcasts. */
	if (msg->msg==htm_complete_init) return;
	if (current) current->handlemsg(msg);
	ht_view::handlemsg(msg);
}

int ht_xgroup::isalone(ht_view *view)
{
	if (group) return group->isalone(this);
	return 0;
}

void ht_xgroup::redraw()
{
	ht_view::redraw();
/* no broadcasts. */
	if (current) current->redraw();
}

void ht_xgroup::selectfirst()
{
	current->selectfirst();
}

void ht_xgroup::selectlast()
{
	current->selectlast();
}

/*
 *	CLASS ht_scrollbar
 */

void ht_scrollbar::init(bounds *b, int *_color)
{
	ht_view::init(b, 0, 0);
	VIEW_DEBUG_NAME("ht_scrollbar");
	color=_color;
	ppos=0;
	enable();	// enabled by default
}

void ht_scrollbar::done()
{
	ht_view::done();
}

void ht_scrollbar::enable()
{
	enable_buffering();
	ht_view::enable();
	dirtyview();
}

void ht_scrollbar::disable()
{
	disable_buffering();
	ht_view::disable();
	dirtyview();
}

void ht_scrollbar::draw()
{
	if (enabled) {
		printchar(0, 0, *color, CHAR_ARROWBIG_UP);
		fill(0, 1, size.w, size.h-2, *color, CHAR_FILLED_M);
		printchar(0, size.h-1, *color, CHAR_ARROWBIG_DOWN);
		printchar(0, ppos+1, *color, CHAR_QUAD_SMALL);
	}
}

void ht_scrollbar::handlemsg(htmsg *msg)
{
	if ((msg->msg!=htm_draw) || (msg->type==htmt_alwaysontop))
		ht_view::handlemsg(msg);
}

void ht_scrollbar::setpos(int percent)
{
	ppos=((size.h-3)*percent+50)/100;
	dirtyview();
}

/*
 *	CLASS ht_frame
 */

void ht_frame::init(bounds *b, char *desc, int _number, int _style)
{
	ht_group::init(b, VO_RESIZE, desc);
	VIEW_DEBUG_NAME("ht_frame");
	growmode_x=GROW_EDP;
	growmode_y=GROW_EDP;
	number=_number;
	style=_style;
	framestate=FST_UNFOCUSED;
}

void ht_frame::done()
{
	ht_group::done();
}

void ht_frame::draw()
{
	int c=getcurcol_normal();
/* "...Ŀ" */
	printchar(0, 0, c, CHAR_CORNERUL);
	for (int i=1; i<size.w-1; i++) printchar(i, 0, c, CHAR_LINEH);
	printchar(0+size.w-1, 0, c, CHAR_CORNERUR);
/* "..." */
	printchar(0, size.h-1, c, CHAR_CORNERLL);
	for (int i=1; i<size.w-1; i++) printchar(i, size.h-1, c, CHAR_LINEH);
	printchar(size.w-1, size.h-1, c, CHAR_CORNERLR);
/* "", "" */
	for (int i=1; i<size.h-1; i++) {
		printchar(0, i, c, CHAR_LINEV);
		printchar(size.w-1, i, c, CHAR_LINEV);
	}
/* "[x]" */
	if (style & FS_KILLER) {
		print(2, 0, c, "[ ]");
		printchar(3, 0, getcurcol_killer(), CHAR_QUAD_SMALL);
	}
/* e.g. "1" */
	if (number>0) {
		int l=number;
		int i=0;
		do {
			l=l/10;
			i++;
		} while (l);
		printf(size.w-4-i, 0, c, "%d", number);
	}
/* <title> */
	if ((desc) && (style & FS_TITLE)) {
		int l=strlen(desc), o=0;
		if (l>size.w-15) {
			o=l-size.w+15;
			l-=o;
			printf((size.w-l-5)/2, 0, c, " ...%s ", desc+o+3);
		} else {
			printf((size.w-l-2)/2, 0, c, " %s ", desc);
		}
	}
}

void ht_frame::getclientarea(bounds *b)
{
	*b=size;
	b->x++;
	b->y++;
	b->w-=2;
	b->h-=2;
}

void ht_frame::getclientcarea(bounds *b)
{
	*b=csize;
	b->x++;
	b->y++;
	b->w-=2;
	b->h-=2;
}

int ht_frame::getcurcol_normal()
{
	switch (framestate) {
		case FST_FOCUSED:
			return windialog_color_frame_focused;
		case FST_UNFOCUSED:
			return windialog_color_frame_unfocused;
		case FST_MOVE:
			return windialog_color_frame_moved;
	}
	return 0;
}

int ht_frame::getcurcol_killer()
{
	return windialog_color_frame_killer;
}

void ht_frame::handlemsg(htmsg *msg)
{
	ht_group::handlemsg(msg);
}

void ht_frame::receivefocus()
{
	ht_group::receivefocus();
	setframestate(FST_FOCUSED);
}

void ht_frame::releasefocus()
{
	ht_group::releasefocus();
	setframestate(FST_UNFOCUSED);
}

void ht_frame::setframestate(int _framestate)
{
	framestate=_framestate;
	dirtyview();
}

/*
 *	CLASS ht_window
 */

void ht_window::init(bounds *b, char *desc, int _number, int framestyle)
{
	ht_group::init(b, VO_SELECTABLE | VO_SELBOUND | VO_BROWSABLE | VO_MOVE | VO_RESIZE, desc);
	number=_number;
	scrollbar=0;
	bounds c=*b;
	c.x=0;
	c.y=0;
	frame=new ht_frame();
	frame->init(&c, desc, _number, framestyle);
	ht_group::insert(frame);
	action_state=WAC_NORMAL;
}

void ht_window::done()
{
	ht_group::done();
}

void ht_window::draw()
{
	clear(windialog_color_body);
}

int ht_window::getnumber()
{
	return number;
}

void ht_window::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		if (options&(VO_MOVE | VO_RESIZE)) {
			if (action_state==WAC_MOVE) {
				if (options & VO_MOVE) {
					switch (msg->data_int) {
						case K_Up:
							if (size.y>group->size.y) move(0, -1);
							break;
						case K_Down:
							if (size.y<group->size.y+group->size.h-1) move(0, 1);
							break;
						case K_Left:
							if (size.x+size.w>group->size.x+1) move(-1, 0);
							break;
						case K_Right:
							if (size.x<group->size.x+group->size.w-1) move(1, 0);
							break;
						case K_Control_Up:
							if (size.y>group->size.y+5-1) move(0, -5); else
								move(0, group->size.y-size.y);
							break;
						case K_Control_Down:
							if (size.y<group->size.y+group->size.h-5) move(0, 5); else
								move(0, group->size.y+group->size.h-size.y-1);
							break;
						case K_Control_Left:
							if (size.x+size.w>group->size.x+5) move(-5, 0); else
								move(-(size.x+size.w)+group->size.x+1, 0);
							break;
						case K_Control_Right:
							if (size.x<group->size.x+group->size.w-5) move(5, 0); else
								move(group->size.x+group->size.w-size.x-1, 0);
							break;
					}
				}
				if (options & VO_RESIZE) {
					switch (msg->data_int) {
						case K_Shift_Up:
							if (size.h>MIN(7, csize.h)) resize(0, -1, 1);
							break;
						case K_Shift_Down:
							if (size.h<group->size.h) resize(0, 1, 1);
							break;
						case K_Shift_Left:
							if ((size.x+size.w>1) && (size.w>MIN(25, csize.w))) resize(-1, 0, 1);
							break;
						case K_Shift_Right:
							if (size.w<group->size.w) resize(1, 0, 1);
							break;
					}
				}
				switch (msg->data_int) {
					case K_Escape:
					case K_Return:
					case K_Control_F5:
						action_state=WAC_NORMAL;
						frame->setframestate(FST_FOCUSED);
						break;
				}
				clearmsg(msg);
				app->sendmsg(htm_dirtyview, 0);
				return;
			} else {
				switch (msg->data_int) {
					case K_Control_F5:
						action_state=WAC_MOVE;
						frame->setframestate(FST_MOVE);
						clearmsg(msg);
						dirtyview();
						return;
				}
			}
		}
	}
	ht_group::handlemsg(msg);
}

void ht_window::insert(ht_view *view)
{
	frame->insert(view);
}

void ht_window::setframe(ht_frame *_frame)
{
/* FIXME: should insert views sorted */
	ht_view *v=frame->first;
	while (v) {
		frame->remove(v);
		_frame->insert(v);
		v=v->next;
	}
	ht_group::remove(frame);
	frame->done();
	delete frame;
	ht_group::insert(_frame);
	frame=_frame;
}

void ht_window::setscrollbar(ht_scrollbar *_scrollbar)
{
	scrollbar=_scrollbar;
	insert(scrollbar);
	putontop(scrollbar);
}

