/*
 *	HT Editor
 *	htobj.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTOBJ_H__
#define __HTOBJ_H__

class ht_view;
class ht_group;

#include "common.h"
#include "htcurses.h"
#include "htpal.h"

/* messages */
#define htm_nothing 		0
#define htm_command			1
#define htm_draw 			2
#define htm_resize 			3
#define htm_keypressed 		4
#define htm_kill			5
#define htm_complete_init	6
#define htm_funcexec		7
#define htm_funcname		8
#define htm_funcname_return	9
#define htm_button_pressed	10
#define htm_dirtyview		11
#define htm_vs_restore		12
#define htm_edit_mode		13
#define htm_view_mode		14

/* message types */
#define htmt_nothing		0
#define htmt_broadcast		1
#define htmt_preprocess		2
#define htmt_postprocess		3
#define htmt_alwaysontop		4

struct htmsg {
	int msg;
	int type;
	union {
		int data_int;
		void *data_ptr;
		char *data_str;
	};
};

/*
 *	CLASS ht_view
 */

/* options */
#define VO_OWNBUFFER		1
#define VO_BROWSABLE		2
#define VO_SELECTABLE		4
#define VO_SELBOUND			8
#define VO_PREPROCESS		16
#define VO_POSTPROCESS		32
#define VO_MOVE			64
#define VO_RESIZE			128
#define VO_ALWAYSONTOP		256

/* grow modes */
#define GROW_CSLT			1	/* corner is static to left/top edge of parent rect */
#define GROW_CSRB			2    /* corner is static to right/bottom edge of parent rect */
#define GROW_EDP			3    /* edge grows dynamically with corresponding parent edge */

class ht_group;

#define VIEW_DEBUG_NAME(name)	ht_view::view_debug_name=name;

class ht_view: public object {
protected:
			int focused;
			int view_is_dirty;

			void cleanview();
public:
	bounds size;
	bounds vsize;	/* visual size */
	bounds csize;	/* creation size */
	palette *pal;
	int pal_class;
	char *pal_name;

	int growmode_x, growmode_y;

	int enabled;
	genericdrawbuf *buf;
	int options;
	char *desc;
	ht_group	*group;
	int browse_idx;
	ht_view 	*prev, *next;

/*debug:*/char *view_debug_name;

			void	init(bounds *b, int options, char *desc);
	virtual	void done();
/* new */
	virtual	int alone();
	virtual	int childcount();
			void clear(int color);
			void clearmsg(htmsg *msg);
			void clipbounds(bounds *b);
	virtual	int countselectables();
	virtual	int datasize();
	virtual	void defaultpalette();
			void dirtyview();
	virtual	void disable();
			void disable_buffering();
STUB	virtual 	void draw();
	virtual	void enable();
			void enable_buffering();
			void fill(int x, int y, int w, int h, int c, int chr);
	virtual	int focus(ht_view *view);
			void getbounds(bounds *b);
	virtual	void getclientarea(bounds *b);
	virtual	void getclientcarea(bounds *b);
	virtual	ht_view *getselected();
	virtual 	ht_view *getfirstchild();
	virtual	void getdata(void *buf);
	virtual	int getnumber();
	virtual 	void handlemsg(htmsg *msg);
			void hidecursor();
			int isviewdirty();
	virtual	int isalone(ht_view *view);
			int lprint(int x, int y, int c, int l, char *text);
			int lprintw(int x, int y, int c, int l, int *text);
	virtual	void move(int x, int y);
			int pointvisible(int x, int y);
			int printf(int x, int y, int c, char *format, ...);
			int print(int x, int y, int c, char *text);
			int printw(int x, int y, int c, int *text);
			void printchar(int x, int y, int c, int ch);
	virtual	void receivefocus();
	virtual	void redraw();
			void relocate_to(ht_view *view);
	virtual	void resize(int rw, int rh, int first_call);
	virtual 	void releasefocus();
	virtual	int select(ht_view *view);
	virtual	void selectfirst();
	virtual	void selectlast();
			void sendmsg(htmsg *msg);
			void sendmsg(int msg, void *data);
			void sendmsg(int msg, int data);
			void setbounds(bounds *b);
			void setvisualbounds(bounds *b);
			void setcursor(int x, int y);
	virtual	void setdata(void *buf);
			void setoptions(int options);
	virtual	void setpalette(char *pal_name);
			void unrelocate_to(ht_view *view);
};

/*
 *	CLASS ht_group
 */

class ht_group: public ht_view {
protected:
	int 		view_count;
public:
	ht_view 	*first, *current, *last;
	void		*shared_data;

			void	init(bounds *b, int options, char *desc);
	virtual	void done();
/* overwritten */
	virtual	int childcount();
	virtual	int countselectables();
	virtual	int focus(ht_view *view);
	virtual	ht_view *getselected();
	virtual 	ht_view *getfirstchild();
STUB	virtual	void handlemsg(htmsg *msg);
	virtual	int isalone(ht_view *view);
			int isviewdirty();
			int load(ht_stream *f);
	virtual	void move(int x, int y);
			void putontop(ht_view *view);
	virtual 	void receivefocus();
	virtual	void redraw();
	virtual	void resize(int rw, int rh, int first_call);
	virtual 	void releasefocus();
	virtual	int select(ht_view *view);
	virtual	void selectfirst();
	virtual	void selectlast();
	virtual	void setpalette(char *pal_name);
	virtual	void store(ht_stream *f);
/* new */
			void remove(ht_view *view);
	virtual	void insert(ht_view *view);
			int focusnext();
			int focusprev();
			ht_view *get_by_browse_idx(int i);
};

/*
 *	CLASS ht_xgroup
 */

class ht_xgroup: public ht_group {
public:
			void	init(bounds *b, int options, char *desc);
	virtual	void done();
/* overwritten */
	virtual	int countselectables();
STUB	virtual	void handlemsg(htmsg *msg);
	virtual	int isalone(ht_view *view);
	virtual	void redraw();
	virtual	void selectfirst();
	virtual	void selectlast();
};

/*
 *	CLASS ht_scrollbar
 */

class ht_scrollbar: public ht_view {
protected:
	int ppos;
	int *color;
public:
			void	init(bounds *b, int *color);
	virtual	void done();
/* overridden */
	virtual	void enable();
	virtual	void disable();
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	void setpos(int percent);
};

/*
 *	CLASS ht_frame
 */

#define FS_KILLER		1
#define FS_TITLE 		2

#define FST_FOCUSED		0
#define FST_UNFOCUSED	1
#define FST_MOVE      	2

class ht_frame: public ht_group {
protected:
	int number;
	int style;
	int framestate;

/* new */
	virtual	int getcurcol_normal();
	virtual	int getcurcol_killer();
public:
			void	init(bounds *b, char *desc, int number, int style);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	void getclientarea(bounds *b);
	virtual	void getclientcarea(bounds *b);
	virtual	void handlemsg(htmsg *msg);
	virtual	void receivefocus();
	virtual 	void releasefocus();
/* new */
			void setframestate(int framestate);
};

/*
 *	CLASS ht_window
 */

#define WAC_NORMAL	0
#define WAC_MOVE	1

class ht_window: public ht_group {
protected:
	ht_frame *frame;
	ht_scrollbar *scrollbar;
	int number;

	int action_state;
public:
			void	init(bounds *b, char *desc, int number, int framestyle);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	int getnumber();
	virtual	void handlemsg(htmsg *msg);
	virtual	void insert(ht_view *view);
/* new */
			void setframe(ht_frame *frame);
			void setscrollbar(ht_scrollbar *scrollbar);
};

#endif /* !__HTOBJ_H__ */
