/*
 *	HT Editor
 *	htpe.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpehead.h"
#include "htpeexp.h"
#include "htpeimp.h"
#include "htpedimp.h"
#include "htpeimg.h"
#include "htperes.h"
#include "htstream.h"

#include <stdlib.h>
#include <string.h>

viewer_if *htpe_ifs[] = {
	&htpeheader_if,
	&htpeexports_if,
	&htpeimports_if,
	&htpedelayimports_if,
	&htperesources_if,
	&htpeimage_if,
	0
};

ht_view *htpe_init(bounds *b, ht_stream *file, ht_group *group)
{
	dword pemagic;
	STREAMOFS h=get_newexe_header_ofs(file);
	file->seek(h);
	file->read(&pemagic, 4);
	if (pemagic!=PE_MAGIC) return 0;

	ht_pe *g=new ht_pe();
	g->init(b, file, htpe_ifs, h);
	return g;
}

viewer_if htpe_if = {
	htpe_init,
	0
};

void ht_pe::init(bounds *b, ht_stream *file, viewer_if **ifs, STREAMOFS h)
{
	ht_viewergroup::init(b, VO_BROWSABLE | VO_SELECTABLE | VO_RESIZE, DESC_PE, file, 0);
	VIEW_DEBUG_NAME("ht_pe");

	LOG("%s: PE: found header at %08x", file->get_name(), h);
	pe_shared=new ht_pe_shared_data;
	pe_shared->hdr_ofs=h;
	pe_shared->v_image=0;
	pe_shared->v_dimports=0;
	pe_shared->v_imports=0;
	pe_shared->v_exports=0;
	pe_shared->v_header=0;

/* headers */
	file->seek(h+4);
	file->read(&pe_shared->coffheader, sizeof pe_shared->coffheader);
	file->read(&pe_shared->opt_magic, sizeof pe_shared->opt_magic);
	file->seek(h+4+sizeof pe_shared->coffheader);
	switch (pe_shared->opt_magic) {
		case COFF_OPTMAGIC_PE32:
			file->read(&pe_shared->pe32.header, sizeof pe_shared->pe32.header);
			file->read(&pe_shared->pe32.header_nt, sizeof pe_shared->pe32.header_nt);
			break;
	}

/* read section headers */
	int os=0;

	/* 1. get section count */
	file->seek(h+6);
	pe_shared->sections.section_count=0;
	file->read(&pe_shared->sections.section_count, 2);
	/* 2. get size of optional header */
	file->seek(h+20);
	file->read(&os, 2);
	/* 3. read sections */
	file->seek(h+os+24);
	pe_shared->sections.sections=(COFF_SECTION_HEADER*)malloc(pe_shared->sections.section_count * sizeof *pe_shared->sections.sections);
	file->read(pe_shared->sections.sections, pe_shared->sections.section_count*sizeof *pe_shared->sections.sections);

	shared_data=pe_shared;

	ht_viewergroup::initall(ifs);
}

void ht_pe::done()
{
	ht_viewergroup::done();
}

/*
 *	rva conversion routines
 */

int pe_rva_to_section(pe_section_headers *section_headers, ADDR rva, int *section)
{
	COFF_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+MAX(s->data_size, s->data_vsize))) {
			*section=i;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_rva_to_ofs(pe_section_headers *section_headers, ADDR rva, dword *ofs)
{
	COFF_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+s->data_size)) {
			*ofs=rva-s->data_address+s->data_offset;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_rva_is_valid(pe_section_headers *section_headers, ADDR rva)
{
	COFF_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+MAX(s->data_size, s->data_vsize))) {
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_rva_is_physical(pe_section_headers *section_headers, ADDR rva)
{
	COFF_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((rva>=s->data_address) &&
		(rva<s->data_address+s->data_size)) {
			return 1;
		}
		s++;
	}
	return 0;
}

/*
 *	ofs conversion routines
 */

int pe_ofs_to_rva(pe_section_headers *section_headers, dword ofs, ADDR *rva)
{
	COFF_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((ofs>=s->data_offset) &&
		(ofs<s->data_offset+s->data_size)) {
			*rva=ofs-s->data_offset+s->data_address;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_ofs_to_section(pe_section_headers *section_headers, dword ofs, ADDR *section)
{
	COFF_SECTION_HEADER *s=section_headers->sections;
	for (int i=0; i<section_headers->section_count; i++) {
		if ((ofs>=s->data_offset) &&
		(ofs<s->data_offset+s->data_size)) {
			*section=i;
			return 1;
		}
		s++;
	}
	return 0;
}

int pe_ofs_to_rva_and_section(pe_section_headers *section_headers, dword ofs, ADDR *rva, ADDR *section)
{
	int r=pe_ofs_to_rva(section_headers, ofs, rva);
	if (r) {
		r=pe_ofs_to_section(section_headers, ofs, section);
	}
	return r;
}

int pe_ofs_is_valid(pe_section_headers *section_headers, dword ofs)
{
	ADDR rva;
	return pe_ofs_to_rva(section_headers, ofs, &rva);
}

