/* 
 *	HT Editor
 *	htpedimp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htcolor.h"
#include "htiobox.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpedimp.h"
#include "htstream.h"
#include "httag.h"
#include "htstring.h"
#include "viewers.h"

#include <stdlib.h>

#define PE_DIMPORT_DESC_LINES 1	/* important ! set to the number
							   of descriptive lines in the beginning
							   of the (hts_)mask */

ht_view *htpedelayimports_init(bounds *b, ht_stream *file, ht_group *group)
{
	int h0=new_timer();
	start_timer(h0);

	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;

	if (pe_shared->opt_magic!=COFF_OPTMAGIC_PE32) return 0;

	STREAMOFS h=pe_shared->hdr_ofs;
	dword sec_rva, sec_size;
	file->seek(h+0x78+PE_DIRECTORY_ENTRY_DELAY_IMPORT*8);
	file->read(&sec_rva, 4);
	file->read(&sec_size, 4);
	if (!sec_rva || !sec_size) return 0;

	dword base=pe_shared->pe32.header_nt.image_base;
	char iline[1024];	/* FIXME: possible buffer overflow */

	int dll_count=0;
	int function_count=0;
	ht_cached_stream *cached_file=0;

	ht_group *g;
	bounds c;
	ht_statictext *head;

	c=*b;
	g=new ht_group();
	g->init(&c, VO_RESIZE, "pe/delay-imports-g");

	c.y++;
	c.h--;
	ht_pe_dimport_viewer *v=new ht_pe_dimport_viewer();
	v->init(&c, DESC_PE_DIMPORTS, file, 0);
	hts_pe_dimport_mask *m=new hts_pe_dimport_mask();
	m->init(file);

	c.y--;
	c.h=1;
/* get delay import directory offset */
	/* 1. get import directory rva */
	dword iofs;
	dword irva=pe_shared->pe32.header_nt.directory[PE_DIRECTORY_ENTRY_DELAY_IMPORT].address;
//	dword isize=pe_shared->pe32.header_nt.directory[PE_DIRECTORY_ENTRY_DELAY_IMPORT].size;
	/* 2. transform it into an offset */
	if (!pe_rva_to_ofs(&pe_shared->sections, irva, &iofs)) goto pe_read_error;
	LOG("%s: PE: reading delay-import directory at offset %08x, rva %08x, size %08x...", file->get_name(), iofs, irva, pe_shared->pe32.header_nt.directory[PE_DIRECTORY_ENTRY_DELAY_IMPORT].size);

//	cached_file=new ht_cached_stream(file, 128*1024, CACHED_STREAM_CACHE_READS, 2048);
//     file=cached_file;		/* we dont need to care about "file" */

	char info[128];
	sprintf(info, "* PE delay-import directory at offset %08lx", iofs);
	m->add_mask(info);
/*** read delay import directory ***/
	PE_DELAY_IMPORT_DESCRIPTOR dimport;
	int dll_index;
	dll_index=0;
	dword f;
	f=0;

	while (1) {
		file->seek(iofs);
		file->read(&dimport, sizeof dimport);
		if (!dimport.name) break;
		dword dllname_ofs;
		if (!pe_rva_to_ofs(&pe_shared->sections, dimport.name-base, &dllname_ofs)) goto pe_read_error;
		file->seek(dllname_ofs);
		char *dllname=getstrz(file);
		ht_pe_dimport_library *lib=new ht_pe_dimport_library(dllname);
		pe_shared->dimports.libs.insert(lib);
		dll_count++;

		dword ntofs, nva;
		dword atofs, ava;
		if (!pe_rva_to_ofs(&pe_shared->sections, dimport.delay_int-base, &ntofs)) goto pe_read_error;
		if (!pe_rva_to_ofs(&pe_shared->sections, dimport.delay_iat-base, &atofs)) goto pe_read_error;
		while (1) {
			ht_pe_dimport_function *func;
			file->seek(ntofs);
			file->read(&nva, 4);
			if (!nva) break;
			function_count++;
			file->seek(atofs);
			file->read(&ava, 4);
			if (nva & 0x80000000) {
				/* import by ordinal */
				char ord[5];
				sprintf(ord, "%04lx", nva&0xffff);
				char *im=iline;
				im+=sprintf(iline, "%-14s %08lx ", dllname, ava);
				tag_make_ref(im, ava, 0, ord);
				m->add_mask(iline);
				func=new ht_pe_dimport_function(dll_index, ava, nva&0xffff);
			} else {
				dword nofs;
				/* import by name */
				if (!pe_rva_to_ofs(&pe_shared->sections, nva-base, &nofs)) goto pe_read_error;
				int hint=0;
				file->seek(nofs);
				file->read(&hint, 2);
				char *name=getstrz(file);
/* FIXME: what about the hint ? */
				char *im=iline;
				im+=sprintf(iline, "%-14s %08lx ", dllname, ava);
				tag_make_ref(im, ava, 0, name);
				m->add_mask(iline);
				func=new ht_pe_dimport_function(dll_index, ava, name, hint);
				delete name;
			}
			pe_shared->dimports.funcs.insert(func);
			ntofs+=4;
			atofs+=4;
			f++;
		}

		delete dllname;
		iofs+=sizeof dimport;
		dll_index++;
	}

	stop_timer(h0);
	LOG("%s: PE: %d ticks (%d msec) to read delay-imports", file->get_name(), get_timer_tick(h0), get_timer_msec(h0));
	delete_timer(h0);

//     delete cached_file;

	if (!function_count) m->add_mask("no delay-imported functions.");
	v->insertsub(m);

	sprintf(iline, "Library        FT_VA    Name/Ordinal    %d delay-imports from %d libraries", function_count, dll_count);
	head=new ht_statictext();
	head->init(&c, iline, align_left);

	g->insert(head);
	g->insert(v);

	g->setpalette(window_palette_str);

	pe_shared->v_dimports=v;

	return g;
pe_read_error:
	errorbox("%s: PE delay-import directory seems to be corrupted.", file->get_name());
	if (cached_file) delete cached_file;
	v->done();
	delete v;
	m->done();
	delete m;
	g->done();
	delete g;
	return 0;
}

viewer_if htpedelayimports_if = {
	htpedelayimports_init,
	0
};

/*
 *	CLASS ht_pe_dimport_viewer
 */

void ht_pe_dimport_viewer::init(bounds *b, char *desc, ht_stream *file, int caps)
{
	ht_uviewer::init(b, desc, file, caps);
	VIEW_DEBUG_NAME("ht_pe_dimport_viewer");
}

void ht_pe_dimport_viewer::done()
{
	ht_uviewer::done();
}

char *ht_pe_dimport_viewer::func(int i, int getname)
{
	switch (i) {
		case 3:
			if (getname) return "sortva"; else {
				((hts_pe_dimport_mask*)cursor_sub)->sortbyva();
				dirtyview();
			}
			return 0;
		case 4:
			if (getname) return "sortname"; else {
				((hts_pe_dimport_mask*)cursor_sub)->sortbyname();
				dirtyview();
			}
			return 0;
		case 5:
			if (getname) return "unsort"; else {
				((hts_pe_dimport_mask*)cursor_sub)->unsort();
				dirtyview();
			}
			return 0;
	}
	return ht_uviewer::func(i, getname);
}

int ht_pe_dimport_viewer::ref_sel(ID id_low, ID id_high)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;
	pe_shared->v_image->jump(id_low, this);
	app->focus(pe_shared->v_image);
	return 1;
}

/*
 *	CLASS ht_pe_dimport_library
 */

ht_pe_dimport_library::ht_pe_dimport_library()
{
}

ht_pe_dimport_library::ht_pe_dimport_library(char *name)
{
	this->name=strdup(name);
}

ht_pe_dimport_library::~ht_pe_dimport_library()
{
}

/*
 *	CLASS ht_pe_dimport_function
 */

ht_pe_dimport_function::ht_pe_dimport_function()
{
	dll=0;
	name.name=0;
	name.hint=0;
	ordinal=0;
	function_address=0;
}

ht_pe_dimport_function::ht_pe_dimport_function(int _dll, int _function_address, int _ordinal)
{
	dll=_dll;
	ordinal=_ordinal;
	function_address=_function_address;
	byname=0;
}

ht_pe_dimport_function::ht_pe_dimport_function(int _dll, int _function_address, char *_name, int _hint)
{
	dll=_dll;
	name.name=strdup(_name);
	name.hint=_hint;
	function_address=_function_address;
	byname=1;
}

ht_pe_dimport_function::~ht_pe_dimport_function()
{
	if ((byname) && (name.name)) delete name.name;
}

/*
 *	CLASS hts_pe_dimport_mask
 */

void hts_pe_dimport_mask::init(ht_stream *file)
{
	hts_mask::init(file);
	sort_va=0;
	sort_name=0;
	sort_path=0;
	sort=0;
}

void hts_pe_dimport_mask::done()
{
	if (sort_va) delete sort_va;
	if (sort_name) delete sort_name;
	hts_mask::done();
}

int hts_pe_dimport_mask::getline(char *line, dword id1, dword id2)
{
	if ((sort) && (sort_path) && (id1>=PE_DIMPORT_DESC_LINES)) {
		return hts_mask::getline(line, sort_path[id1], id2);
	} else {
		return hts_mask::getline(line, id1, id2);
	}
}

void hts_pe_dimport_mask::sortbyva()
{
	if (!sort_va) {
		int c=masks.count();
		byvalue *sv=(byvalue*)malloc(sizeof *sv * c);
		for (int i=PE_DIMPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data *)masks.get(i);
			char *s=d->tagstring;
			dword v=hexd(s+15);
			sv[i].key=v;
			sv[i].value=i;
		}
		qsortbyvalue(sv, PE_DIMPORT_DESC_LINES, c-1);
		sort_va=(dword*)malloc(sizeof *sort_va * c);
		for (int i=PE_DIMPORT_DESC_LINES; i<c; i++) {
			sort_va[i]=sv[i].value;
		}
		delete sv;
	}
	sort=1;
	sort_path=sort_va;
}

void hts_pe_dimport_mask::sortbyname()
{
	if (!sort_name) {
		int c=masks.count();
		byname *sn=(byname*)malloc(sizeof *sn * c);
		for (int i=PE_DIMPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data *)masks.get(i);
			char *s=d->tagstring;
			char *t=tag_get_selectable_tag(s, 1, 0);
			int q=tag_get_seltextlen(t);

			sn[i].key=tag_get_seltext(t);;
			sn[i].keylen=q;
			sn[i].value=i;
		}
		qsortbyname(sn, PE_DIMPORT_DESC_LINES, c-1);
		sort_name=(dword*)malloc(sizeof *sort_name * c);
		for (int i=PE_DIMPORT_DESC_LINES; i<c; i++) {
			sort_name[i]=sn[i].value;
		}
		delete sn;
	}
	sort=1;
	sort_path=sort_name;
}

void hts_pe_dimport_mask::unsort()
{
	sort=0;
}


