/* 
 *	HT Editor
 *	htpeexp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htcolor.h"
#include "htctrl.h"
#include "htdata.h"
#include "htdebug.h"
#include "htiobox.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpeexp.h"
#include "htstream.h"
#include "htstring.h"
#include "httag.h"
#include "tools.h"
#include "viewers.h"

#include <stdlib.h>
#include <string.h>

#define PE_EXPORT_DESC_LINES 1	/* important ! set to the number
							   of descriptive lines in the beginning
							   of the (hts_)mask */

ht_view *htpeexports_init(bounds *b, ht_stream *file, ht_group *group)
{
	int h0=new_timer();
	start_timer(h0);

	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;

	if (pe_shared->opt_magic!=COFF_OPTMAGIC_PE32) return 0;

	STREAMOFS h=pe_shared->hdr_ofs;
	dword sec_rva, sec_size;
	file->seek(h+0x78+PE_DIRECTORY_ENTRY_EXPORT*8);
	file->read(&sec_rva, 4);
	file->read(&sec_size, 4);
	if (!sec_rva || !sec_size) return 0;

	ht_group *g;
	char eline[1024];   /* FIXME: possible buffer overflow */
	char *elinep;
	dword *efunct=0, *enamet=0;
	word *eordt=0;
	bounds c;
	ht_statictext *head;
	char *filename=file->get_name();

	c=*b;
	g=new ht_group();
	g->init(&c, VO_RESIZE, "pe/exports-g");

	c.y++;
	c.h--;
	ht_pe_export_viewer *v=new ht_pe_export_viewer();
	v->init(&c, DESC_PE_EXPORTS, file, 0);
	hts_pe_export_mask *m=new hts_pe_export_mask();
	m->init(file);

	c.y--;
	c.h=1;

/* get export directory offset */
	/* 1. get export directory rva */
	dword eofs;
	dword erva=pe_shared->pe32.header_nt.directory[PE_DIRECTORY_ENTRY_EXPORT].address;
	dword esize=pe_shared->pe32.header_nt.directory[PE_DIRECTORY_ENTRY_EXPORT].size;
	/* 2. transform it into an offset */
	if (!pe_rva_to_ofs(&pe_shared->sections, erva, &eofs)) goto pe_read_error;
	LOG("%s: PE: reading export directory at offset %08x, rva %08x, size %08x...", filename, eofs, erva, esize);

	char info[128];
	sprintf(info, "* PE export directory at offset %08lx", eofs);
	m->add_mask(info);

/* make a memfile out of this section */
	ht_mem_stream *efile;
	char *esectionbuf;
	efile=new ht_mem_stream();
	efile->init(eofs, esize);
	esectionbuf=(char*)malloc(esize);
	file->seek(eofs);
	file->read(esectionbuf, esize);
	efile->write(esectionbuf, esize);
	file=efile;
	delete esectionbuf;
/* read export directory header */
	PE_EXPORT_DIRECTORY edir;
	file->seek(eofs);
	file->read(&edir, sizeof edir);
/* get export filename */
	dword ename_ofs;
	char *ename;
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.name_address, &ename_ofs)) goto pe_read_error;
	file->seek(ename_ofs);
	ename=getstrz(file);
/* ...and add it to the mask */
	sprintf(eline, "Ord  VA       Name      %s exports %ld functions", ename, edir.function_count);
	head=new ht_statictext();
	head->init(&c, eline, align_left);
	delete ename;
/* read in function entrypoint table */
	dword efunct_ofs;
	efunct=(dword*)malloc(edir.function_count*sizeof *efunct);
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.function_table_address, &efunct_ofs)) goto pe_read_error;
	file->seek(efunct_ofs);
	file->read(efunct, edir.function_count*sizeof *efunct);
/* read in name address table */
	dword enamet_ofs;
	enamet=(dword*)malloc(edir.name_count*sizeof *efunct);
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.name_table_address, &enamet_ofs)) goto pe_read_error;
	file->seek(enamet_ofs);
	file->read(enamet, edir.name_count*sizeof *efunct);
/* read in ordinal table */
	dword eordt_ofs;
	eordt=(word*)malloc(edir.name_count*sizeof *efunct);
	if (!pe_rva_to_ofs(&pe_shared->sections, edir.ordinal_table_address, &eordt_ofs)) goto pe_read_error;
	file->seek(eordt_ofs);
	file->read(eordt, edir.name_count*sizeof *efunct);

	int *lord;
	lord=(int*)malloc(sizeof *lord*edir.function_count);
	memset(lord, 0, sizeof *lord*edir.function_count);
	for (dword i=0; i<edir.name_count; i++) {
		*(lord+*(eordt+i))=1;
	}

/* exports by ordinal */
	for (dword i=0; i<edir.function_count; i++) {
		if (*(lord+i)) continue;

		dword f=*(efunct+i)+pe_shared->pe32.header_nt.image_base;

		elinep=eline;
		elinep+=sprintf(elinep, "%04lx %08lx ", i+edir.ordinal_base, f);
		tag_make_ref(elinep, f, 0, "<ordinal only>");

		ht_pe_export_function *efd=new ht_pe_export_function(f, i+edir.ordinal_base);
		pe_shared->exports.funcs.insert(efd);
		m->add_mask(eline);
	}
	delete lord;
/* exports by name */
	for (dword i=0; i<edir.name_count; i++) {
		dword o=*(eordt+i);
		dword f=*(efunct+o)+pe_shared->pe32.header_nt.image_base;
		dword en;
		if (!pe_rva_to_ofs(&pe_shared->sections, *(enamet+i), &en)) goto pe_read_error;
		file->seek(en);
		char *s=getstrz(file);

		elinep=eline;
		elinep+=sprintf(elinep, "%04lx %08lx ", o+edir.ordinal_base, f);
		tag_make_ref(elinep, f, 0, s);

		ht_pe_export_function *efd=new ht_pe_export_function(f, o, s);
		pe_shared->exports.funcs.insert(efd);
		delete s;
		m->add_mask(eline);
	}

	if (!edir.function_count) m->add_mask("<no exported functions>");

	v->insertsub(m);

	g->insert(head);
	g->insert(v);

	g->setpalette(window_palette_str);

	stop_timer(h0);
	LOG("%s: PE: %d ticks (%d msec) to read exports", filename, get_timer_tick(h0), get_timer_msec(h0));
	delete_timer(h0);

	if (efunct) delete efunct;
	if (enamet) delete enamet;
	if (eordt) delete eordt;
	pe_shared->v_exports=v;
	return g;
pe_read_error:
	errorbox("%s: PE export directory seems to be corrupted.", file->get_name());
	if (efunct) delete efunct;
	if (enamet) delete enamet;
	if (eordt) delete eordt;
	g->done();
	delete g;
	m->done();
	delete m;
	v->done();
	delete v;
	return 0;
}

viewer_if htpeexports_if = {
	htpeexports_init,
	0
};

/*
 *	CLASS ht_pe_export_viewer
 */

void ht_pe_export_viewer::init(bounds *b, char *desc, ht_stream *file, int caps)
{
	ht_uviewer::init(b, desc, file, caps);
	VIEW_DEBUG_NAME("ht_pe_export_viewer");
}

void ht_pe_export_viewer::done()
{
	ht_uviewer::done();
}

char *ht_pe_export_viewer::func(int i, int getname)
{
	switch (i) {
		case 2:
			if (getname) return "sortord"; else {
				((hts_pe_export_mask*)cursor_sub)->sortbyord();
				dirtyview();
			}
			return 0;
		case 3:
			if (getname) return "sortva"; else {
				((hts_pe_export_mask*)cursor_sub)->sortbyva();
				dirtyview();
			}
			return 0;
		case 4:
			if (getname) return "sortname"; else {
				((hts_pe_export_mask*)cursor_sub)->sortbyname();
				dirtyview();
			}
			return 0;
		case 5:
			if (getname) return "unsort"; else {
				((hts_pe_export_mask*)cursor_sub)->unsort();
				dirtyview();
			}
			return 0;
	}
	return ht_uviewer::func(i, getname);
}

int ht_pe_export_viewer::ref_sel(ID id_low, ID id_high)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;
	int j=pe_shared->v_image->jump(id_low, this);
	if (j) app->focus(pe_shared->v_image); else errorbox("can't follow: export VA %08x is not valid !", id_low);
	return 1;
}

/*
 *	CLASS hts_pe_export_mask
 */

void hts_pe_export_mask::init(ht_stream *file)
{
	hts_mask::init(file);
	sort_ord=0;
	sort_va=0;
	sort_name=0;
	sort_path=0;
	sort=0;
}

void hts_pe_export_mask::done()
{
	if (sort_ord) delete sort_ord;
	if (sort_va) delete sort_va;
	if (sort_name) delete sort_name;
	hts_mask::done();
}

int hts_pe_export_mask::getline(char *line, dword id1, dword id2)
{
	if ((sort) && (sort_path) && (id1>=PE_EXPORT_DESC_LINES)) {
		return hts_mask::getline(line, sort_path[id1], id2);
	} else {
		return hts_mask::getline(line, id1, id2);
	}
}

void hts_pe_export_mask::sortbyord()
{
	if (!sort_ord) {
		int c=masks.count();
		byvalue *so=(byvalue*)malloc(sizeof *so * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data*)masks.get(i);
			char *s=d->tagstring;
			dword v=hexw(s);
			so[i].key=v;
			so[i].value=i;
		}
		qsortbyvalue(so, PE_EXPORT_DESC_LINES, c-1);
		sort_ord=(dword*)malloc(sizeof *sort_ord * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			sort_ord[i]=so[i].value;
		}
		delete so;
	}
	sort=1;
	sort_path=sort_ord;
}

void hts_pe_export_mask::sortbyva()
{
	if (!sort_va) {
		int c=masks.count();
		byvalue *sv=(byvalue*)malloc(sizeof *sv * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data*)masks.get(i);
			char *s=d->tagstring;
			dword v=hexd(s+5);
			sv[i].key=v;
			sv[i].value=i;
		}
		qsortbyvalue(sv, PE_EXPORT_DESC_LINES, c-1);
		sort_va=(dword*)malloc(sizeof *sort_va * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			sort_va[i]=sv[i].value;
		}
		delete sv;
	}
	sort=1;
	sort_path=sort_va;
}

void hts_pe_export_mask::sortbyname()
{
	if (!sort_name) {
		int c=masks.count();
		byname *sn=(byname*)malloc(sizeof *sn * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			hts_mask_data *d=(hts_mask_data*)masks.get(i);
			char *s=d->tagstring;
			char *t=tag_get_selectable_tag(s, 1, 0);
			int q=tag_get_seltextlen(t);

			sn[i].key=tag_get_seltext(t);;
			sn[i].keylen=q;
			sn[i].value=i;
		}
		qsortbyname(sn, PE_EXPORT_DESC_LINES, c-1);
		sort_name=(dword*)malloc(sizeof *sort_name * c);
		for (int i=PE_EXPORT_DESC_LINES; i<c; i++) {
			sort_name[i]=sn[i].value;
		}
		delete sn;
	}
	sort=1;
	sort_path=sort_name;
}

void hts_pe_export_mask::unsort()
{
	sort=0;
}

/*
 *	class ht_pe_export_function
 */

ht_pe_export_function::ht_pe_export_function()
{
	name=0;
	ordinal=0;
	function_address=0;
}

ht_pe_export_function::ht_pe_export_function(int _function_address, int _ordinal)
{
	ordinal=_ordinal;
	function_address=_function_address;
	byname=0;
}

ht_pe_export_function::ht_pe_export_function(int _function_address, int _ordinal, char *_name)
{
	ordinal=_ordinal;
	name=strdup(_name);
	function_address=_function_address;
	byname=1;
}

ht_pe_export_function::~ht_pe_export_function()
{
	if ((byname) && (name)) delete name;
}
