/*
 *	HT Editor
 *	htpehead.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htatom.h"
#include "htcoffhd.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpehead.h"
#include "httag.h"
#include "htstring.h"
#include "viewers.h"

#include "pestruct.h"

int_hash pe_optional_magics[] =
{
	{COFF_OPTMAGIC_ROMIMAGE, "ROM image"},
	{COFF_OPTMAGIC_PE32, "PE/PE32"},
	{COFF_OPTMAGIC_PE64, "PE32+/PE64"},
	{0, 0}
};

int_hash pe_subsystems[] =
{
	{0, "generic"},
	{PE_SUBSYSTEM_NATIVE, "native"},
	{PE_SUBSYSTEM_WINDOWS_GUI, "Windows GUI"},
	{PE_SUBSYSTEM_WINDOWS_CUI, "Windows CUI"},
	{PE_SUBSYSTEM_OS2_CUI, "OS/2 CUI"},
	{PE_SUBSYSTEM_POSIX_CUI, "POSIX CUI"},
	{9, "Windows CE GUI"},
	{10, "EFI"},
	{11, "EFI/boot"},
	{12, "EFI/runtime"},
	{0, 0}
};

char *pecoffheader[]=
{
	"machine                                          "STATICTAG_EDIT_WORD("00000000")" "STATICTAG_DESC_WORD("00000000", ATOM_PE_MACHINES_STR),
	"number of sections                               "STATICTAG_EDIT_WORD("00000002"),
	"time-date stamp                                  "STATICTAG_EDIT_TIME("00000004"),
	"pointer to symbol table                          "STATICTAG_EDIT_DWORD("00000008"),
	"number of symbols                                "STATICTAG_EDIT_DWORD("0000000c"),
	"size of optional header                          "STATICTAG_EDIT_WORD("00000010"),
	"characteristics                                  "STATICTAG_EDIT_WORD("00000012")" "STATICTAG_FLAGS("00000012", ATOM_PE_CHARACTERISTICS_STR),
	0
};

char *pe32header[] = {
	"optional magic                                   "STATICTAG_EDIT_WORD("00000014")" "STATICTAG_DESC_WORD("00000014", ATOM_PE_OPTIONAL_MAGICS_STR),
	"major linker version                             "STATICTAG_EDIT_BYTE("00000016"),
	"minor linker version                             "STATICTAG_EDIT_BYTE("00000017"),
	"size of code                                     "STATICTAG_EDIT_DWORD("00000018"),
	"size of data                                     "STATICTAG_EDIT_DWORD("0000001c"),
	"size of bss                                      "STATICTAG_EDIT_DWORD("00000020"),
	"entry point                                      "STATICTAG_EDIT_DWORD("00000024"),
	"code base                                        "STATICTAG_EDIT_DWORD("00000028"),
	"data base                                        "STATICTAG_EDIT_DWORD("0000002c")
};

char *pe32header_nt[] = {
	"image base                                       "STATICTAG_EDIT_DWORD("00000030"),
	"section alignment                                "STATICTAG_EDIT_DWORD("00000034"),
	"file alignment                                   "STATICTAG_EDIT_DWORD("00000038"),
	"major OS version                                 "STATICTAG_EDIT_WORD("0000003c"),
	"minor OS version                                 "STATICTAG_EDIT_WORD("0000003e"),
	"major image version                              "STATICTAG_EDIT_WORD("00000040"),
	"minor image version                              "STATICTAG_EDIT_WORD("00000042"),
	"major subsystem version                          "STATICTAG_EDIT_WORD("00000044"),
	"minor subsystem version                          "STATICTAG_EDIT_WORD("00000046"),
	"Win32 version                                    "STATICTAG_EDIT_DWORD("00000048"),
	"size of image                                    "STATICTAG_EDIT_DWORD("0000004c"),
	"size of headers                                  "STATICTAG_EDIT_DWORD("00000050"),
	"checksum                                         "STATICTAG_EDIT_DWORD("00000054"),
	"subsystem                                        "STATICTAG_EDIT_WORD("00000058")" "STATICTAG_DESC_WORD("00000058", ATOM_PE_SUBSYSTEMS_STR),
	"dll characteristics                              "STATICTAG_EDIT_WORD("0000005a"),
	"stack reserve                                    "STATICTAG_EDIT_DWORD("0000005c"),
	"stack commit                                     "STATICTAG_EDIT_DWORD("00000060"),
	"heap reserve                                     "STATICTAG_EDIT_DWORD("00000064"),
	"heap commit                                      "STATICTAG_EDIT_DWORD("00000068"),
	"loader flags                                     "STATICTAG_EDIT_DWORD("0000006c"),
	"number of directory entries                      "STATICTAG_EDIT_DWORD("00000070"),
	"--- directories ---",
	"export directory                             rva "STATICTAG_EDIT_DWORD("00000074")" size "STATICTAG_EDIT_DWORD("00000078"),
	"import directory                             rva "STATICTAG_EDIT_DWORD("0000007c")" size "STATICTAG_EDIT_DWORD("00000080"),
	"resource directory                           rva "STATICTAG_EDIT_DWORD("00000084")" size "STATICTAG_EDIT_DWORD("00000088"),
	"exception directory                          rva "STATICTAG_EDIT_DWORD("0000008c")" size "STATICTAG_EDIT_DWORD("00000090"),
	"security directory                           rva "STATICTAG_EDIT_DWORD("00000094")" size "STATICTAG_EDIT_DWORD("00000098"),
	"base relocation table                        rva "STATICTAG_EDIT_DWORD("0000009c")" size "STATICTAG_EDIT_DWORD("000000a0"),
	"debug directory                              rva "STATICTAG_EDIT_DWORD("000000a4")" size "STATICTAG_EDIT_DWORD("000000a8"),
	"description string                           rva "STATICTAG_EDIT_DWORD("000000ac")" size "STATICTAG_EDIT_DWORD("000000b0"),
	"machine value (GP)                           rva "STATICTAG_EDIT_DWORD("000000b4")" size "STATICTAG_EDIT_DWORD("000000b8"),
	"thread local storage (TLS)                   rva "STATICTAG_EDIT_DWORD("000000bc")" size "STATICTAG_EDIT_DWORD("000000c0"),
	"load configuration directory                 rva "STATICTAG_EDIT_DWORD("000000c4")" size "STATICTAG_EDIT_DWORD("000000c8"),
	"bound import directory                       rva "STATICTAG_EDIT_DWORD("000000cc")" size "STATICTAG_EDIT_DWORD("000000d0"),
	"import address table (IAT)                   rva "STATICTAG_EDIT_DWORD("000000d4")" size "STATICTAG_EDIT_DWORD("000000d8"),
	"delay import descriptor                      rva "STATICTAG_EDIT_DWORD("000000dc")" size "STATICTAG_EDIT_DWORD("000000e0"),
	"COM+ runtime header                          rva "STATICTAG_EDIT_DWORD("000000e4")" size "STATICTAG_EDIT_DWORD("000000e8"),
	"reserved (15)                                rva "STATICTAG_EDIT_DWORD("000000ec")" size "STATICTAG_EDIT_DWORD("000000f0"),
	0
};

char *section[] = {
	"-------------------------------------------------------------------------",
	"name                                             "STATICTAG_EDIT_CHAR("00000000") STATICTAG_EDIT_CHAR("00000001") STATICTAG_EDIT_CHAR("00000002") STATICTAG_EDIT_CHAR("00000003") STATICTAG_EDIT_CHAR("00000004") STATICTAG_EDIT_CHAR("00000005") STATICTAG_EDIT_CHAR("00000006") STATICTAG_EDIT_CHAR("00000007"),
	"virtual size                                     "STATICTAG_EDIT_DWORD("00000008"),
	"virtual address                                  "STATICTAG_EDIT_DWORD("0000000c"),
	"size                                             "STATICTAG_EDIT_DWORD("00000010"),
	"offset                                           "STATICTAG_EDIT_DWORD("00000014"),
	"relocation table offset                          "STATICTAG_EDIT_DWORD("00000018"),
	"line number table offset                         "STATICTAG_EDIT_DWORD("0000001c"),
	"relocation count                                 "STATICTAG_EDIT_WORD("00000020"),
	"line number count                                "STATICTAG_EDIT_WORD("00000022"),
	"characteristics                                  "STATICTAG_EDIT_DWORD("00000024")" "STATICTAG_FLAGS("00000024", ATOM_PE_SECTION_CHARACTERISTICS_STR),
	0
};

ht_view *htpeheader_init(bounds *b, ht_stream *file, ht_group *group)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;

	STREAMOFS h=pe_shared->hdr_ofs;
	ht_uviewer *v=new ht_uviewer();
	v->init(b, DESC_PE_HEADER, file, VC_EDIT);
	hts_mask *m=new hts_mask();
	m->init(file);
	register_atom(ATOM_PE_MACHINES, coff_machines);
	register_atom(ATOM_PE_OPTIONAL_MAGICS, pe_optional_magics);
	register_atom(ATOM_PE_SUBSYSTEMS, pe_subsystems);
	register_atom(ATOM_PE_CHARACTERISTICS, coff_characteristics);
	register_atom(ATOM_PE_SECTION_CHARACTERISTICS, coff_section_characteristics);
	char info[128];
	sprintf(info, "* PE header at offset %08lx", h);
	m->add_mask(info);
/* COFF header */
	m->add_mask("--- COFF header ---");
	m->add_staticmask("magic                                            "STATICTAG_EDIT_DWORD("00000000"), h);
	m->add_staticmask_table(pecoffheader, h+4);
/* optional header */
	m->add_mask("--- optional header ---");
	word opt;
	file->seek(h+24);
	file->read(&opt, 2);
	switch (opt) {
		case COFF_OPTMAGIC_PE32:
			m->add_staticmask_table(pe32header, h+4);
			m->add_mask("--- NT fields ---");
			m->add_staticmask_table(pe32header_nt, h+4);
			break;
		default: {
			m->add_staticmask("optional magic                                   "STATICTAG_EDIT_WORD("00000014")" "STATICTAG_DESC_WORD("00000014", ATOM_PE_OPTIONAL_MAGICS_STR), h+4);
			m->add_mask("-------------------------------------------------------------------------");
			m->add_mask("Unsupported optional magic ! If you get this message in an original");
			m->add_mask("(unmodified) file, please contact us (see help).");
		}
	}
/* section headers */
	int sc=0, os=0;
	file->seek(h+6);
	file->read(&sc, 2);
	file->seek(h+20);
	file->read(&os, 2);
	file->seek(h+os+24);
	for (int i=0; i<sc; i++) m->add_staticmask_table(section, h+24+os+i*40);
	v->insertsub(m);
	return v;
}

viewer_if htpeheader_if = {
	htpeheader_init,
	0
};
