/* 
 *	HT Editor
 *	htpeimg.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htcolor.h"
#include "htnewexe.h"
#include "htpeimg.h"
#include "htstring.h"
#include "viewers.h"

#include "pestruct.h"

#include "htanaly.h"
#include "pe_analy.h"

ht_view *htpeimage_init(bounds *b, ht_stream *file, ht_group *group)
{
	ht_group *pe_group=group;
	while (strcmp(pe_group->desc, DESC_PE)) pe_group=pe_group->group;
	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)pe_group)->shared_data;

	if (pe_shared->opt_magic!=COFF_OPTMAGIC_PE32) return 0;

	LOG("%s: PE: loading image (starting analyser)...", file->get_name());
	pe_analyser *p = new pe_analyser();
	p->init(pe_shared, file);
	p->beginanalysis(pe_shared->pe32.header.entrypoint_address+pe_shared->pe32.header_nt.image_base);

	bounds c=*b;
	ht_group *g=new ht_group();
	g->init(&c, VO_RESIZE, "pe/image-g");
	analy_infoline *head;

	c.y+=2;
	c.h-=2;
	ht_aviewer *v=new ht_aviewer();
	v->init(&c, DESC_PE_IMAGE, file, VC_EDIT | VC_GOTO | VC_SEARCH, p);

	c.y-=2;
	c.h=2;
	head=new analy_infoline();
	head->init(&c, v, ANALY_STATUS_DEFAULT);

	v->attach_infoline(head);

/* lowest/highest finden */
	ADDR l=(ADDR)-1, h=0;
	COFF_SECTION_HEADER *s=pe_shared->sections.sections;
	for (int i=0; i<pe_shared->sections.section_count; i++) {
		if (s->data_address<l) l=s->data_address;
		if (s->data_address+s->data_size>h) h=s->data_address+s->data_size;
		s++;
	}
/**/
	l+=pe_shared->pe32.header_nt.image_base;
	h+=pe_shared->pe32.header_nt.image_base;

	hts_analy *aaa=new hts_analy();
	aaa->init(file, p, l, h);
	v->insertsub(aaa);


/* FIXME: s/jump/jumpstr/ */
	tlabel *z=p->findlabel("entrypoint");
	v->sendmsg(htm_complete_init, 0);
	if (z) v->jump(z->addr->addr, 0);

	g->insert(head);
	g->insert(v);

	g->setpalette(window_palette_str);

	pe_shared->v_image=v;
	return g;
}

viewer_if htpeimage_if = {
	htpeimage_init,
	0
};

