/*
 *	HT Editor
 *	htstring.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "global.h"
#include "htdebug.h"
#include "htstring.h"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#ifndef HAVE_STRDUP

#include <string.h>

char *strdup(char *str)
{
	   if (str) {
			 int len=strlen(str)+1;
			 char *s=(char*)malloc(len);
			 memmove(s, str, len);
			 return s;
	   } else {
			 return 0;
	   }
}

#endif /* HAVE_STRDUP */

/* common string parsing functions */

void whitespaces(char **str)
{
	while ((unsigned char)**str<=32) {
		if (!**str) return;
		(*str)++;
	}
}

void non_whitespaces(char **str)
{
	while ((unsigned char)**str>32) {
		(*str)++;
	}
}

int waitforchar(char **str, char b)
{
	while (**str!=b) {
		if (!**str) return 0;
		(*str)++;
	}
	return 1;
}

int bnstr2bin(char *str, char *p, int base, dword *v)
{
	*v=0;
	do {
		int c=hexdigit(*str);
		if ((c==-1) || (c>=base)) return 0;
		(*v)*=base;
		*v+=c;
		str++;
	} while (str<p);
	return 1;
}

int bnstr(char **str, dword *v, int defaultbase)
{
	int base=defaultbase;
	int t=0;
	char *p=*str;
	while (!strchr("+-*/%()[] \t#.,:;", *p) && (*p)) p++;
	if (p==*str) return 0; /* zero length */
	if (strncmp("0x", *str, 2)==0) {
		(*str)+=2;
		base=16;
	} else {
		switch (*(p-1)) {
			case 'b':
				base=2;
				p--;
				t++;
				break;
			case 'o':
				base=8;
				p--;
				t++;
				break;
			case 'd':
				base=10;
				p--;
				t++;
				break;
			case 'h':
				base=16;
				p--;
				t++;
				break;
			default:
				if (**str=='0') base=8;
		}
	}
	if (bnstr2bin(*str, p, base, v)) {
		*str=p+t;
		return 1;
	}
	return 0;
}

/*
--- C/ASM strings ---
	tokens:
		"..."		C strings (allows escape sequences: \\, \a,\b,\e,\f,\n,\r,\t,\v,\xHH,\0OOO)
		'...'		pascal strings
		#32			pascal chars
		32			decimal numbers
		0xff			hexadecimal numbers (C notation)
		077			octal numbers 		(C notation)
		ffh 			hexadecimal numbers (ASM notation)
		77o			octal numbers 		(ASM notation)
		10011110b		binary numbers 	(ASM notation)
	separators:
		commas and/or whitespaces
*/

#define MAX_CASM_STR_LEN	1024

byte casmstr[MAX_CASM_STR_LEN];

char _casmerrstr[128];

char *casmerrstr=_casmerrstr;

void casmerrf(char *format, ...)
{
	va_list arg;
	va_start(arg, format);
	vsprintf(_casmerrstr, format, arg);
	va_end(arg);
}

/*
 *	eval_casmstr()
 */

byte *eval_casmstr(char *str, int *ret_strlen)
{
	char *origstr=str;
	byte *c=casmstr;
	whitespaces(&str);
	while (*str) {
		switch (*str) {
			case '"':
				str++;
				while (*str!='"') {
					if (!*str) {
						casmerrf("unterminated string (\"...\") at pos %d", str-origstr);
						return 0;
					} else if (*str=='\\') {
						str++;
						switch (*str) {
							case '\\':
								*(c++)='\\'; str++;
								break;
							case 'a': /* Alarm/bell, 7 */
								*(c++)='\a'; str++;
								break;
							case 'b': /* Backspace, 8 */
								*(c++)='\b'; str++;
								break;
							case 'e': /* Escape, 27 (1b) */
								*(c++)='\e'; str++;
								break;
							case 'f': /* form Feed, 12 (0c) */
								*(c++)='\f'; str++;
								break;
							case 'n': /* liNe feed, 10 (0a) */
								*(c++)='\n'; str++;
								break;
							case 'r': /* carriage Return, 13 (0d) */
								*(c++)='\r'; str++;
								break;
							case 't': /* (horizontal) Tab, 9 */
								*(c++)='\t'; str++;
								break;
							case 'v': /* Vertical tab, 11 */
								*(c++)='\v'; str++;
								break;
							case 'x': /* heX */ {
								dword v=0;
								str++;
								char *p=str+1;
								if (hexdigit(*str)==-1) {
									casmerrf("\\x used with no following hex digits at pos %d", str-origstr);
									return 0;
								}
								if (hexdigit(*(str+1))!=-1) p++;
								if (!bnstr2bin(str, p, 16, &v)) {
									casmerrf("syntax error at pos %d", str-origstr);
									return 0;
								}
								*(c++)=(byte)v;
								str=p;
								break;
							}
							default:
								if ((*str>='0') && (*str<='7')) {
									dword v=0, w;
									while (1) {
										if ((*str<'0') || (*str>'7')) break;
										w=v;
										v<<=3;
										v+=*str-'0';
										if (v>255) {
											v=w;
											break;
										}
										str++;
									}
									*(c++)=(byte)v;
								} else {
									*(c++)=*str;
								}
								break;
						}
					} else *(c++)=*(str++);
				}
				str++;
				break;
			case '\'': {
				char *s=++str;
				if (!waitforchar(&str, '\'')) {
					casmerrf("unterminated string ('...')  at pos %d", str-origstr);
					return 0;
				}
				memmove(c, s, str-s);
				c+=str-s;
				str++;
				break;
			}
			case '#': {
				str++;
				char *p=str;
				while (!strchr("#, ", *p)) p++;
				dword v;
				if (!bnstr2bin(str, p, 10, &v)) {
					casmerrf("invalid string for character at pos %d", str-origstr);
					return 0;
				}
				if (v>255) {
					casmerrf("value out of range for character (>255) at pos %d", str-origstr);
					return 0;
				}
				*(c++)=(byte)v;
				str=p;
				break;
			}
			default: {
				dword v;
				if (!bnstr(&str, &v, 10)) {
					casmerrf("invalid string for character at pos %d", str-origstr);
					return 0;
				}
				if (v>255) {
					casmerrf("value out of range (>255) for character at pos %d", str-origstr);
					return 0;
				}
				*(c++)=(byte)v;
				break;
			}
		}
		while (*str && (((unsigned char)*str<=32) || (*str==','))) str++;
	}
	if (ret_strlen) *ret_strlen=c-casmstr;
	*c=0;
	return casmstr;
}

/*
 *	eval_intexprstr()
 */

char _eval_int_errstr[128];
char *eval_int_errstr=_eval_int_errstr;

void eval_int_expr_error(char *format, ...)
{
	va_list arg;
	va_start(arg, format);
	vsprintf(eval_int_errstr, format, arg);
	va_end(arg);
}

int (*global_int_eval_proc)(void *context, char **s, dword *v);
void *global_int_eval_proc_context;

int eval_int_expr(char **str, dword *v, char terminator);

int eval_int_token(char **str, dword *v)
{
	if ((global_int_eval_proc) && (global_int_eval_proc(global_int_eval_proc_context, str, v))) return 1;
	return bnstr(str, v, 10);
}

int eval_int_sign_and_token(char **str, dword *v)
{
	int sign=0;
	while ((**str=='-') || (**str=='+')) {
		if (**str=='-') sign=sign^1;
		(*str)++;
		whitespaces(str);
	}
	if (**str=='(') {
		(*str)++;
		if (!eval_int_expr(str, v, ')')) {
/*			char *e=t;
			non_whitespaces(&e);
			char temp=*e;
			*e=0;
			eval_int_expr_error("invalid token '%s'", t);
			*e=temp;*/
			return 0;
		}
		(*str)++;
	} else {
		char *t=*str;
		if (!eval_int_token(str, v)) {
			char *e=t;
			non_whitespaces(&e);
			char temp=*e;
			*e=0;
			eval_int_expr_error("invalid token '%s'", t);
			*e=temp;
			return 0;
		}
	}
	if (sign) *v=-*v;
	return 1;
}

int eval_int_expr_mul(char **str, dword *v)
{
	int brk=0;
	dword value=0, sc_value=0;
	whitespaces(str);
	if (!eval_int_sign_and_token(str, &value)) return 0;
	while (!brk) {
		whitespaces(str);
		switch (**str) {
			case '*':
				(*str)++;
				whitespaces(str);
				if (!eval_int_sign_and_token(str, &sc_value)) return 0;
				value*=sc_value;
				break;
			case '/':
				(*str)++;
				whitespaces(str);
				if (!eval_int_sign_and_token(str, &sc_value)) return 0;
				if (!sc_value) {
					eval_int_expr_error("division by zero");
					return 0;
				}
				value/=sc_value;
				break;
			case '%':
				(*str)++;
				whitespaces(str);
				if (!eval_int_sign_and_token(str, &sc_value)) return 0;
				if (!sc_value) {
					eval_int_expr_error("division by zero");
					return 0;
				}
				value%=sc_value;
				break;
			default:
				brk=1;
				break;
		}
	}
	*v=value;
	return 1;
}

int eval_int_expr_add(char **str, dword *v)
{
	int brk=0;
	dword value=0, sc_value=0;
	whitespaces(str);
	if (!eval_int_expr_mul(str, &value)) return 0;
	while (!brk) {
		whitespaces(str);
		switch (**str) {
			case '+':
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_mul(str, &sc_value)) return 0;
				value+=sc_value;
				break;
			case '-':
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_mul(str, &sc_value)) return 0;
				value-=sc_value;
				break;
			default:
				brk=1;
				break;
		}
	}
	*v=value;
	return 1;
}

int eval_int_expr_shift(char **str, dword *v)
{
	int brk=0;
	dword value=0, sc_value=0;
	whitespaces(str);
	if (!eval_int_expr_add(str, &value)) return 0;
	while (!brk) {
		whitespaces(str);
		switch (**str) {
			case '<':
				(*str)++;
				if (**str!='<') break;
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_add(str, &sc_value)) return 0;
				value<<=sc_value;
				break;
			case '>':
				(*str)++;
				if (**str!='>') break;
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_add(str, &sc_value)) return 0;
				value>>=sc_value;
				break;
			default:
				brk=1;
				break;
		}
	}
	*v=value;
	return 1;
}

int eval_int_expr(char **str, dword *v, char terminator)
{
	int brk=0;
	dword value=0, sc_value=0;
	whitespaces(str);
	if (!eval_int_expr_shift(str, &value)) return 0;
	while (!brk) {
		whitespaces(str);
		if (**str==terminator) break;
		switch (**str) {
			case '|':
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_shift(str, &sc_value)) return 0;
				value|=sc_value;
				break;
			case '&':
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_shift(str, &sc_value)) return 0;
				value&=sc_value;
				break;
			case '^':
				(*str)++;
				whitespaces(str);
				if (!eval_int_expr_add(str, &sc_value)) return 0;
				value^=sc_value;
				break;
			default:
				eval_int_expr_error("invalid operator '%c'", **str);
				return 0;
		}
	}
	*v=value;
	return 1;
}

int eval_int_expr(char *str, dword *v)
{
	return eval_int_expr_custom(str, v, 0, 0);
}

int eval_int_expr_custom(char *str, dword *v, void *eval_proc_context, int (*eval_proc)(void *context, char **s, dword *v))
{
	global_int_eval_proc=eval_proc;
	global_int_eval_proc_context=eval_proc_context;
	return eval_int_expr(&str, v, 0);
}

int hex2int(char *s, dword *v)
{
	dword n=0;
	while (*s) {
		unsigned char b=*s;
		if ((b>='0') && (b<='9')) b-='0'; else
		if ((b>='a') && (b<='f')) b-='a'-10; else
		if ((b>='A') && (b<='F')) b-='A'-10; else return 0;
		n<<=4;
		n|=b;
		s++;
	}
	*v=n;
	return 1;
}

/* hex/string functions */

int hexdigit(char a)
{
	if ((a>='0') && (a<='9')) {
		return a-'0';
	} else if ((a>='a') && (a<='f')) {
		return a-'a'+10;
	} else if ((a>='A') && (a<='F')) {
		return a-'A'+10;
	}
	return -1;
}

dword hexb(char *s)
{
	byte b=*(byte*)s;
	b-='0';
	if (b>9) b-='a'-'0'-10;
	byte c=*(byte*)(s+1);
	c-='0';
	if (c>9) c-='a'-'0'-10;
	return (b<<4)+c;
}

dword hexw(char *s)
{
	return (hexb(s)<<8) | hexb(s+2);
}

dword hexd(char *s)
{
	return (hexw(s)<<16) | hexw(s+4);
}

char hexchars[17]="0123456789abcdef";

char *mkhexb(char *buf, byte d)
{
	*buf++=hexchars[(d>>4)&0xf];
	*buf++=hexchars[d&0xf];
	return buf;
}

char *mkhexw(char *buf, word d)
{
	*buf++=hexchars[(d>>12)&0xf];
	*buf++=hexchars[(d>>8)&0xf];
	*buf++=hexchars[(d>>4)&0xf];
	*buf++=hexchars[d&0xf];
	return buf;
}

char *mkhexd(char *buf, dword d)
{
	*buf++=hexchars[(d>>28)&0xf];
	*buf++=hexchars[(d>>24)&0xf];
	*buf++=hexchars[(d>>20)&0xf];
	*buf++=hexchars[(d>>16)&0xf];
	*buf++=hexchars[(d>>12)&0xf];
	*buf++=hexchars[(d>>8)&0xf];
	*buf++=hexchars[(d>>4)&0xf];
	*buf++=hexchars[d&0xf];
	return buf;
}

