/* 
 *	HT Editor
 *	httag.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTTAG_H__
#define __HTTAG_H__

#include "global.h"
#include "htstream.h"

/* SELECTION-TAG */
#define HT_TAG_SEL				0x01
#define HT_TAG_SEL_LEN(n)		(sizeof(ht_tag_sel)+n)
#define HT_TAG_SEL_VLEN(n)		(n)

#define HT_STATICTAG_SEL			0x81
#define HT_STATICTAG_SEL_CH		"\e\x81"

struct ht_tag_sel {
	byte escape;
	byte magic;
	ID id64_low, id64_high;
	byte strlen;
};

/* FLAGS-TAG */
#define HT_TAG_FLAGS			0x02
#define HT_TAG_FLAGS_LEN           sizeof(ht_tag_flags)
#define HT_TAG_FLAGS_VLEN		1

#define HT_STATICTAG_FLAGS		0x82
#define HT_STATICTAG_FLAGS_CH		"\e\x82"

struct ht_tag_flags {
	byte escape;
	byte magic;
	STREAMOFS offset;
	ID id;
};

struct ht_tag_flags_s {
	char bitidx;
	char *desc;
};

/* GROUP-TAG */
#define HT_TAG_GROUP			0x03
#define HT_TAG_GROUP_LEN			sizeof(ht_tag_group)

#define HT_STATICTAG_GROUP		0x83
#define HT_STATICTAG_GROUP_CH		"\e\x83"

struct ht_tag_group {
	byte escape;
	byte magic;
};

/* COLOR-TAG */
#define HT_TAG_COLOR			0x04
#define HT_TAG_COLOR_LEN			sizeof(ht_tag_color)

#define HT_STATICTAG_COLOR		0x84
#define HT_STATICTAG_COLOR_CH		"\e\x84"

struct ht_tag_color {
	byte escape;
	byte magic;
	int color;
};

/* EDIT-BYTE-TAG */
#define HT_TAG_EDIT_BYTE			0x10
#define HT_TAG_EDIT_BYTE_LEN		sizeof(ht_tag_edit_byte)
#define HT_TAG_EDIT_BYTE_VLEN		(2)
#define HT_TAG_EDIT_BYTE_SIZE		(1)

#define HT_STATICTAG_EDIT_BYTE	0x90
#define HT_STATICTAG_EDIT_BYTE_CH	"\e\x90"

struct ht_tag_edit_byte {
	byte escape;
	byte magic;
	STREAMOFS offset;
};

/* EDIT-WORD-TAG */
#define HT_TAG_EDIT_WORD			0x11
#define HT_TAG_EDIT_WORD_LEN		sizeof(ht_tag_edit_word)
#define HT_TAG_EDIT_WORD_VLEN		(4)
#define HT_TAG_EDIT_WORD_SIZE		(2)

#define HT_STATICTAG_EDIT_WORD	0x91
#define HT_STATICTAG_EDIT_WORD_CH	"\e\x91"

struct ht_tag_edit_word {
	byte escape;
	byte magic;
	STREAMOFS offset;
};

/* EDIT-DWORD-TAG */
#define HT_TAG_EDIT_DWORD		0x12
#define HT_TAG_EDIT_DWORD_LEN		sizeof(ht_tag_edit_dword)
#define HT_TAG_EDIT_DWORD_VLEN	(8)
#define HT_TAG_EDIT_DWORD_SIZE	(4)

#define HT_STATICTAG_EDIT_DWORD	0x92
#define HT_STATICTAG_EDIT_DWORD_CH	"\e\x92"

struct ht_tag_edit_dword {
	byte escape;
	byte magic;
	STREAMOFS offset;
};

/* EDIT-TIME-TAG */
#define HT_TAG_EDIT_TIME			0x13
#define HT_TAG_EDIT_TIME_LEN		sizeof(ht_tag_edit_time)
#define HT_TAG_EDIT_TIME_VLEN		(19)
#define HT_TAG_EDIT_TIME_SIZE		(4)

#define HT_STATICTAG_EDIT_TIME	0x93
#define HT_STATICTAG_EDIT_TIME_CH	"\e\x93"

struct ht_tag_edit_time {
	byte escape;
	byte magic;
	STREAMOFS offset;
};

/* EDIT-CHAR-TAG */
#define HT_TAG_EDIT_CHAR			0x14
#define HT_TAG_EDIT_CHAR_LEN		sizeof(ht_tag_edit_char)
#define HT_TAG_EDIT_CHAR_VLEN		(1)
#define HT_TAG_EDIT_CHAR_SIZE		(1)

#define HT_STATICTAG_EDIT_CHAR 	0x94
#define HT_STATICTAG_EDIT_CHAR_CH	"\e\x94"

struct ht_tag_edit_char {
	byte escape;
	byte magic;
	STREAMOFS offset;
};

/* EDIT-BIT-TAG */
#define HT_TAG_EDIT_BIT			0x15
#define HT_TAG_EDIT_BIT_LEN		sizeof(ht_tag_edit_bit)
#define HT_TAG_EDIT_BIT_VLEN		(1)
#define HT_TAG_EDIT_BIT_SIZE		(1)

#define HT_STATICTAG_EDIT_BIT 	0x95
#define HT_STATICTAG_EDIT_BIT_CH	"\e\x95"

struct ht_tag_edit_bit {
	byte escape;
	byte magic;
	STREAMOFS offset;
	byte bitidx;
};

/* EDIT-SELVIS-TAG */
#define HT_TAG_EDIT_SELVIS			0x16
#define HT_TAG_EDIT_SELVIS_LEN		sizeof(ht_tag_edit_selvis)
#define HT_TAG_EDIT_SELVIS_VLEN		(1)

#define HT_STATICTAG_EDIT_SELVIS		0x96
#define HT_STATICTAG_EDIT_SELVIS_CH	"\e\x96"

struct ht_tag_edit_selvis {
	byte escape;
	byte magic;
	STREAMOFS offset;
	char ch;
};

/* DESC-BYTE-TAG */
#define HT_TAG_DESC_BYTE			0x20
#define HT_TAG_DESC_BYTE_LEN		sizeof(ht_tag_desc_byte)

#define HT_STATICTAG_DESC_BYTE	0xa0
#define HT_STATICTAG_DESC_BYTE_CH	"\e\xa0"

struct ht_tag_desc_byte {
	byte escape;
	byte magic;
	ID id;
	STREAMOFS offset;
};

/* DESC-WORD-TAG */
#define HT_TAG_DESC_WORD			0x21
#define HT_TAG_DESC_WORD_LEN		sizeof(ht_tag_desc_word)

#define HT_STATICTAG_DESC_WORD	0xa1
#define HT_STATICTAG_DESC_WORD_CH	"\e\xa1"

struct ht_tag_desc_word {
	byte escape;
	byte magic;
	ID id;
	STREAMOFS offset;
};

/* DESC-DWORD-TAG */
#define HT_TAG_DESC_DWORD		0x22
#define HT_TAG_DESC_DWORD_LEN		sizeof(ht_tag_desc_dword)

#define HT_STATICTAG_DESC_DWORD	0xa2
#define HT_STATICTAG_DESC_DWORD_CH	"\e\xa2"

struct ht_tag_desc_dword {
	byte escape;
	byte magic;
	ID id;
	STREAMOFS offset;
};

/* tag color indices */

#define tag_coloridx_normal					0

#define tag_coloridx_edit_cursor_sel			1
#define tag_coloridx_edit_cursor_edit			2
#define tag_coloridx_edit_selected				3
#define tag_coloridx_edit_modified				4
#define tag_coloridx_edit_normal				5

#define tag_coloridx_sel_cursor  				6
#define tag_coloridx_sel_normal				7

#define STATICTAG_SEL(len8, str) HT_STATICTAG_SEL_CH "0000000000000000" len8 str
#define STATICTAG_REF(id64, len8, str) HT_STATICTAG_SEL_CH id64 len8 str
#define STATICTAG_FLAGS(ofs32, id32) HT_STATICTAG_FLAGS_CH ofs32 id32
#define STATICTAG_GROUP() HT_STATICTAG_GROUP_CH
#define STATICTAG_COLOR(color8) HT_STATICTAG_COLOR_CH color8

#define STATICTAG_EDIT_BYTE(ofs32) HT_STATICTAG_EDIT_BYTE_CH ofs32
#define STATICTAG_EDIT_WORD(ofs32) HT_STATICTAG_EDIT_WORD_CH ofs32
#define STATICTAG_EDIT_DWORD(ofs32) HT_STATICTAG_EDIT_DWORD_CH ofs32
#define STATICTAG_EDIT_TIME(ofs32) HT_STATICTAG_EDIT_TIME_CH ofs32
#define STATICTAG_EDIT_CHAR(ofs32) HT_STATICTAG_EDIT_CHAR_CH ofs32
#define STATICTAG_EDIT_BIT(ofs32, bitidx8) HT_STATICTAG_EDIT_BIT_CH ofs32 bitidx8
#define STATICTAG_EDIT_SELVIS(ofs32, char8) HT_STATICTAG_EDIT_SELVIS_CH ofs32 char8

#define STATICTAG_DESC_BYTE(ofs32, id32) HT_STATICTAG_DESC_BYTE_CH ofs32 id32
#define STATICTAG_DESC_WORD(ofs32, id32) HT_STATICTAG_DESC_WORD_CH ofs32 id32
#define STATICTAG_DESC_DWORD(ofs32, id32) HT_STATICTAG_DESC_DWORD_CH ofs32 id32

#define tag_class_no	0
#define tag_class_edit	1
#define tag_class_sel	2

typedef char TAGSTRING;

void statictag_to_tag(char *statictag_str, TAGSTRING *tag_str, dword relocation);

TAGSTRING *tag_findnext(TAGSTRING *tagstring);
void tag_get_id(TAGSTRING *tagstring, dword *id64_low, dword *id64_high);
int tag_get_len(TAGSTRING *tagstring);
dword tag_get_offset(TAGSTRING *tagstring);
int tag_get_size(TAGSTRING *tagstring);
int tag_get_vlen(TAGSTRING *tagstring);
int tag_get_seltextlen(TAGSTRING *tagstring);
char* tag_get_seltext(TAGSTRING *tagstring);
int tag_get_micropos(TAGSTRING *tagstring, int i);
int tag_get_microsize(TAGSTRING *tagstring);
int tag_get_desc_id(TAGSTRING *tagstring, dword *id);

void tag_set_offset(TAGSTRING *tagstring, dword offset);
void tag_set_value(TAGSTRING *tagstring, dword value);

int tag_is_editable(TAGSTRING *tagstring);

void tag_strcat(TAGSTRING *dest, TAGSTRING *src);
void tag_strcpy(TAGSTRING *dest, TAGSTRING *src);
TAGSTRING *tag_strdup(TAGSTRING *tagstring);
int tag_strlen(TAGSTRING *tagstring);

int tag_count_selectable_tags_in_group(TAGSTRING *tagstring, int group);
int tag_count_selectable_tags(TAGSTRING *tagstring);
int tag_count_groups(TAGSTRING *tagstring);
TAGSTRING *tag_get_selectable_tag(TAGSTRING *tagstring, int n, int group);
TAGSTRING *tag_get_group(TAGSTRING *tagstring, int group);
int tag_get_class(TAGSTRING *tagstring);

TAGSTRING *tag_make_sel(TAGSTRING *buf, char *string);
TAGSTRING *tag_make_ref(TAGSTRING *buf, dword id64_low, dword id64_high, char *string);
TAGSTRING *tag_make_flags(TAGSTRING *buf, dword id, dword offset);
TAGSTRING *tag_make_group(TAGSTRING *buf);
TAGSTRING *tag_make_color(TAGSTRING *buf, int color);
TAGSTRING *tag_make_edit_byte(TAGSTRING *buf, STREAMOFS ofs32);
TAGSTRING *tag_make_edit_word(TAGSTRING *buf, STREAMOFS ofs32);
TAGSTRING *tag_make_edit_dword(TAGSTRING *buf, STREAMOFS ofs32);
TAGSTRING *tag_make_edit_time(TAGSTRING *buf, STREAMOFS ofs32);
TAGSTRING *tag_make_edit_char(TAGSTRING *buf, STREAMOFS ofs32);
TAGSTRING *tag_make_edit_bit(TAGSTRING *buf, STREAMOFS ofs32, int bitidx);
TAGSTRING *tag_make_edit_selvis(TAGSTRING *buf, STREAMOFS offset, char ch);
TAGSTRING *tag_make_desc_byte(TAGSTRING *buf, STREAMOFS ofs32, dword id32);
TAGSTRING *tag_make_desc_word(TAGSTRING *buf, STREAMOFS ofs32, dword id32);
TAGSTRING *tag_make_desc_dword(TAGSTRING *buf, STREAMOFS ofs32, dword id32);

#endif /* !__HTTAG_H__ */
