/*
 *	HT Editor
 *	htviewer.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "htapp.h"
#include "htatom.h"
#include "htclipboard.h"
#include "htcolor.h"
#include "htctrl.h"
#include "htdebug.h"
#include "htdialog.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htsearch.h"
#include "htstring.h"
#include "httag.h"
#include "htviewer.h"
#include "tools.h"

#define zerofillchar '.'

/*
 *	CLASS ht_viewer
 */

void ht_viewer::init(bounds *b, char *desc, ht_stream *_stream, int _caps)
{
	ht_view::init(b, VO_OWNBUFFER | VO_BROWSABLE | VO_SELECTABLE | VO_RESIZE, desc);
	VIEW_DEBUG_NAME("ht_viewer");
	growmode_x=GROW_EDP;
	growmode_y=GROW_EDP;
	caps=_caps;
	stream=_stream;
	shared_data=0;
	vs_history=new vs_stack();
}

void ht_viewer::done()
{
	delete vs_history;
	ht_view::done();
}

char *ht_viewer::func(int i, int getname)
{
// FIXME: dirty hack !
	if (baseview==app) {
		return ((ht_app*)app)->func(i, getname);
	}
	return 0;
}

int ht_viewer::convert_str_to_addr(char *str, VIEWERADDR *addr)
{
	return 0;
}

int ht_viewer::convert_addr_to_ofs(VIEWERADDR addr, STREAMOFS *ofs)
{
	return 0;
}

int ht_viewer::get_current_offset(STREAMOFS *offset)
{
	return 0;
}

STUB void ht_viewer::handlemsg(htmsg *msg)
{
	int i=0;
	switch (msg->msg) {
		case htm_keypressed: {
			switch (msg->data_int) {
				case K_F12: i++;
				case K_F11: i++;
				case K_F10: i++;
				case K_F9: i++;
				case K_F8: i++;
				case K_F7: i++;
				case K_F6: i++;
				case K_F5: i++;
				case K_F4: i++;
				case K_F3: i++;
				case K_F2: i++;
				case K_F1: {
					i++;
					htmsg m;
					m.msg=htm_funcname;
					m.data_int=i;
					sendmsg(&m);
					if ((m.msg==htm_funcname_return) && (m.data_str)) {
						sendmsg(htm_funcexec, i);
						clearmsg(msg);
						return;
					}
					break;
				}
			}
			break;
		}
		case htm_vs_restore:
			view_state_load(msg->data_ptr);
			clearmsg(msg);
			return;
		case htm_funcexec:
			func(msg->data_int, 0);
			clearmsg(msg);
			return;
		case htm_funcname:
			msg->msg=htm_funcname_return;
			msg->data_str=func(msg->data_int, 1);
			break;
	}
	ht_view::handlemsg(msg);
}

int ht_viewer::jump(VIEWERADDR addr, ht_view *to_history_focused)
{
	return 0;
}

int ht_viewer::jumpstr(char *str)
{
	VIEWERADDR a;
	if (convert_str_to_addr(str, &a)) {
		int j=jump(a, this);
		if (!j) errorbox("invalid address %08lx", a);
		return j;
	} else errorbox(globalerror);
	return 0;
}

int ht_viewer::push_vs_history(ht_view *focused)
{
	void *vs=view_state_create();
	if (vs) {
		vstate *v=new vstate;
		v->focused=focused;
		v->view=this;
		v->data=vs;
		vs_history->push(v);
		return 1;
	}
	return 0;
}

int ht_viewer::pop_vs_history()
{
	vstate *vs=vs_history->pop();
	if (vs) {
		app->focus(vs->focused);
		vs->view->sendmsg(htm_vs_restore, (int)vs->data);
		delete vs;
		return 1;
	}
	return 0;
}

void ht_viewer::view_state_load(void *view_state)
{
}

int ht_viewer::search(ht_search *search, STREAMOFS start, STREAMOFS end)
{
	return 0;
}

void *ht_viewer::view_state_create()
{
	return 0;
}

void ht_viewer::view_state_destroy(void *view_state)
{
	free(view_state);
}

/*
 *	CLASS ht_uviewer
 */

void ht_uviewer::init(bounds *b, char *desc, ht_stream *stream, int caps)
{
	ht_viewer::init(b, desc, stream, caps);
	VIEW_DEBUG_NAME("ht_uviewer");
	edit=0;
	first_sub=0;
	last_sub=0;
	xscroll=0;
	cursor_ypos=0;
	cursor_visual_length=0;
	cursor_visual_xpos=0;
	sel_start=0;
	sel_end=0;
	has_last_search=0;
	initialized=0;
	isdirty_cursor_line=0;
}

void ht_uviewer::done()
{
	edit_end();
	ht_viewer::done();
}

void ht_uviewer::adjust_cursor_group()
{
	cursorline_get();
	int g=tag_count_groups(cursor_line);
	if (cursor_tag_group>=g) cursor_tag_group=0;
}

void ht_uviewer::adjust_cursor_idx()
{
	cursorline_get();
	int c=tag_count_selectable_tags_in_group(cursor_line, cursor_tag_group);
	if (cursor_tag_idx>c-1) cursor_tag_idx=c-1;
}

int ht_uviewer::center_view(hts_sub *sub, ID id1, ID id2)
{
	int r=prev_line(&sub, &id1, &id2, size.h/2);
	top_sub=sub;
	top_id1=id1;
	top_id2=id2;
	cursorline_dirty();
	return r;
}

void ht_uviewer::check_cursor_visibility()
{
	if (cursor_state!=cursor_state_disabled) {
		if ((cursor_ypos<0) || (cursor_ypos>=size.h)) {
			cursor_state=cursor_state_invisible;
		} else {
			cursor_state=cursor_state_visible;
		}
	}
}

void ht_uviewer::complete_init()
{
	if (!first_sub) return;
	if (initialized) return;
/*  initialize top_*  */
	top_sub=first_sub;
	top_sub->first_line_id(&top_id1, &top_id2);
/*  initialize cursor_*  */
	cursor_state=cursor_state_disabled;
	cursor_tag_idx=0;
	cursor_tag_group=0;

	cursor_sub=first_sub;
	cursor_sub->first_line_id(&cursor_id1, &cursor_id2);
	char line[1024];	/* FIXME: possible buffer overflow ! */
	cursor_ypos--;
	do {
		cursor_ypos++;
		if (!cursor_sub->getline(line, cursor_id1, cursor_id2)) break;
		if (tag_count_selectable_tags(line)) {
			cursor_state=(cursor_ypos<size.h) ? cursor_state_visible : cursor_state_invisible;
			break;
		}
	} while ((next_line(&cursor_sub, &cursor_id1, &cursor_id2, 1)) && (cursor_ypos<size.h));
	if (cursor_state==cursor_state_disabled) {
		cursor_sub=first_sub;
		cursor_sub->first_line_id(&cursor_id1, &cursor_id2);
		if (cursor_sub->getline(line, cursor_id1, cursor_id2)) {
			cursor_ypos=0;
			cursor_state=cursor_state_invisible;
		}
	}
/* get cursorline */
	cursorline_dirty();
	cursorline_get();
/* initialize visual */
	update_visual_info();
/* initialize misc */
	update_misc_info();
/**/
	initialized=1;
}

int ht_uviewer::continue_search()
{
	if (has_last_search) {
		STREAMOFS o;
		if (get_current_offset(&o)) {
			return search(&last_search, o+1, last_search_end);
		}
	}
	return 0;
}

int ht_uviewer::convert_str_to_addr(char *str, VIEWERADDR *addr)
{
	int ev=eval_int_expr(str, addr);
	if (!ev) strcpy(globalerror, eval_int_errstr);
	return ev;
}

int ht_uviewer::convert_addr_to_ofs(VIEWERADDR addr, STREAMOFS *ofs)
{
/* FIXME: hack! */
	*ofs=addr;
	return 1;
}

int ht_uviewer::cursor_left()
{
	if (cursor_tag_idx) {
		cursor_tag_idx--;
		update_visual_info();
		update_misc_info();
		return 1;
	} else {
		if (cursor_up(1)) {
			cursor_end();
			return 1;
		}
	}
	return 0;
}

int ht_uviewer::cursor_right()
{
	cursorline_get();
	if (cursor_tag_idx<tag_count_selectable_tags_in_group(cursor_line, cursor_tag_group)-1) {
		cursor_tag_idx++;
		update_visual_info();
		update_misc_info();
		return 1;
	} else {
		if (cursor_down(1)) {
			cursor_home();
			return 1;
		}
	}
	return 0;
}

int ht_uviewer::cursor_up(int n)
{
	switch (cursor_state) {
		case cursor_state_invisible:
		case cursor_state_visible: {
			if ((n==1) && (cursor_state==cursor_state_visible)) {
				int r=0;
				hts_sub *c_sub=cursor_sub;
				ID c_id1=cursor_id1, c_id2=cursor_id2;
				char c_line[1024];
				int c_ypos=cursor_ypos;
				int c_tag_idx=cursor_tag_idx;
				int c_tag_group=cursor_tag_group;
				int d_tag_group=cursor_tag_group;

				while (prev_line(&c_sub, &c_id1, &c_id2, 1) && (c_ypos>=0)) {
					c_ypos--;
					c_sub->getline(c_line, c_id1, c_id2);
					int g=tag_count_groups(c_line);
					if (d_tag_group<g) c_tag_group=d_tag_group;
					int s;
					if (c_tag_group>=g) {
						c_tag_group=g-1;
						s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
						if (s) {
							c_tag_idx=s-1;
							r=1;
							break;
						}
					} else {
						s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
						if (s) {
							if (c_tag_idx>=s) c_tag_idx=s-1;
							r=1;
							break;
						}
					}
				}
				if (r) {
					cursor_sub=c_sub;
					cursor_id1=c_id1;
					cursor_id2=c_id2;
					memmove(cursor_line, c_line, sizeof cursor_line);
					cursorline_dirty();
					cursor_ypos=c_ypos;
					cursor_tag_idx=c_tag_idx;
					cursor_tag_group=c_tag_group;
					if (cursor_ypos<=-1) scroll_up(-cursor_ypos);
					update_misc_info();
					update_visual_info();
					if (edit) update_micropos();
				} else scroll_up(n);
				return r;
			} else {
				int r=0;
				char c_line[1024];
				int c_tag_idx=cursor_tag_idx;
				int c_tag_group=cursor_tag_group;
				int d_tag_group=cursor_tag_group;
				hts_sub *c_sub;
				ID c_id1;
				ID c_id2;
				int c_ypos;
				if (cursor_state==cursor_state_invisible) {
					c_sub=top_sub;
					c_id1=top_id1;
					c_id2=top_id2;
					c_ypos=0;
				} else {
					c_sub=cursor_sub;
					c_id1=cursor_id1;
					c_id2=cursor_id2;
					c_ypos=cursor_ypos;
				}
				int nc=prev_line(&c_sub, &c_id1, &c_id2, n);
				c_ypos-=nc;

				while (nc--) {
					c_sub->getline(c_line, c_id1, c_id2);
					int g=tag_count_groups(c_line);
					if (d_tag_group<g) c_tag_group=d_tag_group;
					int s;
					if (c_tag_group>=g) {
						c_tag_group=g-1;
						s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
						if (s) {
							c_tag_idx=s-1;
							r=1;
							break;
						}
					}
					s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
					if (s) {
						if (c_tag_idx>=s) c_tag_idx=s-1;
						r=1;
						break;
					}
					if (!next_line(&c_sub, &c_id1, &c_id2, 1)) break;
					c_ypos++;
				}
				if (r) {
					cursor_sub=c_sub;
					cursor_id1=c_id1;
					cursor_id2=c_id2;
					memmove(cursor_line, c_line, sizeof cursor_line);
					cursorline_dirty();
					cursor_ypos=c_ypos;
					cursor_tag_idx=c_tag_idx;
					cursor_tag_group=c_tag_group;
					if (-cursor_ypos+n-nc-1>0) scroll_up(-cursor_ypos+n-nc-1);
					update_misc_info();
					update_visual_info();
					if (edit) update_micropos();
				} else scroll_up(n);
				break;
			}
		}
		case cursor_state_disabled:
			scroll_up(n);
			return n;
	}
	return 0;
}

int ht_uviewer::cursor_down(int n)
{
	switch (cursor_state) {
		case cursor_state_invisible:
		case cursor_state_visible: {
			if ((n==1) && (cursor_state==cursor_state_visible)) {
				int r=0;
				hts_sub *c_sub=cursor_sub;
				ID c_id1=cursor_id1, c_id2=cursor_id2;
				char c_line[1024];
				int c_ypos=cursor_ypos;
				int c_tag_idx=cursor_tag_idx;
				int c_tag_group=cursor_tag_group;
				int d_tag_group=cursor_tag_group;
				int nls;	/* controls scrolling beyond end */

				while ((nls=next_line(&c_sub, &c_id1, &c_id2, 1)) && (c_ypos<=size.h-1)) {
					c_ypos++;
					c_sub->getline(c_line, c_id1, c_id2);
					int g=tag_count_groups(c_line);
					if (d_tag_group<g) c_tag_group=d_tag_group;
					int s;
					if (c_tag_group>=g) {
						c_tag_group=g-1;
						s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
						if (s) {
							c_tag_idx=s-1;
							r=1;
							break;
						}
					}
					s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
					if (s) {
						if (c_tag_idx>=s) c_tag_idx=s-1;
						r=1;
						break;
					}
				}
				if (r) {
					cursor_sub=c_sub;
					cursor_id1=c_id1;
					cursor_id2=c_id2;
					memmove(cursor_line, c_line, sizeof cursor_line);
					cursorline_dirty();
					cursor_ypos=c_ypos;
					cursor_tag_idx=c_tag_idx;
					cursor_tag_group=c_tag_group;
					if (cursor_ypos>=size.h) scroll_down(cursor_ypos-size.h+1);
					update_misc_info();
					update_visual_info();
					if (edit) update_micropos();
				} else if (nls) scroll_down(1);
				return r;
			} else {
				int r=0;
				char c_line[1024];
				int c_tag_idx=cursor_tag_idx;
				int c_tag_group=cursor_tag_group;
				int d_tag_group=cursor_tag_group;
				hts_sub *c_sub;
				ID c_id1;
				ID c_id2;
				int c_ypos;
				if (cursor_state==cursor_state_invisible) {
					c_sub=top_sub;
					c_id1=top_id1;
					c_id2=top_id2;
					c_ypos=next_line(&c_sub, &c_id1, &c_id2, size.h-1);
				} else {
					c_sub=cursor_sub;
					c_id1=cursor_id1;
					c_id2=cursor_id2;
					c_ypos=cursor_ypos;
				}

				int nc=next_line(&c_sub, &c_id1, &c_id2, n);
				int onc=c_ypos+nc-size.h+1;
				c_ypos+=nc;

				while (nc--) {
					c_sub->getline(c_line, c_id1, c_id2);
					int g=tag_count_groups(c_line);
					if (d_tag_group<g) c_tag_group=d_tag_group;
					int s;
					if (c_tag_group>=g) {
						c_tag_group=g-1;
						s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
						if (s) {
							c_tag_idx=s-1;
							r=1;
							break;
						}
					} else {
						s=tag_count_selectable_tags_in_group(c_line, c_tag_group);
						if (s) {
							if (c_tag_idx>=s) c_tag_idx=s-1;
							r=1;
							break;
						}
					}
					if (!prev_line(&c_sub, &c_id1, &c_id2, 1)) break;
					c_ypos--;
				}
				if (r) {
					cursor_sub=c_sub;
					cursor_id1=c_id1;
					cursor_id2=c_id2;
					memmove(cursor_line, c_line, sizeof cursor_line);
					cursorline_dirty();
					cursor_ypos=c_ypos;
					cursor_tag_idx=c_tag_idx;
					cursor_tag_group=c_tag_group;
					if (cursor_ypos-size.h+1>0) scroll_down(cursor_ypos-size.h+1);
					update_misc_info();
					update_visual_info();
					if (edit) update_micropos();
				} else if (onc>0) scroll_down(onc);
//				} else scroll_down(n);
				break;
			}
		}
		case cursor_state_disabled:
			scroll_down(n);
			return n;
	}
	return 0;
}

int ht_uviewer::cursor_home()
{
	if (cursor_tag_idx) {
		cursor_tag_idx=0;
		if (edit) cursor_tag_micropos=0;
		update_visual_info();
		update_misc_info();
		return 1;
	}
	return 0;
}

int ht_uviewer::cursor_end()
{
	cursorline_get();
	int c=tag_count_selectable_tags_in_group(cursor_line, cursor_tag_group);
	if (cursor_tag_idx!=c-1) {
		cursor_tag_idx=c-1;
		if (edit) {
			char *e=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
			if (e) cursor_tag_micropos=tag_get_microsize(e)-1;
		}
		update_visual_info();
		update_misc_info();
		return 1;
	}
	return 0;
}

void ht_uviewer::cursor_tab()
{
	cursor_tag_group++;
	adjust_cursor_group();
}

void ht_uviewer::cursorline_dirty()
{
	isdirty_cursor_line=1;
}

void ht_uviewer::cursorline_get()
{
	if (isdirty_cursor_line) {
		cursor_sub->getline(cursor_line, cursor_id1, cursor_id2);
		isdirty_cursor_line=0;
	}
}

int ht_uviewer::cursormicroedit_forward()
{
	cursorline_get();
	char *t=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	int s=tag_get_microsize(t);
	if (cursor_tag_micropos+1>=s) {
		hts_sub *sub=cursor_sub;
		ID id1=cursor_id1;
		ID id2=cursor_id2;
		int tag_idx=cursor_tag_idx+1;
		char c_line[1024];  /* FIXME: possible buffer overflow ! */
		int g=cursor_tag_group;
		for (int y=0; y<size.h; y++) {
			if (!sub->getline(c_line, id1, id2)) break;
			int c=tag_count_selectable_tags_in_group(c_line, cursor_tag_group);
			while (tag_idx<c) {
				char *t=tag_get_selectable_tag(c_line, tag_idx, cursor_tag_group);
				if (tag_get_class(t)==tag_class_edit) {
					cursor_tag_micropos=0;
					set_cursor(sub, id1, id2);
					cursorline_dirty();
					cursor_tag_group=g;
					cursor_tag_idx=tag_idx;
					update_misc_info();
					update_visual_info();
					return 1;
				}
				tag_idx++;
			}
			tag_idx=0;
			if (!next_line(&sub, &id1, &id2, 1)) break;
		}
		if (cursor_right()) cursor_tag_micropos=0;
	} else cursor_tag_micropos++;
	return 0;
}

int ht_uviewer::cursormicro_forward()
{
	cursorline_get();
	char *t=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	int s=tag_get_microsize(t);
	if (cursor_tag_micropos+1>=s) {
		if (cursor_right()) cursor_tag_micropos=0; else return 0;
	} else cursor_tag_micropos++;
	return 1;
}

int ht_uviewer::cursormicro_backward()
{
	if (cursor_tag_micropos>0) {
		cursor_tag_micropos--;
	} else {
		if (cursor_left()) {
			cursorline_get();
			char *t=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
			cursor_tag_micropos=tag_get_microsize(t)-1;
		} else return 0;
	}
	return 1;
}

void ht_uviewer::draw()
{
	int max_sdown_count=0;
	if (!first_sub) return;
	if (!initialized) ERROR("complete_init() not called !");
restart:
	clear(windialog_color_body);
	ID id1=top_id1, id2=top_id2;
	hts_sub *sub=top_sub;
	ID pid1, pid2;
	hts_sub *psub=0;
	char line[1024];	/* FIXME: possible buffer overflow ! */
	int is_cursor, cursor_found=-1;
	int coloridx;
	if (focused) hidecursor();
	for (int y=0; y<size.h; y++) {
		if (y==cursor_ypos) {
			pid1=id1;
			pid2=id2;
			psub=sub;
		}
		if (!sub->getline(line, id1, id2)) break;
		if ((cursor_state==cursor_state_visible) && (sub==cursor_sub) &&
		(id1==cursor_id1) && (id2==cursor_id2)) {
			cursor_found=y;
			cursorline_dirty();
			update_misc_info();
			int c=tag_count_selectable_tags_in_group(line, cursor_tag_group);
			if (cursor_tag_idx>=c) cursor_tag_idx=c-1;
			char *t=tag_get_selectable_tag(line, cursor_tag_idx, cursor_tag_group);
			is_cursor=1;
			switch (cursor_tag_class) {
				case tag_class_edit:
					if (edit) {
						coloridx=tag_coloridx_edit_cursor_edit;
						int x=cursor_visual_xpos+tag_get_micropos(t, cursor_tag_micropos)-xscroll;
						if (focused) if (x>=0) setcursor(x, y);
					} else {
						coloridx=tag_coloridx_edit_cursor_sel;
					}
					break;
				case tag_class_sel:
					coloridx=tag_coloridx_sel_cursor;
					break;
				default:
					coloridx=tag_coloridx_normal;
					break;
			}
		} else is_cursor=0;
		print_tagstring(0, y, size.w, xscroll, line, is_cursor, coloridx);
		if (!next_line(&sub, &id1, &id2, 1)) break;
	}
	if ((cursor_state==cursor_state_visible) && (cursor_found==-1)) {
/*		scroll_down(1);
		max_sdown_count++;
		if (max_sdown_count>=size.h) assert(0);
		goto restart;*/
		if (!psub) {
			pid1=top_id1;
			pid2=top_id2;
			psub=top_sub;
		}
		cursor_id1=pid1;
		cursor_id2=pid2;
		cursor_sub=psub;
		cursorline_dirty();
		update_misc_info();
		int c=tag_count_selectable_tags_in_group(line, cursor_tag_group);
		if (cursor_tag_idx>c-1) cursor_tag_idx=c-1;

		if (max_sdown_count++) {

		} else goto restart;
	}
	if (cursor_found!=-1) cursor_ypos=cursor_found;
//     printf(1, 1, VCP(VC_GREEN, VC_BLACK), "%s", focused ? "focused" : "unfocused");
//     printf(20, 6, 7, "csize.x=%2d, csize.y=%2d, csize.w=%2d, csize.h=%2d", csize.x, csize.y, csize.w, csize.h);
//     printf(20, 7, 7, "size.x=%2d, size.y=%2d, size.w=%2d, size.h=%2d", size.x, size.y, size.w, size.h);
//     printf(20, 8, 7, "vsize.x=%2d, vsize.y=%2d, vsize.w=%2d, vsize.h=%2d", vsize.x, vsize.y, vsize.w, vsize.h);
//     printf(20, 7, 7, "size.x=%2d, size.y=%2d, size.w=%2d, size.h=%2d", group->group->size.x, group->group->size.y, group->group->size.w, group->group->size.h);
//     printf(20, 8, 7, "vsize.x=%2d, vsize.y=%2d, vsize.w=%2d, vsize.h=%2d", group->group->vsize.x, group->group->vsize.y, group->group->vsize.w, group->group->vsize.h);
//     printf(0, 7, 7, "vx=%2d, vl=%2d, ypos=%2d", cursor_visual_xpos, cursor_visual_length, cursor_ypos);
//     printf(0, 7, 7, "cursor_id1=%x, cursor_id2=%x", cursor_id1, cursor_id2);
//     printf(0, 6, 7, "cursor_tag_idx=%d", cursor_tag_idx);
/*     if (cursor_tag_class==tag_class_edit) {
		printf(30, 7, 7, "class=edit, addr=%08x", cursor_tag_offset);
	} else {
		printf(30, 7, 7, "class=sel, id_low=%08x, id_high=%08x", cursor_tag_id.low, cursor_tag_id.high);
	}*/
}

void ht_uviewer::edit_end()
{
	if (edit) {
		stream->invalidate();
		edit=0;
		stream->set_access_mode(SAM_READ);
		hidecursor();
		if (find_first_tag(&cursor_sub, &cursor_id1, &cursor_id2, size.h)) {
			set_cursor(cursor_sub, cursor_id1, cursor_id2);
		} else {
			if (cursor_state!=cursor_state_disabled) {
/* is that all we have to do ??? testing needed */
				cursor_state=cursor_state_invisible;
				assert(0);	/* FIXME: not yet implemented */
			}
		}
		dirtyview();
	}
}

void ht_uviewer::edit_input(byte b)
{
	cursorline_get();
	char *t=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	switch (t[1]) {
		case HT_TAG_EDIT_BYTE: {
			int nibval=edit_input_c2h(b);
			if (nibval==-1) break;
			int shift=(8-4-cursor_tag_micropos*4);
			byte v, m=~(0xf<<shift), o=nibval<<shift;
			stream->seek(cursor_tag_offset);
			stream->read(&v, 1);
			v&=m;
			v|=o;
			stream->seek(cursor_tag_offset);
			stream->write(&v, 1);
			edit_input_correctpos();
			cursormicroedit_forward();
			break;
		}
		case HT_TAG_EDIT_WORD:{
			int nibval=edit_input_c2h(b);
			if (nibval==-1) break;
			int shift=(16-4-cursor_tag_micropos*4);
			word v, m=~(0xf<<shift), o=nibval<<shift;
			stream->seek(cursor_tag_offset);
			stream->read(&v, 2);
			v&=m;
			v|=o;
			stream->seek(cursor_tag_offset);
			stream->write(&v, 2);
			cursormicroedit_forward();
			break;
		}
		case HT_TAG_EDIT_DWORD:{
			int nibval=edit_input_c2h(b);
			if (nibval==-1) break;
			int shift=(32-4-cursor_tag_micropos*4);
			dword v, m=~(0xf<<shift), o=nibval<<shift;
			stream->seek(cursor_tag_offset);
			stream->read(&v, 4);
			v&=m;
			v|=o;
			stream->seek(cursor_tag_offset);
			stream->write(&v, 4);
			cursormicroedit_forward();
			break;
		}
		case HT_TAG_EDIT_CHAR: {
			if ((b>=32) && (b<=0xff)) {
				stream->seek(cursor_tag_offset);
				stream->write(&b, 1);
				cursormicroedit_forward();
			}
			break;
		}
		case HT_TAG_EDIT_BIT: {
			if ((b=='1') || (b=='0') || b==' ') {
				byte d;
				cursorline_get();
				char *t=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
				int shift=((ht_tag_edit_bit*)t)->bitidx;
				dword mask=1 << (shift%8);
				int op=shift/8;
				stream->seek(cursor_tag_offset+op);
				stream->read(&d, 1);
				switch (b) {
					case '0':
						d&=~mask;
						break;
					case '1':
						d|=mask;
						break;
					case ' ':
						d^=mask;
						break;
				}
				stream->seek(cursor_tag_offset+op);
				stream->write(&d, 1);
				cursormicroedit_forward();
			}
			break;
		}
	}
}

int ht_uviewer::edit_input_c2h(byte b)
{
	int h=-1;
	if ((b>='0') && (b<='9')) {
		h=b-'0';
	} else if ((b>='a') && (b<='f')) {
		h=b-'a'+10;
	} else if ((b>='A') && (b<='F')) {
		h=b-'A'+10;
	}
	return h;
}

void ht_uviewer::edit_input_correctpos()
{
	hts_sub *sub;
	ID id1, id2;

/* try finding edited tag by its offset */
	sub=top_sub;
	id1=top_id1;
	id2=top_id2;
	char c_line[1024];  /* FIXME: possible buffer overflow ! */
	int g=cursor_tag_group;
	for (int y=0; y<size.h+10; y++) {
		if (!sub->getline(c_line, id1, id2)) break;
		int c=tag_count_selectable_tags_in_group(c_line, cursor_tag_group);
		for (int i=0; i<c; i++) {
			char *t=tag_get_selectable_tag(c_line, i, cursor_tag_group);
			if ((tag_get_class(t)==tag_class_edit) && (tag_get_offset(t)==cursor_tag_offset)) {
				set_cursor(sub, id1, id2);
				cursorline_dirty();
				cursor_tag_idx=i;
				cursor_tag_group=g;
				update_misc_info();
				update_visual_info();
				return;
			}
		}
		if (!next_line(&sub, &id1, &id2, 1)) break;
	}

/* try finding edited tag by cursor_ypos */
	sub=top_sub;
	id1=top_id1;
	id2=top_id2;
	next_line(&sub, &id1, &id2, cursor_ypos);
	int ci=cursor_tag_idx;
	set_cursor(sub, id1, id2);
	cursorline_dirty();
	cursorline_get();
	int x=tag_count_selectable_tags(cursor_line);
	if (ci>=x) ci=x-1;
	cursor_tag_idx=ci;
	update_misc_info();
	update_visual_info();
}

void ht_uviewer::edit_start()
{
	if (!edit) {
		if (stream->set_access_mode(SAM_READ | SAM_WRITE)) {
			edit=1;
			cursor_tag_micropos=0;
			cursorline_dirty();
			adjust_cursor_idx();
			update_visual_info();
			update_misc_info();
			dirtyview();
		} else errorbox("can't open file %s in write mode ! (error %08x)", stream->get_name(), stream->get_error());
	}
}

void ht_uviewer::edit_update()
{
	if (edit) {
		stream->flush();
		dirtyview();
	}
}

int ht_uviewer::find_first_tag(hts_sub **_sub, ID *_id1, ID *_id2, int limit)
{
	char line[1024];	/* FIXME: possible buffer overflow ! */
	int i=0;
	hts_sub *sub=(*_sub);
	ID id1=*_id1, id2=*_id2;
	do {
		sub->getline(line, id1, id2);
		if (tag_get_selectable_tag(line, 0, 0)) {
			*_sub=sub;
			*_id1=id1;
			*_id2=id2;
			return 1;
		}
	} while ((next_line(&sub, &id1, &id2, 1)) && (i++<limit));
	return 0;
}

int ht_uviewer::find_first_edit_tag_with_offset(hts_sub **_sub, ID *_id1, ID *_id2, int *tag_idx, int limit, STREAMOFS offset)
{
	char line[1024];	/* FIXME: possible buffer overflow ! */
	int i=0;
	hts_sub *sub=(*_sub);
	ID id1=*_id1, id2=*_id2;
	do {
		sub->getline(line, id1, id2);
		int c=tag_count_selectable_tags(line);
		char *t=line;
		for (int j=0; j<c; j++) {
			t=tag_get_selectable_tag(t, 0, -1);
			if ((tag_get_class(t)==tag_class_edit) && (tag_get_offset(t)==offset)) {
				*tag_idx=j;	/* FIXME: what about groups ??? */
				*_sub=sub;
				*_id1=id1;
				*_id2=id2;
				return 1;
			}
			t+=tag_get_len(t);
		}
	} while ((next_line(&sub, &id1, &id2, 1)) && (i++<limit));
	return 0;
}

void ht_uviewer::focus_cursor()
{
	if (cursor_state==cursor_state_invisible) {
		center_view(cursor_sub, cursor_id1, cursor_id2);
		cursor_state=cursor_state_visible;
	}
	update_visual_info();
	if (cursor_visual_xpos+cursor_visual_length-xscroll>size.w) {
		xscroll=cursor_visual_xpos+cursor_visual_length-size.w;
	} else if (cursor_visual_xpos-xscroll<0) {
		xscroll=cursor_visual_xpos;
	}
}

char *ht_uviewer::func(int i, int getname)
{
	switch (i) {
		case 4:
			if (caps & VC_EDIT) {
				if (edit) {
					if (getname) return "view"; else sendmsg(htm_view_mode, 0);
				} else {
					if (getname) return "edit"; else sendmsg(htm_edit_mode, 0);
				}
			}
			break;
		case 5:
			if (caps & VC_GOTO) {
				if (getname) return "goto"; else jump_dialog(this);
			}
			break;
		case 7:
			if (caps & VC_SEARCH) {
				if (getname) return "search"; else
					if (cursor_tag_class==tag_class_edit) {
						search_dialog(this, cursor_tag_offset);
					} else {
						VIEWERADDR a;
						if (cursor_sub->convert_id_to_addr(cursor_id1, cursor_id2, &a)) {
							STREAMOFS o;
							if (convert_addr_to_ofs(a, &o)) {
								search_dialog(this, o);
							} else errorbox("can't search...(internal error)"); /* FIXME: internal error */
						} else errorbox("can't search here"); /* FIXME: start at 0 ? */
					}
			}
			break;
		case 9:
			if (caps & VC_EDIT) {
				if (edit) {
					if (getname) return "update"; else
						edit_update();
				} else {
					if (getname) return "~update"; else return 0;
				}
			}
			break;
	}
	return ht_viewer::func(i, getname);
}

int ht_uviewer::get_current_offset(STREAMOFS *offset)
{
	if (cursor_tag_class==tag_class_edit) {
		*offset=cursor_tag_offset;
		return 1;
	}
	return 0;
}

int ht_uviewer::get_current_tag(char **tag)
{
	cursorline_get();
	char *e=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	if (e) {
		*tag=e;
		return 1;
	}
	return 0;
}

int ht_uviewer::get_current_tag_size(dword *size)
{
	if (cursor_tag_class==tag_class_edit) {
		cursorline_get();
		char *e=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
		if (e) {
			*size=tag_get_size(e);
			return 1;
		}
	}
	return 0;
}

void ht_uviewer::handlemsg(htmsg *msg)
{
	switch (msg->msg) {
		case htm_complete_init:
			complete_init();
			clearmsg(msg);
			return;
		case htm_keypressed:
			switch (msg->data_int) {
				case K_Up:
					cursor_up(1);
					clearmsg(msg);
					dirtyview();
					return;
				case K_Down:
					cursor_down(1);
					clearmsg(msg);
					dirtyview();
					return;
				case K_PageUp:
					cursor_up(size.h);
					clearmsg(msg);
					dirtyview();
					return;
				case K_PageDown:
					cursor_down(size.h);
					clearmsg(msg);
					dirtyview();
					return;
				case K_Left:
					if (cursor_state!=cursor_state_disabled) {
						if (edit) cursormicro_backward(); else cursor_left();
						focus_cursor();
						clearmsg(msg);
						dirtyview();
						return;
					}
					break;
				case K_Right:
					if (cursor_state!=cursor_state_disabled) {
						if (edit) cursormicro_forward(); else cursor_right();
						focus_cursor();
						clearmsg(msg);
						dirtyview();
						return;
					}
					break;
				case K_Space:
					switch (cursor_tag_class) {
						case tag_class_sel:
							ref();
							break;
						case tag_class_edit:
							if (edit) edit_input(msg->data_int);
							break;
					}
					dirtyview();
					clearmsg(msg);
					return;
				case K_BackSpace:
					pop_vs_history();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Tab:
					cursor_tab();
					focus_cursor();
					update_visual_info();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Home:
					if (cursor_state!=cursor_state_disabled) {
						cursor_home();
						focus_cursor();
						clearmsg(msg);
						dirtyview();
						return;
					}
					break;
				case K_End:
					if (cursor_state!=cursor_state_disabled) {
						cursor_end();
						focus_cursor();
						clearmsg(msg);
						dirtyview();
						return;
					}
					break;
				case K_Control_Left:
					xscroll-=2;
					if (xscroll<0) xscroll=0;
					dirtyview();
					clearmsg(msg);
					return;
				case K_Control_Right:
					xscroll+=2;
					dirtyview();
					clearmsg(msg);
					return;
				case K_Control_PageUp: {
					top_sub=first_sub;
					top_sub->first_line_id(&top_id1, &top_id2);

					hts_sub *sub=top_sub;
					ID id1=top_id1, id2=top_id2;
					if (find_first_tag(&sub, &id1, &id2, size.h)) {
						set_cursor(sub, id1, id2);
					} else {
						cursor_ypos=0x7fffffff;
						update_misc_info();
						update_visual_info();
						check_cursor_visibility();
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Control_PageDown: {
					top_sub=last_sub;
					top_sub->last_line_id(&top_id1, &top_id2);

					hts_sub *sub=top_sub;
					ID id1=top_id1, id2=top_id2;
					if (find_first_tag(&sub, &id1, &id2, 1)) {
						set_cursor(sub, id1, id2);
					} else {
						cursor_ypos=0x7fffffff;
						update_misc_info();
						update_visual_info();
						check_cursor_visibility();
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_Up: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					cursor_up(1);
					if (s!=0xffffffff) {
						if (cursor_tag_class==tag_class_edit) {
							select_add(s, cursor_tag_offset);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(s, e);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_Down: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					cursor_down(1);
					if (s!=0xffffffff) {
						if (cursor_tag_class==tag_class_edit) {
							select_add(s, cursor_tag_offset);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(s, e);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_Left: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					int r;
					if (edit) r=cursormicro_backward(); else r=cursor_left();
					if ((s!=0xffffffff) & (r)) {
						if (cursor_tag_class==tag_class_edit) {
							select_add(s, cursor_tag_offset);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(s, e);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_Right: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					int r;
					if (edit) r=cursormicro_forward(); else r=cursor_right();
					if (s!=0xffffffff) {
						if (r) {
							if (cursor_tag_class==tag_class_edit) {
								select_add(s, cursor_tag_offset);
							} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
								select_add(s, e);
							}
						} else {
							if (cursor_tag_class==tag_class_edit) {
								select_add(s, e);
							}
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_PageUp: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					cursor_up(size.h);
					if (s!=0xffffffff) {
						if (cursor_tag_class==tag_class_edit) {
							select_add(s, cursor_tag_offset);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(s, e);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_PageDown: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					cursor_down(size.h);
					if (s!=0xffffffff) {
						if (cursor_tag_class==tag_class_edit) {
							select_add(s, cursor_tag_offset);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(s, e);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_Home: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					int r=cursor_home();
					if ((s!=0xffffffff) && (r)) {
						if (cursor_tag_class==tag_class_edit) {
							select_add(cursor_tag_offset, s);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(e, s);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Shift_End: {
					focus_cursor();
					STREAMOFS s, e;
					dword ts;
					if (cursor_tag_class==tag_class_edit)
						s=cursor_tag_offset;
					else
						s=(sel_end>sel_start) ? sel_end : 0xffffffff;
					e=get_current_tag_size(&ts) ? s+ts : 0xffffffff;
					int r=cursor_end();
					if ((s!=0xffffffff) && (r)){
						if (cursor_tag_class==tag_class_edit) {
							select_add(s, cursor_tag_offset);
						} else if ((cursor_tag_class==tag_class_sel) && (e!=0xffffffff)) {
							select_add(s, e);
						}
					}
					clearmsg(msg);
					dirtyview();
					return;
				}
				case K_Escape:
					if (edit) {
						sendmsg(htm_view_mode, 0);
						return;
					}
					break;
				case K_Shift_F7:
					continue_search();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Control_Insert:
					sendmsg(htm_command, cmd_clipboard_copy);
					return;
				case K_Shift_Insert:
					sendmsg(htm_command, cmd_clipboard_paste);
					return;
				default: {
					if (((dword)msg->data_int<=255) && (edit)) {
						if (cursor_tag_class==tag_class_edit) {
							focus_cursor();
							edit_input(msg->data_int);
							dirtyview();
							return;
						}
					}
				}
			}
			break;
		case htm_edit_mode:
			edit_start();
			clearmsg(msg);
			dirtyview();
			return;
		case htm_view_mode:
			edit_end();
			clearmsg(msg);
			dirtyview();
			return;
		case htm_command:
			switch (msg->data_int) {
				case cmd_clipboard_copy:
					if (sel_end>sel_start) {
						char dsc[1024]; /* FIXME: possible buffer overflow */
						sprintf(dsc, "%s::%s", stream->get_name(), desc);
						clipboard_copy(dsc, stream, sel_start, sel_end-sel_start);
					}
					clearmsg(msg);
					return;
				case cmd_clipboard_paste:
					STREAMOFS ofs;
					if (get_current_offset(&ofs)) {
						if (clipboard_paste(stream, ofs)) {
							edit_start();
						}
					}
					dirtyview();
					clearmsg(msg);
					return;
			}
			break;
	}
	ht_viewer::handlemsg(msg);
}

void ht_uviewer::insertsub(hts_sub *sub)
{
	if (last_sub) last_sub->next=sub;
	sub->prev=last_sub;
	last_sub=sub;
	if (!first_sub) first_sub=sub;
}

int ht_uviewer::jump(VIEWERADDR addr, ht_view *focused_to_history)
{
	ID id1, id2;
	hts_sub *sub=first_sub;
	while (sub) {
		if (sub->convert_addr_to_id(addr, &id1, &id2)) {
			int tag_idx;
/* FIXME: hack... (ADDR as param to find_first_tag_with_OFFSET) ! */
			if (find_first_edit_tag_with_offset(&sub, &id1, &id2, &tag_idx, size.h, addr)) {
				if (focused_to_history) push_vs_history(focused_to_history);
				set_cursor_ex(sub, id1, id2, tag_idx, 0);
				return 1;
			} else if (find_first_tag(&sub, &id1, &id2, size.h)) {
				if (focused_to_history) push_vs_history(focused_to_history);
				set_cursor_ex(sub, id1, id2, 0, 0);
				return 1;
			}
			break;
		}
		sub=sub->next;
	}
	return 0;
}

int ht_uviewer::next_line(hts_sub **sub, ID *id1, ID *id2, int n)
{
	int in=n;
	while (n-=(*sub)->next_line_id(id1, id2, n)) {
		if (!(*sub)->next) return in-n;
		*sub=(*sub)->next;
		(*sub)->first_line_id(id1, id2);
		n--;
	}
	return in-n;
}

int ht_uviewer::prev_line(hts_sub **sub, ID *id1, ID *id2, int n)
{
	int in=n;
	while (n-=(*sub)->prev_line_id(id1, id2, n)) {
		if (!(*sub)->prev) return in-n;
		*sub=(*sub)->prev;
		(*sub)->last_line_id(id1, id2);
		n--;
	}
	return in-n;
}

void ht_uviewer::print_tagstring(int x, int y, int maxlen, int xscroll, char *tagstring, int iscursor, int coloridx)
{
	char *n=tagstring;
	int c=0;
	int is_cursor, i=0, g=0;
	palette *tag_pal=pal_grp->get(tag_palette_str);
	int normal_color=tag_pal[tag_coloridx_normal].value;
	do {
		int l=0;
		while (n[l] && n[l]!='\e') { l++; }
		c+=print_tagstring_final(x+c-xscroll, y, n, l, normal_color);
		n+=l;
		is_cursor=(i==cursor_tag_idx) && (g==cursor_tag_group) && iscursor;
		if (*n=='\e') {
			STREAMOFS tag_offset;
			vc tag_color;
			char str[32];
			switch (n[1]) {
				case HT_TAG_EDIT_BYTE: {
					byte d;
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;
					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					} else if (stream->isdirty(tag_offset, 1)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_modified].value);
					}
					if (is_cursor) tag_color=vcp_mix(tag_color, tag_pal[coloridx].value);
					stream->seek(tag_offset);
					stream->read(&d, 1);
					mkhexb(str, d);
					c+=print_tagstring_final(x+c-xscroll, y, str, 2, tag_color);
					n+=HT_TAG_EDIT_BYTE_LEN;
					break;
				}
				case HT_TAG_EDIT_WORD: {
					dword d;
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;
					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					} else if (stream->isdirty(tag_offset, 2)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_modified].value);
					}
					if (is_cursor) tag_color=vcp_mix(tag_color, tag_pal[coloridx].value);
					stream->seek(tag_offset);
					stream->read(&d, 2);
					mkhexw(str, d);
					c+=print_tagstring_final(x+c-xscroll, y, str, 4, tag_color);
					n+=HT_TAG_EDIT_WORD_LEN;
					break;
				}
				case HT_TAG_EDIT_DWORD: {
					dword d;
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;
					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					} else if (stream->isdirty(tag_offset, 4)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_modified].value);
					}
					if (is_cursor) tag_color=vcp_mix(tag_color, tag_pal[coloridx].value);
					stream->seek(tag_offset);
					stream->read(&d, 4);
					mkhexd(str, d);
					c+=print_tagstring_final(x+c-xscroll, y, str, 8, tag_color);
					n+=HT_TAG_EDIT_DWORD_LEN;
					break;
				}
				case HT_TAG_EDIT_TIME: {
					dword d;
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;
					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					} else if (stream->isdirty(tag_offset, 4)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_modified].value);
					}
					if (is_cursor) tag_color=vcp_mix(tag_color, tag_pal[coloridx].value);
					stream->seek(tag_offset);
					stream->read(&d, 4);
					tm *t=gmtime((time_t*)&d);
					sprintf(str, "%02d:%02d:%02d %02d.%02d.%04d", t->tm_hour, t->tm_min, t->tm_sec, t->tm_mday, t->tm_mon+1, t->tm_year+1900);
					c+=print_tagstring_final(x+c-xscroll, y, str, strlen(str), tag_color);
					n+=HT_TAG_EDIT_TIME_LEN;
					break;
				}
				case HT_TAG_EDIT_CHAR: {
					char d;
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;
					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					} else if (stream->isdirty(tag_offset, 1)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_modified].value);
					}
					if (is_cursor) tag_color=vcp_mix(tag_color, tag_pal[coloridx].value);
					stream->seek(tag_offset);
					stream->read(&d, 1);
					if (!d) d=' ';
					c+=print_tagstring_final(x+c-xscroll, y, &d, 1, tag_color);
					n+=HT_TAG_EDIT_CHAR_LEN;
					break;
				}
				case HT_TAG_EDIT_BIT: {
					int shift=((ht_tag_edit_bit*)n)->bitidx;
					int op=shift/8;
					byte d;
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;
					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					} else if (stream->isdirty(tag_offset+op, ISDIRTY_SINGLEBIT | (shift & 7))) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_modified].value);
					}
					if (is_cursor) tag_color=vcp_mix(tag_color, tag_pal[coloridx].value);
					stream->seek(tag_offset+op);
					stream->read(&d, 1);
					str[0]=(d& (1 << (shift%8))) ? '1' : '0';
					str[1]=0;
					c+=print_tagstring_final(x+c-xscroll, y, str, 1, tag_color);
					n+=HT_TAG_EDIT_BIT_LEN;
					break;
				}
				case HT_TAG_EDIT_SELVIS: {
					tag_offset=tag_get_offset(n);
					tag_color=tag_pal[tag_coloridx_edit_normal].value;

					if ((tag_offset>=sel_start) && (tag_offset<sel_end)) {
						tag_color=vcp_mix(tag_color, tag_pal[tag_coloridx_edit_selected].value);
					}

					c+=print_tagstring_final(x+c-xscroll, y, &((ht_tag_edit_selvis*)n)->ch, 1, tag_color);
					n+=HT_TAG_EDIT_SELVIS_LEN;
					continue;
				}
				case HT_TAG_SEL: {
					int tag_textlen=tag_get_seltextlen(n);
					tag_color=tag_pal[tag_coloridx_sel_normal].value;
					if (is_cursor) tag_color=tag_pal[coloridx].value;
					n+=HT_TAG_SEL_LEN(0);
					c+=print_tagstring_final(x+c-xscroll, y, n, tag_textlen, tag_color);
					n+=tag_textlen;
					break;
				}
				case HT_TAG_DESC_BYTE: {
					ID id;
					tag_color=tag_pal[tag_coloridx_sel_normal].value;
					if (is_cursor) tag_color=tag_pal[coloridx].value;
					if (tag_get_desc_id(n, &id)) {
						int_hash *tbl;
						if ((tbl=(int_hash*)find_atom(id))) {
							char *str;
							byte d;
							tag_offset=tag_get_offset(n);
							stream->seek(tag_offset);
							stream->read(&d, 1);
							if ((str=matchhash(d, tbl))) {
								int l=strlen(str);
								c+=print_tagstring_final(x+c-xscroll, y, str, l, tag_color);
							} else c+=print_tagstring_final(x+c-xscroll, y, "?", 1, tag_color);
						}
					}
					n+=HT_TAG_DESC_WORD_LEN;
					break;
				}
				case HT_TAG_DESC_WORD: {
					ID id;
					tag_color=tag_pal[tag_coloridx_sel_normal].value;
					if (is_cursor) tag_color=tag_pal[coloridx].value;
					if (tag_get_desc_id(n, &id)) {
						int_hash *tbl;
						if ((tbl=(int_hash*)find_atom(id))) {
							char *str;
							word d;
							tag_offset=tag_get_offset(n);
							stream->seek(tag_offset);
							stream->read(&d, 2);
							if ((str=matchhash(d, tbl))) {
								int l=strlen(str);
								c+=print_tagstring_final(x+c-xscroll, y, str, l, tag_color);
							} else c+=print_tagstring_final(x+c-xscroll, y, "?", 1, tag_color);
						}
					}
					n+=HT_TAG_DESC_WORD_LEN;
					break;
				}
				case HT_TAG_DESC_DWORD: {
					ID id;
					tag_color=tag_pal[tag_coloridx_sel_normal].value;
					if (is_cursor) tag_color=tag_pal[coloridx].value;
					if (tag_get_desc_id(n, &id)) {
						int_hash *tbl;
						if ((tbl=(int_hash*)find_atom(id))) {
							char *str;
							dword d;
							tag_offset=tag_get_offset(n);
							stream->seek(tag_offset);
							stream->read(&d, 4);
							if ((str=matchhash(d, tbl))) {
								int l=strlen(str);
								c+=print_tagstring_final(x+c-xscroll, y, str, l, tag_color);
							} else c+=print_tagstring_final(x+c-xscroll, y, "?", 1, tag_color);
						}
					}
					n+=HT_TAG_DESC_DWORD_LEN;
					break;
				}
				case HT_TAG_FLAGS:
					n+=HT_TAG_FLAGS_LEN;
					tag_color=tag_pal[tag_coloridx_sel_normal].value;
					if (is_cursor) tag_color=tag_pal[coloridx].value;
					c+=print_tagstring_final(x+c-xscroll, y, "details", 7, tag_color);
					break;
				case HT_TAG_GROUP:
					n+=HT_TAG_GROUP_LEN;
					g++;
					i=0;
					continue;
				case HT_TAG_COLOR:
					normal_color=((ht_tag_color*)n)->color;
					n+=HT_TAG_COLOR_LEN;
					continue;
				default:
					assert(0);
			}
		}
		i++;
	} while (*n);
}

int ht_uviewer::print_tagstring_final(int x, int y, char *string, int len, int color)
{
	int l=len;
	while (x<0) {
		string++;
		len--;
		x++;
	}
	if (len>0) {
		if (x+len>size.w) len=size.w-x;
		lprint(x, y, color, len, string);
	}
	return l;
}

int ht_uviewer::ref()
{
	cursorline_get();
	char *e=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	if (!e) return 0;
	if (tag_get_class(e)==tag_class_sel) {
		switch (e[1]) {
			case HT_TAG_SEL:
				return ref_sel(cursor_tag_id.low, cursor_tag_id.high);
			case HT_TAG_FLAGS:
				return ref_flags(((ht_tag_flags*)e)->offset, ((ht_tag_flags*)e)->id);
			case HT_TAG_DESC_BYTE:
				return ref_desc(((ht_tag_desc_byte*)e)->id, ((ht_tag_desc_byte*)e)->offset, 1);
			case HT_TAG_DESC_WORD:
				return ref_desc(((ht_tag_desc_byte*)e)->id, ((ht_tag_desc_byte*)e)->offset, 2);
			case HT_TAG_DESC_DWORD:
				return ref_desc(((ht_tag_desc_byte*)e)->id, ((ht_tag_desc_byte*)e)->offset, 4);
		}
	}
	return 0;
}

int ht_uviewer::ref_desc(ID id, STREAMOFS offset, int size)
{
	int_hash *desc=(int_hash*)find_atom(id);
	if (desc) {
		bounds b;
		b.w=60;
		b.h=14;
		b.x=(screen->size.w-b.w)/2;
		b.y=(screen->size.h-b.h)/2;
		ht_dialog *d=new ht_dialog();
		d->init(&b, "desc", FS_KILLER);

		b.x=0;
		b.y=0;
		b.w-=2;
		b.h-=2;
		ht_text_listbox *l=new ht_text_listbox();
		l->init(&b, 2, 1);

		int curvalue=0;
		int curpos=0, i=0;

		stream->seek(offset);
		stream->read(&curvalue, size);

		int_hash *dsc=desc;
		while (dsc->desc) {
			char buf[1024]; /* FIXME: possible buffer overflow */
			switch (size) {
				case 1:
					sprintf(buf, "0x%02x", dsc->value);
					break;
				case 2:
					sprintf(buf, "0x%04x", dsc->value);
					break;
				case 4:
					sprintf(buf, "0x%08x", dsc->value);
					break;
			}
			if (dsc->value==curvalue) {
				curpos=i;
			}
			l->insert_str(i, buf, dsc->desc);
			dsc++;
			i++;
		}

		l->update();
		l->seek(curpos);

		d->insert(l);

		if (execute_dialog(d)==button_ok) {
			ht_listbox_data da;
			l->getdata(&da);
			int i=((ht_text_listbox_item*)da.cursor)->id;
			if (i!=curpos) {
				edit_start();
				stream->seek(offset);
				stream->write(&desc[i].value, size);
				dirtyview();
			}
		}

		d->done();
		delete d;
		return 1;
	}
	return 0;
}

int ht_uviewer::ref_flags(ID id, STREAMOFS offset)
{
	ht_tag_flags_s *flags=(ht_tag_flags_s*)find_atom(id), *fl;
	if (flags) {
		bounds b;
		b.w=60;
		b.h=14;
		b.x=(screen->size.w-b.w)/2;
		b.y=(screen->size.h-b.h)/2;
		ht_dialog *d=new ht_dialog();
		d->init(&b, (flags->bitidx==-1) ? flags->desc : 0, FS_KILLER | FS_TITLE);

		b.x=0;
		b.y=0;
		b.w-=2;
		b.h-=2;
		ht_uviewer *u=new ht_uviewer();
		u->init(&b, 0, stream, VC_EDIT);

		hts_mask *m=new hts_mask();
		m->init(stream);
		char *t, x[256];

		int width=0;
		fl=flags;
		if (fl->bitidx==-1) fl++;
		do {
			int l=strlen(fl->desc);
			if (l>width) width=l;
			fl++;
		} while (fl->desc);

		width=MAX(width, 25);
		width++;

		fl=flags;
		if (fl->bitidx==-1) fl++;
		do {
			t=x;
			int l=strlen(fl->desc);
			memmove(t, fl->desc, l);
			t+=l;
			while (t<x+width) *(t++)=' ';
			t=tag_make_edit_bit(t, offset, fl->bitidx);
			*t=0;
			m->add_mask(x);
			fl++;
		} while (fl->desc);

		u->insertsub(m);
		u->sendmsg(htm_complete_init, 0);
		d->insert(u);
		d->setpalette(window_palette_str);

		if (execute_dialog(d)==button_ok) u->edit_update();

		d->done();
		delete d;
		return 1;
	}
	return 0;
}

int ht_uviewer::ref_sel(ID id_low, ID id_high)
{
	return 0;
}

int ht_uviewer::search(ht_search *search, STREAMOFS start, STREAMOFS end)
{
	if (search!=&last_search) {
		if (has_last_search) free(last_search.pattern);
		has_last_search=1;
		last_search_end=end;
		last_search=*search;
		last_search.pattern=(byte*)malloc(last_search.pattern_length);
		memmove(last_search.pattern, search->pattern, last_search.pattern_length);
	}
	hts_sub *sub=first_sub;
	STREAMOFS ofs;
	dword size;
	while (sub) {
		if (sub->search(search, start, end, &ofs, &size)==SR_FOUND) {
			jump(ofs, this);
			if (size) select_set(ofs, ofs+size);
			return 1;
		}
		sub=sub->next;
	}
	infobox("string not found.");
	return 0;
}

void ht_uviewer::select_add(STREAMOFS start, STREAMOFS end)
{
	if (end<start) {
		STREAMOFS temp=start;
		start=end;
		end=temp;
	}
	if (end==sel_start) {
		sel_start=start;
	} else if (start==sel_end) {
		sel_end=end;
	} else if (start==sel_start) {
		sel_start=end;
	} else if (end==sel_end) {
		sel_end=start;
	} else {
		sel_start=start;
		sel_end=end;
	}
	if (sel_start>sel_end) {
		STREAMOFS temp=sel_start;
		sel_start=sel_end;
		sel_end=temp;
	}
}

void ht_uviewer::select_get(STREAMOFS *start, STREAMOFS *end)
{
	*start=sel_start;
	*end=sel_end;
}

void ht_uviewer::select_set(STREAMOFS start, STREAMOFS end)
{
	sel_start=start;
	sel_end=end;
}

int ht_uviewer::set_cursor(hts_sub *sub, ID id1, ID id2)
{
	return set_cursor_ex(sub, id1, id2, -1, -1);
}

int ht_uviewer::set_cursor_ex(hts_sub *sub, ID id1, ID id2, int tag_idx, int tag_group)
{
	cursorline_dirty();
	hts_sub *tsub=top_sub;
	ID tid1=top_id1, tid2=top_id2;
	int ty=0;
/* test if cursor is already on screen */
	if (cursor_state==cursor_state_visible) {
		do {
			if ((tsub==sub) && (tid1==id1) && (tid2==id2)) {
				cursor_sub=sub;
				cursor_id1=id1;
				cursor_id2=id2;
				if (tag_group!=-1) cursor_tag_group=tag_group;
				adjust_cursor_group();
				if (tag_idx!=-1) cursor_tag_idx=tag_idx;
				adjust_cursor_idx();
				cursor_ypos=ty;
				update_misc_info();
				update_visual_info();
				check_cursor_visibility();
				cursorline_dirty();
				dirtyview();
				return 1;
			}
		} while ((next_line(&tsub, &tid1, &tid2, 1)) && (ty++<size.h-1));
	}
/**/
	char line[1024];	/* FIXME: possible buffer overflow ! */
	char *e;
	sub->getline(line, id1, id2);
	e=tag_get_selectable_tag(line, 0, 0);
	if (!e) return 0;
	cursor_sub=sub;
	cursor_id1=id1;
	cursor_id2=id2;
	if (tag_group!=-1) cursor_tag_group=tag_group;
	adjust_cursor_group();
	if (tag_idx!=-1) cursor_tag_idx=tag_idx;
	adjust_cursor_idx();
	cursor_ypos=0;
//	cursor_ypos=center_view(sub, id1, id2);
	top_sub=sub;
	top_id1=id1;
	top_id2=id2;
	update_misc_info();
	update_visual_info();
	check_cursor_visibility();
	cursorline_dirty();
	dirtyview();
	return 1;
}

void ht_uviewer::scroll_up(int n)
{
	cursor_ypos+=prev_line(&top_sub, &top_id1, &top_id2, n);
	cursorline_dirty();
	check_cursor_visibility();
}

void ht_uviewer::scroll_down(int n)
{
	cursor_ypos-=next_line(&top_sub, &top_id1, &top_id2, n);
	cursorline_dirty();
	check_cursor_visibility();
}

void ht_uviewer::update_micropos()
{
	cursorline_get();
	char *e=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	if (e) {
		int s=tag_get_microsize(e);
		if (cursor_tag_micropos>=s) cursor_tag_micropos=s-1;
	}
}

void ht_uviewer::update_misc_info()
{
	cursorline_get();
	char *e=tag_get_selectable_tag(cursor_line, cursor_tag_idx, cursor_tag_group);
	if (e) {
		cursor_tag_class=tag_get_class(e);
		switch (cursor_tag_class) {
			case tag_class_edit:
				cursor_tag_offset=tag_get_offset(e);
				break;
			case tag_class_sel:
				tag_get_id(e, &cursor_tag_id.low, &cursor_tag_id.high);
				break;
		}
	}
}

void ht_uviewer::update_visual_info()
{
	cursorline_get();
	char *s, *t=cursor_line;
	int v=0, vl;
	int i=0, g=0;
	while ((s=tag_findnext(t))) {
		int cl=tag_get_class(s);
		if (s[1]==HT_TAG_GROUP) {
			i=0;
			g++;
		}
		v+=s-t;
		vl=tag_get_vlen(s);
		if ((i==cursor_tag_idx) && (g==cursor_tag_group) && (cl!=tag_class_no)) break;
		v+=vl;
		t=s+tag_get_len(s);
		if (cl!=tag_class_no) i++;
	}

	cursor_visual_xpos=v;
	cursor_visual_length=vl;
}

void ht_uviewer::update_ypos()
{
	hts_sub *sub=top_sub;
	ID id1=top_id1, id2=top_id2;
	int y=0;
	while ((next_line(&sub, &id1, &id2, 1)) && (y<size.h)) {
		if ((cursor_sub==sub) && (cursor_id1==id1) && (cursor_id2==id2)) {
			cursor_ypos=y;
			break;
		}
		y++;
	}
}

struct ht_uviewer_vs {
	int edit;
	hts_sub *first_sub, *last_sub;
/* top line position */
	hts_sub *top_sub;
	ID top_id1, top_id2;
/* cursor line and tag position */
	int cursor_state;
	int cursor_ypos;
	hts_sub *cursor_sub;
	ID cursor_id1, cursor_id2;
	int cursor_tag_idx;
	int cursor_tag_group;
/* selection*/
	STREAMOFS sel_start;
	STREAMOFS sel_end;
};

void ht_uviewer::view_state_load(void *data)
{
	ht_uviewer_vs *vs=(ht_uviewer_vs *)data;
	edit=vs->edit;
	first_sub=vs->first_sub;
	last_sub=vs->last_sub;
/* top line position */
	top_sub=vs->top_sub;
	top_id1=vs->top_id1;
	top_id2=vs->top_id2;
/* cursor line and tag position */
	cursor_state=vs->cursor_state;
	cursor_ypos=vs->cursor_ypos;
	cursor_sub=vs->cursor_sub;
	cursor_id1=vs->cursor_id1;
	cursor_id2=vs->cursor_id2;
	cursor_tag_idx=vs->cursor_tag_idx;
	cursor_tag_group=vs->cursor_tag_group;
/* selection*/
	sel_start=vs->sel_start;
	sel_end=vs->sel_end;
/**/
	cursorline_dirty();
	update_misc_info();
	update_visual_info();
}

void *ht_uviewer::view_state_create()
{
	ht_uviewer_vs *vs=new ht_uviewer_vs;
	vs->edit=edit;
	vs->first_sub=first_sub;
	vs->last_sub=last_sub;
/* top line position */
	vs->top_sub=top_sub;
	vs->top_id1=top_id1;
	vs->top_id2=top_id2;
/* cursor line and tag position */
	vs->cursor_state=cursor_state;
	vs->cursor_ypos=cursor_ypos;
	vs->cursor_sub=cursor_sub;
	vs->cursor_id1=cursor_id1;
	vs->cursor_id2=cursor_id2;
	vs->cursor_tag_idx=cursor_tag_idx;
	vs->cursor_tag_group=cursor_tag_group;
/* selection*/
	vs->sel_start=sel_start;
	vs->sel_end=sel_end;
/**/
	return vs;
}

/*
 *	CLASS ht_logviewer
 */

void ht_logviewer::init(bounds *b, int _maxlinecount)
{
	ht_viewer::init(b, "log", 0, 0);
	VIEW_DEBUG_NAME("ht_logviewer");
	maxlinecount=_maxlinecount;
	lines=new ht_string_collection();
	ofs=0;
	xofs=0;
}

void ht_logviewer::done()
{
	delete lines;
	ht_viewer::done();
}

int ht_logviewer::cursor_up(int n)
{
	ofs-=n;
	if (ofs<0) ofs=0;
	return n;
}

int ht_logviewer::cursor_down(int n)
{
	int c=lines->count();
	if (c>=size.h) {
		ofs+=n;
		if (ofs>c-size.h) ofs=lines->count()-size.h;
	}
	return n;
}

void ht_logviewer::deletefirstline()
{
	lines->del(0);
}

void ht_logviewer::draw()
{
	clear(windialog_color_body);
	int c=lines->count();
	for (int i=0; i<size.h; i++) {
		if (i+ofs>=c) break;
		char *s=lines->get(i+ofs);
		int l=strlen(s);
		if (xofs<l) print(0, i, windialog_color_body, s+xofs);
	}
}

void ht_logviewer::insertline(char *line)
{
	int c=lines->count();
	if (c>=maxlinecount) deletefirstline();
	lines->insert(line);
	if (c>=size.h-1) {
		ofs=c-size.h+1;
	}
	dirtyview();
}

void ht_logviewer::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Up:
				cursor_up(1);
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
				cursor_down(1);
				dirtyview();
				clearmsg(msg);
				return;
			case K_PageUp:
				cursor_up(size.h);
				dirtyview();
				clearmsg(msg);
				return;
			case K_PageDown:
				cursor_down(size.h);
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right: case K_Control_Right:
				xofs+=2;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Left: case K_Control_Left:
				if (xofs-2>=0) xofs-=2;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Control_PageUp:
				ofs=0;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Control_PageDown:
				ofs=lines->count()-size.h;
				if (ofs<0) ofs=0;
				dirtyview();
				clearmsg(msg);
				return;
		}
	}
	ht_viewer::handlemsg(msg);
}

/*
 *	CLASS hts_sub
 */

void hts_sub::init(ht_stream *_stream)
{
	object::init();
	prev=0;
	next=0;
	stream=_stream;
}

void hts_sub::done()
{
	object::done();
}

int hts_sub::convert_addr_to_id(VIEWERADDR addr, ID *id1, ID *id2)
{
	return 0;
}

int hts_sub::convert_id_to_addr(ID id1, ID id2, VIEWERADDR *addr)
{
	return 0;
}

int hts_sub::getline(char *line, ID id1, ID id2)
{
	abstract();
	return 0;
}

void hts_sub::first_line_id(ID *id1, ID *id2)
{
	abstract();
}

void hts_sub::last_line_id(ID *id1, ID *id2)
{
	abstract();
}

int hts_sub::prev_line_id(ID *id1, ID *id2, int n)
{
	abstract();
	return 0;
}

int hts_sub::next_line_id(ID *id1, ID *id2, int n)
{
	abstract();
	return 0;
}

int hts_sub::search(ht_search *s, STREAMOFS start, STREAMOFS end, STREAMOFS *offset, dword *size)
{
	return SR_NOT_SUPPORTED;
}

/*
 *	CLASS hts_linear_sub
 */

void hts_linear_sub::init(ht_stream *stream, STREAMOFS ofs, int size)
{
	hts_sub::init(stream);
	fofs=ofs;
	fsize=size;
}

void hts_linear_sub::done()
{
	hts_sub::done();
}

int hts_linear_sub::search(ht_search *s, STREAMOFS start, STREAMOFS end, STREAMOFS *offset, dword *size)
{
	if (s->type==ST_BIN) {
		bounds b;
		get_std_progress_indicator_metrics(&b);
		ht_progress_indicator *progress_indicator=new ht_progress_indicator();
		progress_indicator->init(&b, "searching, %2d%% complete...");
		int ret=SR_NOT_FOUND;
		int fl=(s->caseinsensitive) ? SF_CASEINSENSITIVE: 0;
		if (start<fofs) start=fofs;
		if (end>fofs+fsize-1) end=fofs+fsize-1;
		if ((fsize) && (start<end) && (search_bin(stream, start, end,
		s->pattern_length, s->pattern, offset, fl, progress_indicator))) {
			ret=SR_FOUND;
			*size=s->pattern_length;
		}
		progress_indicator->done();
		delete progress_indicator;
		return ret;
	}
	return SR_NOT_SUPPORTED;
}

/*
 *	CLASS hts_hex
 */

void hts_hex::init(ht_stream *stream, STREAMOFS ofs, int size)
{
	hts_linear_sub::init(stream, ofs, size);
}

void hts_hex::done()
{
	hts_linear_sub::done();
}

int hts_hex::convert_addr_to_id(VIEWERADDR addr, ID *id1, ID *id2)
{
	if ((addr>=fofs) && (addr<fofs+fsize)) {
		*id1=addr&~15;
		*id2=0;
		return 1;
	}
	return 0;
}

int hts_hex::convert_id_to_addr(ID id1, ID id2, VIEWERADDR *addr)
{
	*addr=id1;
	return 1;
}

int hts_hex::getline(char *line, ID id1, ID id2)
{
	ID ofs=id1;
	dword c=MIN(16, (int)(fofs+fsize-ofs));
	if (c<=0) return 0;
	char *l=line;
	l=mkhexd(l, ofs);
	*l++=' ';
	for (dword i=0; i<16; i++) {
		if (i<c) {
			l=tag_make_edit_byte(l, ofs+i);
		} else {
			*l++=' ';
			*l++=' ';
		}
		if (i<c-1) {
			if (i==7) {
				l=tag_make_edit_selvis(l, ofs+i, '-');
			} else {
				l=tag_make_edit_selvis(l, ofs+i, ' ');
			}
		} else *l++=' ';
	}
	l=tag_make_group(l);
	for (dword i=0; i<16; i++) {
		if (i<c) {
			l=tag_make_edit_char(l, ofs+i);
		} else {
			*l++=' ';
		}
	}
	*l=0;
	return 1;
}

void hts_hex::first_line_id(ID *id1, ID *id2)
{
	(*id1)=fofs;
	(*id2)=0;
}

void hts_hex::last_line_id(ID *id1, ID *id2)
{
	if (fsize) {
		(*id1)=((fsize-1) & ~0xf)+fofs;
		(*id2)=0;
	} else {
		(*id1)=fofs;
		(*id2)=0;
	}
}

int hts_hex::prev_line_id(ID *id1, ID *id2, int n)
{
	int c=0;
	while (n--) {
		if (*id1-16>*id1) break;
		if (*id1-16<fofs) break;
		(*id1)-=16;
		c++;
	}
	return c;
}

int hts_hex::next_line_id(ID *id1, ID *id2, int n)
{
	int c=0;
	while (n--) {
		if (*id1+16>=fofs+fsize) break;
		(*id1)+=16;
		c++;
	}
	return c;
}

/*
 *	CLASS hts_mask
 */

void hts_mask::init(ht_stream *stream)
{
	hts_sub::init(stream);
}

void hts_mask::done()
{
	hts_sub::done();
}

int hts_mask::getline(char *line, ID id1, ID id2)
{
	tag_strcpy(line, ((hts_mask_data*)masks.get(id1))->tagstring);
	return 1;
}

void hts_mask::first_line_id(ID *id1, ID *id2)
{
	*id1=0;
	*id2=0;
}

void hts_mask::last_line_id(ID *id1, ID *id2)
{
	*id1=masks.count()-1;
	*id2=0;
}

int hts_mask::prev_line_id(ID *id1, ID *id2, int n)
{
	int r;
	if (*id1<(dword)n) {
		r=*id1;
		*id1=0;
	} else {
		r=n;
		*id1-=n;
	}
	return r;
}

int hts_mask::next_line_id(ID *id1, ID *id2, int n)
{
	int r=n;
	*id1+=n;
	if ((int)*id1>masks.count()-1) {
		r-=*id1-masks.count()+1;
		*id1=masks.count()-1;
	}
	return r;
}

void hts_mask::add_mask(char *tagstr)
{
	hts_mask_data *d=new hts_mask_data();
	d->tagstring=tag_strdup(tagstr);
	masks.insert(d);
}

void hts_mask::add_mask_table(char **tagstr)
{
	while (*tagstr) add_mask(*(tagstr++));
}

void hts_mask::add_staticmask(char *statictag_str, STREAMOFS relocation_ofs)
{
	char tag_str[1024];	/* FIXME: possible buffer overflow */
	statictag_to_tag(statictag_str, tag_str, relocation_ofs);
	hts_mask_data *d=new hts_mask_data();
	d->tagstring=tag_strdup(tag_str);
	masks.insert(d);
}

void hts_mask::add_staticmask_table(char **statictag_str, STREAMOFS relocation_ofs)
{
	while (*statictag_str) add_staticmask(*(statictag_str++), relocation_ofs);
}

/*
 *	CLASS hts_unasm
 */

void hts_unasm::init(ht_stream *file, STREAMOFS ofs, int size, disassembler *dis, int _display_style)
{
	hts_linear_sub::init(file, ofs, size);
	unasm=dis;
	display_style=_display_style;
}

void hts_unasm::done()
{
	delete unasm;
	hts_linear_sub::done();
}

int hts_unasm::convert_addr_to_id(VIEWERADDR addr, ID *id1, ID *id2)
{
	if (((addr&~15)>=fofs) && ((addr&~15)<fofs+fsize)) {
		*id1=addr&~15;
		*id2=0;
		return 1;
	}
	return 0;
}

int hts_unasm::convert_id_to_addr(ID id1, ID id2, VIEWERADDR *addr)
{
	*addr=id1;
	return 1;
}

int hts_unasm::getline(char *line, ID id1, ID id2)
{
	dword ofs=id1;
	unsigned char buf[15];
	int c=MIN(15, (int)(fofs+fsize-ofs));
	if (c<=0) return 0;
	stream->seek(ofs);
	c=stream->read(buf, c);
	CPU_ADDR caddr;
	caddr.addr32=0;
	char *s;
	char *l=line;
	if (c) {
		dis_insn *insn=unasm->decode(buf, c, caddr);
		s=unasm->str(insn, display_style);
		c=unasm->getsize(insn);
	} else {
		s="db ?";
		c=0;
	}
	l=mkhexd(l, ofs);
	*l++=' ';
	for (int i=0; i<15; i++) {
		if (i<c) {
			l=tag_make_edit_byte(l, ofs+i);
		} else {
			*l++=' ';
			*l++=' ';
		}
	}
	*l++=' ';
	strcpy(l, s);
//     l+=sprintf(l, " %s", s);
	return 1;
}

void hts_unasm::first_line_id(ID *id1, ID *id2)
{
	(*id1)=fofs;
	(*id2)=0;
}

void hts_unasm::last_line_id(ID *id1, ID *id2)
{
	(*id1)=fofs+fsize;
	(*id2)=0;
	prev_line_id(id1, id2, 1);
}

int hts_unasm::prev_line_id(ID *id1, ID *id2, int n)
{
	dword *ofs=id1;
	unsigned char buf[15*50], *bufp=buf;
	int offsets[15*50];
	int *of=offsets;
	int r=n<6 ? 6*15 : n*15;
	ADDR o=*ofs-r;
	int c=r, d;
	int s;
	if (*ofs<fofs+r) {
		c-=fofs-o;
		o=fofs;
	}
	if (o+c>fofs+fsize) {
		c=fofs+fsize-o;
	}
	stream->seek(o);
	d=stream->read(buf, c);

	CPU_ADDR caddr;
	caddr.addr32=0;
	do {
		if (d>0) {
			dis_insn *insn=unasm->decode(bufp, d, caddr);
			s=unasm->getsize(insn);
			d-=s;
		} else {
			s=1;
		}
		*(of++)=o;
		o+=s;
		bufp+=s;
	} while (o<=*ofs);
	if (of-n-1<offsets) {
		*ofs=*(offsets);
		return of-offsets-1;
	} else {
		*ofs=*(of-n-1);
		return n;
	}
}

int hts_unasm::next_line_id(ID *id1, ID *id2, int n)
{
	dword *ofs=id1;
	unsigned char buf[15];
	int c=0, s, z;
	CPU_ADDR caddr;
	caddr.addr32=0;
	while (n--) {
		z=MIN(15, (int)(fofs+fsize-*ofs));
		stream->seek(*ofs);
		z=stream->read(buf, z);
		if (z) {
			dis_insn *insn=unasm->decode(buf, z, caddr);
			s=unasm->getsize(insn);
		} else {
			s=1;
		}
		if (*ofs+s>fofs+fsize-1) return c;
		*ofs+=s;
		c++;
	}
	return c;
}

