/* 
 *	HT Editor
 *	htcurses.cc (WIN32 implementation)
 *
 *	Copyright (C) 1999, 2000 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "global.h"
#include "htcurses.h"

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* virtual color pairs (fg/bg) */

void put_vc(char *dest1, word *dest2, char ch, int vc)
{
	int fg, bg;
	if (VC_GET_BASECOLOR(VCP_BACKGROUND(vc))==VC_TRANSPARENT) {
		bg=(byte)(((byte)*dest2)>>4);
	} else if (VC_GET_LIGHT(VCP_BACKGROUND(vc))) {
		bg=VC_GET_BASECOLOR(VCP_BACKGROUND(vc))+8;
	} else {
		bg=VC_GET_BASECOLOR(VCP_BACKGROUND(vc));
	}
	if (VC_GET_BASECOLOR(VCP_FOREGROUND(vc))==VC_TRANSPARENT) {
		fg=(byte)(((byte)*dest2)&0xf);
	} else if (VC_GET_LIGHT(VCP_FOREGROUND(vc))) {
		fg=VC_GET_BASECOLOR(VCP_FOREGROUND(vc))+8;
	} else {
		fg=VC_GET_BASECOLOR(VCP_FOREGROUND(vc));
	}
	*dest1=(unsigned char)ch;
	*dest2=(unsigned char)((bg<<4)|fg);
}

vcp vcp_mix(vcp base, vcp layer)
{
	int fg, bg;
	if (VC_GET_BASECOLOR(VCP_FOREGROUND(layer))==VC_TRANSPARENT) {
		fg=VCP_FOREGROUND(base);
	} else {
		fg=VCP_FOREGROUND(layer);
	}
	if (VC_GET_BASECOLOR(VCP_BACKGROUND(layer))==VC_TRANSPARENT) {
		bg=VCP_BACKGROUND(base);
	} else {
		bg=VCP_BACKGROUND(layer);
	}
	return VCP(fg, bg);
}

/*
 *	CLASS screendrawbuf
 */

screendrawbuf::screendrawbuf()
{
	bounds b;
	CONSOLE_SCREEN_BUFFER_INFO screen_info;

	buf_attr=NULL;
	buf_char=NULL;
	cursorx=0;
	cursory=0;

	output = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(output, &screen_info);

	b.x=0;
	b.y=0;
	b.w=screen_info.dwSize.X;
	b.h=screen_info.dwSize.Y;
	setbounds(&b);

	show();
}

screendrawbuf::~screendrawbuf()
{
	if (buf_attr) delete buf_attr;
	if (buf_char) delete buf_char;
}

void screendrawbuf::drawbuffer(drawbuf *b, int x, int y, bounds *clipping)
{
	drawbufch *ch=b->buf;
	for (int iy=0; iy<b->size.h; iy++) {
		word *k_attr = buf_attr+x+(iy+y)*size.w;
		char *k_char = buf_char+x+(iy+y)*size.w;
		if (y+iy>=clipping->y+clipping->h) break;
		if (y+iy>=clipping->y)
		for (int ix=0; ix<b->size.w; ix++) {
			if ((x+ix<clipping->x+clipping->w) && (x+ix>=clipping->x)) {
				put_vc(k_char, k_attr, ch->ch, ch->c);
			}
			k_attr++;
			k_char++;
			ch++;
		}
	}
}

void screendrawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
	for (int i=0; i<h; i++) {
		word *k_attr = buf_attr+x+(i+y)*size.w;
		char *k_char = buf_char+x+(i+y)*size.w;
		if (y+i>=size.h) break;
		if (y+i>=0)
		for (int j=0; j<w; j++) {
			if (x+j>=0) put_vc(k_char, k_attr, ch, c);
			if (x+j>=size.w-1) break;
			k_attr++;
			k_char++;
		}
	}
}

void screendrawbuf::printchar(int x, int y, int c, int ch)
{
	word *k_attr = buf_attr+x+y*size.w;
	char *k_char = buf_char+x+y*size.w;
	put_vc(k_char, k_attr, ch, c);
}

int screendrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int n=0;
	word *k_attr = buf_attr+x+y*size.w;
	char *k_char = buf_char+x+y*size.w;
	while ((*text) && (n<l)) {
		put_vc(k_char, k_attr, (unsigned char)*text, c);
		k_attr++;
		k_char++;
		text++;
		n++;
	}
	return n;
}

int screendrawbuf::lprintw(int x, int y, int c, int l, int *text)
{
	int n=0;
	word *k_attr = buf_attr+x+y*size.w;
	char *k_char = buf_char+x+y*size.w;
	while ((*text) && (n<l)) {
		put_vc(k_char, k_attr, (unsigned char)*text, c);
		k_attr++;
		k_char++;
		text++;
		n++;
	}
	return n;
}

void screendrawbuf::resize(int rw, int rh)
{
    /* screens are not sizeable (?) */
}

void screendrawbuf::rmove(int rx, int ry)
{
    /* screens are not movable */
}

void screendrawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf_attr) delete buf_attr;
	if (buf_char) delete buf_char;
	buf_attr=(word *)malloc(size.w * size.h * sizeof(word));
	buf_char=(char *)malloc(size.w * size.h * sizeof(byte));
	fill(size.x, size.y, size.w, size.h, 7, ' ');
}

void screendrawbuf::show()
{
	COORD xy;
	dword wrr;

	xy.X=cursorx;
	xy.Y=cursory;
	SetConsoleCursorPosition(output, xy);
	xy.X=0;
	for (xy.Y=0; xy.Y<size.h; xy.Y++) {
		WriteConsoleOutputAttribute(output, buf_attr+(xy.Y*size.w), size.w, xy, &wrr);
		WriteConsoleOutputCharacter(output, buf_char+(xy.Y*size.w), size.w, xy, &wrr);
	}
}

void screendrawbuf::getcursor(int *x, int *y)
{
	*x=cursorx;
	*y=cursory;
}

void screendrawbuf::hidecursor()
{
	setcursor(0, 0);
}

void screendrawbuf::setcursor(int x, int y)
{
	cursorx=x;
	cursory=y;
}

void screendrawbuf::showcursor()
{
}


/*
 *	CLASS drawbuf
 */

drawbuf::drawbuf(bounds *b)
{
	buf=0;
	setbounds(b);
}

drawbuf::~drawbuf()
{
	if (buf) {
		delete buf;
	}
}

void drawbuf::printchar(int x, int y, int c, int ch)
{
	drawbufch *b=buf+(x-size.x)+(y-size.y)*size.w;
	b->ch=ch;
	b->c=vcp_mix(b->c, c);
}

int drawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int ox=0;
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
//	if (y<size.h) {
		while ((*text) && (ox<l)) {
//          	if (x+ox<size.w) {
				b->ch=(byte)*text;
				b->c=vcp_mix(b->c, c);
//			}
			text++;
			ox++;
			b++;
		}
//	}
	return ox;
}

int drawbuf::lprintw(int x, int y, int c, int l, int *text)
{
	int ox=0;
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
//	if (y<size.h) {
		while ((*text) && (ox<l)) {
//          	if (x+ox<size.w) {
				b->ch=(byte)*text;
				b->c=vcp_mix(b->c, c);
//			}
			text++;
			ox++;
			b++;
		}
//	}
	return ox;
}

void drawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	for (int iy=y; iy<y+h; iy++) {
		for (int ix=x; ix<x+w; ix++) {
			b->ch=ch;
			b->c=vcp_mix(b->c, c);
			b++;
		}
	}
}

void drawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) {
		delete buf;
		buf=0;
	}
	if (size.w * size.h) {
		buf=(drawbufch*)malloc(sizeof *buf * size.w * size.h);
		drawbufch *bb=buf;
		for (int iy=0; iy<size.h; iy++) {
			for (int ix=0; ix<size.w; ix++) {
				bb->ch=' ';
				bb->c=VCP(VC_TRANSPARENT, VC_TRANSPARENT);
//               	bb->c=VC_BLACK;
				bb++;
			}
		}
	}
}

/*****************************************************************************
 ****************************************************************************/

genericdrawbuf::genericdrawbuf()
{
}

genericdrawbuf::~genericdrawbuf()
{
}

void genericdrawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
}

int genericdrawbuf::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return print(x, y, c, buf);
}

int genericdrawbuf::print(int x, int y, int c, char *text)
{
	return lprint(x, y, c, 0x7fffffff, text);
}

int genericdrawbuf::printw(int x, int y, int c, int *text)
{
	return lprintw(x, y, c, 0x7fffffff, text);
}

void genericdrawbuf::printchar(int x, int y, int c, int ch)
{
}

int genericdrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	return 0;
}

int genericdrawbuf::lprintw(int x, int y, int c, int l, int *text)
{
	return 0;
}

void genericdrawbuf::resize(int rw, int rh)
{
	size.w+=rw;
	size.h+=rh;
	setbounds(&size);
}

void genericdrawbuf::rmove(int rx, int ry)
{
	size.x+=rx;
	size.y+=ry;
}

void genericdrawbuf::setbounds(bounds *b)
{
	size=*b;
}

