/* 
 *	HT Editor
 *	htkeyb.cc (WIN32 implementation)
 *
 *	Copyright (C) 1999, 2000 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htkeyb.h"

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>

HANDLE input_handle;
INPUT_RECORD key_event_record;
bool key_pending=false;
bool k_shift_state=false;
bool k_ctrl_state=false;
bool k_alt_state=false;

int htkeypressed()
{
	DWORD read;
	if (key_pending) return true;
	while (1) {
		PeekConsoleInputA(input_handle, &key_event_record, 1, &read);
		if (!read) return false;
		ReadConsoleInputA(input_handle, &key_event_record, 1, &read);
		if (key_event_record.EventType & KEY_EVENT) {
			switch (key_event_record.Event.KeyEvent.wVirtualKeyCode) {
				case VK_SHIFT:
					k_shift_state=key_event_record.Event.KeyEvent.bKeyDown;
					break;
				case VK_CONTROL:
					k_ctrl_state=key_event_record.Event.KeyEvent.bKeyDown;
					break;
				case VK_MENU:
					k_alt_state=key_event_record.Event.KeyEvent.bKeyDown;
					break;
			}
			if (key_event_record.Event.KeyEvent.bKeyDown) {
				key_pending = true;
				return true;
			}
		}
	}
}

int ht_key_meta(bool shift)
{
    return (((k_shift_state) && (shift)) ? HT_VK_SHIFT : 0)
		+((k_ctrl_state) ? HT_VK_CTRL : 0)
		+((k_alt_state) ? HT_VK_ALT : 0);
}

int htgetkey()
{
	while (!htkeypressed());
	key_pending = false;
	if (key_event_record.Event.KeyEvent.uChar.AsciiChar) {
		return key_event_record.Event.KeyEvent.uChar.AsciiChar + ht_key_meta(false);
	} else {
		return key_event_record.Event.KeyEvent.wVirtualKeyCode + HT_VK + ht_key_meta(true);
	}
}

int htalt2key(int key)
{
	switch (key) {
		case K_Alt_A: return 'a';
		case K_Alt_B: return 'b';
		case K_Alt_C: return 'c';
		case K_Alt_D: return 'd';
		case K_Alt_E: return 'e';
		case K_Alt_F: return 'f';
		case K_Alt_G: return 'g';
		case K_Alt_H: return 'h';
		case K_Alt_I: return 'i';
		case K_Alt_J: return 'j';
		case K_Alt_K: return 'k';
		case K_Alt_L: return 'l';
		case K_Alt_M: return 'm';
		case K_Alt_N: return 'n';
		case K_Alt_O: return 'o';
		case K_Alt_P: return 'p';
		case K_Alt_Q: return 'q';
		case K_Alt_R: return 'r';
		case K_Alt_S: return 's';
		case K_Alt_T: return 't';
		case K_Alt_U: return 'u';
		case K_Alt_V: return 'v';
		case K_Alt_W: return 'w';
		case K_Alt_X: return 'x';
		case K_Alt_Y: return 'y';
		case K_Alt_Z: return 'z';
	}
	return -1;
}

int htlkey2alt(int key)
{
	switch (key) {
		case 'a': return K_Alt_A;
		case 'b': return K_Alt_B;
		case 'c': return K_Alt_C;
		case 'd': return K_Alt_D;
		case 'e': return K_Alt_E;
		case 'f': return K_Alt_F;
		case 'g': return K_Alt_G;
		case 'h': return K_Alt_H;
		case 'i': return K_Alt_I;
		case 'j': return K_Alt_J;
		case 'k': return K_Alt_K;
		case 'l': return K_Alt_L;
		case 'm': return K_Alt_M;
		case 'n': return K_Alt_N;
		case 'o': return K_Alt_O;
		case 'p': return K_Alt_P;
		case 'q': return K_Alt_Q;
		case 'r': return K_Alt_R;
		case 's': return K_Alt_S;
		case 't': return K_Alt_T;
		case 'u': return K_Alt_U;
		case 'v': return K_Alt_V;
		case 'w': return K_Alt_W;
		case 'x': return K_Alt_X;
		case 'y': return K_Alt_Y;
		case 'z': return K_Alt_Z;
	}
	return -1;
}

int htkey2alt(int key)
{
	if ((key>='a') && (key<='z')) {
		return htlkey2alt(key);
	} else if ((key>='A') && (key<='Z')) {
		return htlkey2alt(key+32);
	}
	return -1;
}

void init_keyb()
{
	input_handle = GetStdHandle(STD_INPUT_HANDLE);
}

void done_keyb()
{
}
