/* 
 *	HT Editor
 *	pestruct.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __PESTRUCT_H_
#define __PESTRUCT_H_

#include "global.h"
#include "tools.h"

#include <time.h>

#include "coff_s.h"

#define PE_MAGIC MAGICD("PE\0\0")

//
// Directory format.
//

typedef struct	PE_DATA_DIRECTORY {
    dword	  address __attribute__ ((packed));
    dword	  size __attribute__ ((packed));
};

#define PE_NUMBEROF_DIRECTORY_ENTRIES	  	16

//
// Optional header format.
//

typedef struct	PE_OPTIONAL_HEADER32_NT {
// NT additional fields.
	dword image_base __attribute__ ((packed));
	dword section_alignment __attribute__ ((packed));
	dword file_alignment __attribute__	((packed));
	word major_os_version	__attribute__ ((packed));
	word minor_os_version	__attribute__ ((packed));
	word major_image_version	__attribute__ ((packed));
	word minor_image_version	__attribute__ ((packed));
	word major_subsystem_version __attribute__ ((packed));
	word minor_subsystem_version __attribute__ ((packed));
	dword win32_version __attribute__ ((packed));
	dword image_size __attribute__ ((packed));
	dword header_size __attribute__ ((packed));
	dword checksum __attribute__	((packed));
	word	subsystem __attribute__ ((packed));
	word dll_characteristics __attribute__ ((packed));
	dword stack_reserve_size __attribute__ ((packed));
	dword stack_commit_size __attribute__ ((packed));
	dword heap_reserve_size __attribute__ ((packed));
	dword heap_commit_size __attribute__ ((packed));
	dword loader_flags __attribute__ ((packed));
	dword directory_count __attribute__ ((packed));
	PE_DATA_DIRECTORY directory[PE_NUMBEROF_DIRECTORY_ENTRIES] __attribute__ ((packed));
};

// Subsystem Values

#define PE_SUBSYSTEM_NATIVE			   1	  // Image doesn't require a subsystem.
#define PE_SUBSYSTEM_WINDOWS_GUI		   2	  // Image runs in the Windows GUI subsystem.
#define PE_SUBSYSTEM_WINDOWS_CUI		   3	  // Image runs in the Windows character subsystem.
#define PE_SUBSYSTEM_OS2_CUI			   5	  // image runs in the OS/2 character subsystem.
#define PE_SUBSYSTEM_POSIX_CUI		   7	  // image run  in the Posix character subsystem.
#define PE_SUBSYSTEM_RESERVED8		   8	  // image run  in the 8 subsystem.


// Directory Entries

#define PE_DIRECTORY_ENTRY_EXPORT			 0	  // Export Directory
#define PE_DIRECTORY_ENTRY_IMPORT			 1	  // Import Directory
#define PE_DIRECTORY_ENTRY_RESOURCE		 2	  // Resource Directory
#define PE_DIRECTORY_ENTRY_EXCEPTION		 3	  // Exception Directory
#define PE_DIRECTORY_ENTRY_SECURITY		 4	  // Security Directory
#define PE_DIRECTORY_ENTRY_BASERELOC		 5	  // Base Relocation Table
#define PE_DIRECTORY_ENTRY_DEBUG			 6	  // Debug Directory
#define PE_DIRECTORY_ENTRY_COPYRIGHT		 7	  // Description String
#define PE_DIRECTORY_ENTRY_GLOBALPTR		 8	  // Machine Value (MIPS GP)
#define PE_DIRECTORY_ENTRY_TLS			 9	  // TLS Directory
#define PE_DIRECTORY_ENTRY_LOAD_CONFIG		10	  // Load Configuration Directory
#define PE_DIRECTORY_ENTRY_BOUND_IMPORT		11	  // Bound Import Directory in headers
#define PE_DIRECTORY_ENTRY_IAT			12	  // Import Address Table
#define PE_DIRECTORY_ENTRY_DELAY_IMPORT		13	  // Delay Import Directory
#define PE_DIRECTORY_ENTRY_COM			14	  // COM+

/*
 *	Export
 */

struct PE_EXPORT_DIRECTORY {
	dword characteristics __attribute__ ((packed));
	time_t timestamp __attribute__ ((packed));
	word major_version __attribute__ ((packed));
	word minor_version __attribute__ ((packed));
	dword name_address __attribute__ ((packed));
	dword ordinal_base __attribute__ ((packed));
	dword function_count __attribute__ ((packed));
	dword name_count __attribute__ ((packed));
	dword function_table_address __attribute__ ((packed));
	dword name_table_address __attribute__ ((packed));
	dword ordinal_table_address __attribute__ ((packed));
};

/*
 *	Import
 */

struct PE_THUNK_DATA	{
	union {
		dword forwarder_string __attribute__ ((packed));
		dword function_desc_address __attribute__ ((packed));
		dword ordinal __attribute__ ((packed));
		dword data_address __attribute__ ((packed));
	};
};

struct PE_IMPORT_DESCRIPTOR {
	union {
		dword characteristics __attribute__ ((packed)); 		// 0 for terminating null import descriptor
		dword original_first_thunk __attribute__ ((packed));  	// rva to original unbound IAT
	};
	time_t timestamp __attribute__ ((packed));		// 0 if not bound,
											// -1 if bound, and real date\time stamp
											//     in IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT (new BIND)
											// O.W. date/time stamp of DLL bound to (Old BIND)

	dword forwarder_chain __attribute__ ((packed));  	// -1 if no forwarders
	dword name __attribute__ ((packed));
	dword first_thunk __attribute__ ((packed)); 		// rva to IAT (if bound this IAT has actual addresses)
};

/*
 *	Delay Import
 */

struct PE_DELAY_IMPORT_DESCRIPTOR {
	dword attributes __attribute__ ((packed));
	dword name __attribute__ ((packed));
	dword module_handle __attribute__ ((packed));
	dword delay_iat __attribute__ ((packed));
	dword delay_int __attribute__ ((packed));
	dword bound_delay_import_table __attribute__ ((packed));
	dword unload_delay_import_table __attribute__ ((packed));
	time_t timestamp __attribute__ ((packed));
};

/*
 *	Resource
 */

struct PE_RESOURCE_DIRECTORY {
	dword characteristics __attribute__ ((packed));
	time_t timedate_stamp __attribute__ ((packed));
	word major_version __attribute__ ((packed));
	word minor_version __attribute__ ((packed));
	word name_count __attribute__ ((packed));
	word id_count __attribute__ ((packed));
//    PE_RESOURCE_DIRECTORY_ENTRY DirectoryEntries[];
};

#define PE_RESOURCE_NAME_IS_STRING			0x80000000
#define PE_RESOURCE_DATA_IS_DIRECTORY		0x80000000

struct PE_RESOURCE_DIRECTORY_ENTRY {
	dword   name __attribute__ ((packed));				// or id
	dword   offset_to_directory __attribute__ ((packed));	// or data
};

/*typedef struct PE_RESOURCE_DIRECTORY_STRING {
    word    Length;
    CHAR    NameString[ 1 ];
};


typedef struct PE_RESOURCE_DIR_STRING_U {
    word    Length;
    WCHAR   NameString[ 1 ];
};*/

struct PE_RESOURCE_DATA_ENTRY {
    dword	  offset_to_data __attribute__ ((packed));
    dword	  size __attribute__ ((packed));
    dword	  codepage __attribute__	((packed));
    dword	  reserved __attribute__	((packed));
};

#endif /* !__PESTRUCT_H_ */
