/* 
 *	HT Editor
 *	viewers.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "viewers.h"

#include "hthex.h"
#include "htelf.h"
#include "htcoff.h"
#include "htle.h"
#include "htmz.h"
#include "htne.h"
#include "htpe.h"

viewer_if *viewer_ifs[] =
{
	&hthex_if,
	&htelf_if,
	&htmz_if,
	&htcoff_if,
	&htpe_if,
	&htne_if,
	&htle_if,
	0
};

/*
 *	CLASS ht_viewergroup
 */

void	ht_viewergroup::init(bounds *b, int options, char *desc, ht_stream *file, viewer_if **_ifs)
{
	ht_xgroup::init(b, options, desc);
	VIEW_DEBUG_NAME("ht_viewergroup");
	ifs=_ifs;
	ht_viewergroup_stream=file;
	if (ifs) initall(ifs);
}

void ht_viewergroup::done()
{
	if (ifs) doneall();
	ht_xgroup::done();
}

void ht_viewergroup::initall(viewer_if **ifs)
{
/* init & insert all ifs */
	bounds b;
	getclientarea(&b);
	b.x=0;
	b.y=0;
	int j=0;
	viewer_if **i=ifs;
	while (*i) {
		if ((*i)->init) {
			views[j]=(*i)->init(&b, ht_viewergroup_stream, this);
			if (views[j]) {
				views[j]->sendmsg(htm_complete_init, 0);
				insert(views[j]);
			}
		} else views[j]=0;
		i++;
		j++;
	}
}

void ht_viewergroup::doneall()
{
/* destroy & remove all ifs */
	int j=0;
	viewer_if **i=ifs;
	while (*i) {
		if (views[j]) remove(views[j]);
		if ((*i)->done) {
			if (!views[j]) assert(0);
			(*i)->done(views[j]);
		} else if (views[j]) {
			views[j]->done();
			delete views[j];
		}
		i++;
		j++;
	}
}

